from dataclasses import dataclass, field

import numpy as np


@dataclass
class CameraData:
    """
    CameraData stores the parameters and configuration settings for an RGB-D camera.

    Attributes:
    ----------
    name : str
        The name of the camera instance. Default is an empty string.
    enable : bool
        Boolean flag to enable or disable the camera. Default is True.
    type : str
        Type of the camera. Default is "rgbd_camera".
    relative_to : str
        The reference frame relative to which the camera is placed. Default is "base_link".

    width : int
        The image width (in pixels) captured by the camera. Default is 128.
    height : int
        The image height (in pixels) captured by the camera. Default is 128.
    image_format : str
        The image format used (e.g., "R8G8B8"). Default is "R8G8B8".
    update_rate : int
        The rate (in Hz) at which the camera provides updates. Default is 10 Hz.
    horizontal_fov : float
        The horizontal field of view (in radians). Default is pi / 3.0.
    vertical_fov : float
        The vertical field of view (in radians). Default is pi / 3.0.

    clip_color : tuple[float, float]
        The near and far clipping planes for the color camera. Default is (0.01, 1000.0).
    clip_depth : tuple[float, float]
        The near and far clipping planes for the depth camera. Default is (0.05, 10.0).

    noise_mean : float | None
        The mean value of the Gaussian noise applied to the camera's data. Default is None (no noise).
    noise_stddev : float | None
        The standard deviation of the Gaussian noise applied to the camera's data. Default is None (no noise).

    publish_color : bool
        Whether or not to publish color data from the camera. Default is False.
    publish_depth : bool
        Whether or not to publish depth data from the camera. Default is False.
    publish_points : bool
        Whether or not to publish point cloud data from the camera. Default is False.

    spawn_position : tuple[float, float, float]
        The initial spawn position (x, y, z) of the camera relative to the reference frame. Default is (0, 0, 1).
    spawn_quat_xyzw : tuple[float, float, float, float]
        The initial spawn orientation of the camera in quaternion (x, y, z, w). Default is (0, 0.70710678118, 0, 0.70710678118).

    random_pose_rollouts_num : int
        The number of random pose rollouts. Default is 1.
    random_pose_mode : str
        The mode for random pose generation (e.g., "orbit"). Default is "orbit".
    random_pose_orbit_distance : float
        The fixed distance from the object in "orbit" mode. Default is 1.0.
    random_pose_orbit_height_range : tuple[float, float]
        The range of vertical positions (z-axis) in "orbit" mode. Default is (0.1, 0.7).
    random_pose_orbit_ignore_arc_behind_robot : float
        The arc angle (in radians) behind the robot to ignore when generating random poses. Default is pi / 8.
    random_pose_select_position_options : list[tuple[float, float, float]]
        A list of preset position options for the camera in random pose mode. Default is an empty list.
    random_pose_focal_point_z_offset : float
        The offset in the z-direction of the focal point when generating random poses. Default is 0.0.
    random_pose_rollout_counter : float
        A counter tracking the number of rollouts completed for random poses. Default is 0.0.
    """

    name: str = field(default_factory=str)
    enable: bool = field(default=True)
    type: str = field(default="rgbd_camera")
    relative_to: str = field(default="base_link")

    width: int = field(default=128)
    height: int = field(default=128)
    image_format: str = field(default="R8G8B8")
    update_rate: int = field(default=10)
    horizontal_fov: float = field(default=np.pi / 3.0)
    vertical_fov: float = field(default=np.pi / 3.0)

    clip_color: tuple[float, float] = field(default=(0.01, 1000.0))
    clip_depth: tuple[float, float] = field(default=(0.05, 10.0))

    noise_mean: float | None = None
    noise_stddev: float | None = None

    publish_color: bool = field(default=False)
    publish_depth: bool = field(default=False)
    publish_points: bool = field(default=False)

    spawn_position: tuple[float, float, float] = field(default=(0.0, 0.0, 1.0))
    spawn_quat_xyzw: tuple[float, float, float, float] = field(
        default=(0.0, 0.70710678118, 0.0, 0.70710678118)
    )

    random_pose_rollouts_num: int = field(default=1)
    random_pose_mode: str = field(default="orbit")
    random_pose_orbit_distance: float = field(default=1.0)
    random_pose_orbit_height_range: tuple[float, float] = field(default=(0.1, 0.7))
    random_pose_orbit_ignore_arc_behind_robot: float = field(default=np.pi / 8)
    random_pose_select_position_options: list[tuple[float, float, float]] = field(
        default_factory=list
    )
    random_pose_focal_point_z_offset: float = field(default=0.0)
    random_pose_rollout_counter: float = field(default=0.0)
