from dataclasses import dataclass, field

from env_manager.utils.types import Pose


@dataclass
class ObjectRandomizerData:
    """
    ObjectRandomizerData contains parameters for randomizing object properties during simulation.

    Attributes:
    ----------
    count : int
        The number of objects to randomize. Default is 0.
    random_pose : bool
        Flag indicating whether to randomize both the position and orientation of objects. Default is False.
    random_position : bool
        Flag indicating whether to randomize the position of objects. Default is False.
    random_orientation : bool
        Flag indicating whether to randomize the orientation of objects. Default is False.
    random_model : bool
        Flag indicating whether to randomize the model of the objects. Default is False.
    random_spawn_position_segments : list
        List of segments within which the object can be randomly spawned. Default is an empty list.
    random_spawn_position_update_workspace_centre : bool
        Flag indicating whether to update the workspace center during random spawning. Default is False.
    random_spawn_volume : tuple[float, float, float]
        The volume within which objects can be randomly spawned, defined by (x, y, z). Default is (0.5, 0.5, 0.5).
    models_rollouts_num : int
        The number of rollouts for randomizing models. Default is 0.
    """

    count: int = field(default=0)
    random_pose: bool = field(default=False)
    random_position: bool = field(default=False)
    random_orientation: bool = field(default=False)
    random_model: bool = field(default=False)
    random_spawn_position_segments: list = field(default_factory=list)
    random_spawn_position_update_workspace_centre: bool = field(default=False)
    random_spawn_volume: tuple[float, float, float] = field(default=(0.5, 0.5, 0.5))
    models_rollouts_num: int = field(default=0)
    random_color: bool = field(default=False)


@dataclass
class ObjectData:
    """
    ObjectData stores the base properties for any object in the simulation environment.

    Attributes:
    ----------
    name : str
        The name of the object. Default is "object".
    type : str
        The type of the object (e.g., "sphere", "box"). Default is an empty string.
    relative_to : str
        The reference frame relative to which the object is positioned. Default is "world".
    position : tuple[float, float, float]
        The position of the object in (x, y, z) coordinates. Default is (0.0, 0.0, 0.0).
    orientation : tuple[float, float, float, float]
        The orientation of the object as a quaternion (x, y, z, w). Default is (1.0, 0.0, 0.0, 0.0).
    static : bool
        Flag indicating if the object is static in the simulation (immovable). Default is False.
    randomize : ObjectRandomizerData
        Object randomizer settings for generating randomized object properties. Default is an empty ObjectRandomizerData instance.
    """

    name: str = field(default="object")
    type: str = field(default_factory=str)
    relative_to: str = field(default="world")
    position: tuple[float, float, float] = field(default=(0.0, 0.0, 0.0))
    orientation: tuple[float, float, float, float] = field(default=(1.0, 0.0, 0.0, 0.0))
    color: tuple[float, float, float, float] | None = field(default=None)
    static: bool = field(default=False)
    randomize: ObjectRandomizerData = field(default_factory=ObjectRandomizerData)


@dataclass
class PrimitiveObjectData(ObjectData):
    """
    PrimitiveObjectData defines the base properties for primitive objects (e.g., spheres, boxes) in the simulation.

    Attributes:
    ----------
    collision : bool
        Flag indicating whether the object participates in collision detection. Default is True.
    visual : bool
        Flag indicating whether the object has a visual representation in the simulation. Default is True.
    color : tuple[float, float, float, float]
        The color of the object, represented in RGBA format. Default is (0.8, 0.8, 0.8, 1.0).
    mass : float
        The mass of the object. Default is 0.1.
    """

    collision: bool = field(default=True)
    visual: bool = field(default=True)
    mass: float = field(default=0.1)


@dataclass
class SphereObjectData(PrimitiveObjectData):
    """
    SphereObjectData defines the specific properties for a spherical object in the simulation.

    Attributes:
    ----------
    radius : float
        The radius of the sphere. Default is 0.025.
    friction : float
        The friction coefficient of the sphere when in contact with surfaces. Default is 1.0.
    """

    radius: float = field(default=0.025)
    friction: float = field(default=1.0)


@dataclass
class PlaneObjectData(PrimitiveObjectData):
    """
    PlaneObjectData defines the specific properties for a planar object in the simulation.

    Attributes:
    ----------
    size : tuple[float, float]
        The size of the plane, defined by its width and length. Default is (1.0, 1.0).
    direction : tuple[float, float, float]
        The normal vector representing the direction the plane faces. Default is (0.0, 0.0, 1.0).
    friction : float
        The friction coefficient of the plane when in contact with other objects. Default is 1.0.
    """

    size: tuple = field(default=(1.0, 1.0))
    direction: tuple = field(default=(0.0, 0.0, 1.0))
    friction: float = field(default=1.0)


@dataclass
class CylinderObjectData(PrimitiveObjectData):
    """
    CylinderObjectData defines the specific properties for a cylindrical object in the simulation.

    Attributes:
    ----------
    radius : float
        The radius of the cylinder. Default is 0.025.
    length : float
        The length of the cylinder. Default is 0.1.
    friction : float
        The friction coefficient of the cylinder when in contact with surfaces. Default is 1.0.
    """

    radius: float = field(default=0.025)
    length: float = field(default=0.1)
    friction: float = field(default=1.0)


@dataclass
class BoxObjectData(PrimitiveObjectData):
    """
    BoxObjectData defines the specific properties for a box-shaped object in the simulation.

    Attributes:
    ----------
    size : tuple[float, float, float]
        The dimensions of the box in (width, height, depth). Default is (0.05, 0.05, 0.05).
    friction : float
        The friction coefficient of the box when in contact with surfaces. Default is 1.0.
    """

    size: tuple = field(default=(0.05, 0.05, 0.05))
    friction: float = field(default=1.0)


@dataclass
class ModelData(ObjectData):
    """
    MeshObjectData defines the specific properties for a mesh-based object in the simulation.

    Attributes:
    ----------
    texture : list[float]
        A list of texture coordinates or texture properties applied to the mesh. Default is an empty list.
    """

    texture: list[float] = field(default_factory=list)


@dataclass
class MeshData(ModelData):
    mass: float = field(default_factory=float)
    inertia: tuple[float, float, float, float, float, float] = field(
        default_factory=tuple
    )
    cm: Pose = field(default_factory=Pose)
    collision: str = field(default_factory=str)
    visual: str = field(default_factory=str)
    friction: float = field(default_factory=float)
    density: float = field(default_factory=float)
