from dataclasses import dataclass, field
from enum import Enum


@dataclass
class ToolData:
    name: str = field(default_factory=str)
    type: str = field(default_factory=str)


@dataclass
class GripperData(ToolData):
    pass


@dataclass
class ParallelGripperData(GripperData):
    pass


@dataclass
class MultifingerGripperData(GripperData):
    pass


@dataclass
class VacuumGripperData(GripperData):
    pass


class GripperEnum(Enum):
    parallel = ParallelGripperData
    mulrifinger = MultifingerGripperData
    vacuum = VacuumGripperData


@dataclass
class RobotRandomizerData:
    """
    RobotRandomizerData stores configuration parameters for randomizing robot properties during simulation.

    Attributes:
    ----------
    pose : bool
        Flag indicating whether to randomize the robot's pose (position and orientation). Default is False.
    spawn_volume : tuple[float, float, float]
        The volume within which the robot can be spawned, defined by (x, y, z) dimensions. Default is (1.0, 1.0, 0.0).
    joint_positions : bool
        Flag indicating whether to randomize the robot's joint positions. Default is False.
    joint_positions_std : float
        The standard deviation for randomizing the robot's joint positions. Default is 0.1.
    joint_positions_above_object_spawn : bool
        Flag indicating whether the robot's joint positions should be randomized to place the robot above the object's spawn position. Default is False.
    joint_positions_above_object_spawn_elevation : float
        The elevation above the object's spawn position when placing the robot's joints. Default is 0.2.
    joint_positions_above_object_spawn_xy_randomness : float
        The randomness in the x and y coordinates when placing the robot's joints above the object's spawn position. Default is 0.2.
    """

    pose: bool = field(default=False)
    spawn_volume: tuple[float, float, float] = field(default=(1.0, 1.0, 0.0))
    joint_positions: bool = field(default=False)
    joint_positions_std: float = field(default=0.1)
    joint_positions_above_object_spawn: bool = field(default=False)
    joint_positions_above_object_spawn_elevation: float = field(default=0.2)
    joint_positions_above_object_spawn_xy_randomness: float = field(default=0.2)


@dataclass
class RobotData:
    """
    RobotData stores the base properties and configurations for a robot in the simulation.

    Attributes:
    ----------
    name : str
        The name of the robot. Default is an empty string.
    urdf_string : str
        Optional parameter that can store a URDF. This parameter is overridden by the node parameters if set in the node configuration.
    spawn_position : tuple[float, float, float]
        The position where the robot will be spawned in (x, y, z) coordinates. Default is (0.0, 0.0, 0.0).
    spawn_quat_xyzw : tuple[float, float, float, float]
        The orientation of the robot in quaternion format (x, y, z, w) at spawn. Default is (0.0, 0.0, 0.0, 1.0).
    joint_positions : list[float]
        A list of the robot's joint positions. Default is an empty list.
    with_gripper : bool
        Flag indicating whether the robot is equipped with a gripper. Default is False.
    gripper_joint_positions : list[float] | float
        The joint positions for the gripper. Can be a list of floats or a single float. Default is an empty list.
    randomizer : RobotRandomizerData
        The randomization settings for the robot, allowing various properties to be randomized in simulation. Default is an instance of RobotRandomizerData.
    """

    name: str = field(default_factory=str)
    urdf_string: str = field(default_factory=str)
    spawn_position: tuple[float, float, float] = field(default=(0.0, 0.0, 0.0))
    spawn_quat_xyzw: tuple[float, float, float, float] = field(
        default=(0.0, 0.0, 0.0, 1.0)
    )
    joint_positions: list[float] = field(default_factory=list)
    with_gripper: bool = field(default=False)
    gripper_joint_positions: list[float] | float = field(default_factory=list)
    randomizer: RobotRandomizerData = field(default_factory=RobotRandomizerData)
