from dataclasses import dataclass, field

from .camera import CameraData
from .light import LightData
from .objects import ObjectData
from .robot import RobotData
from .terrain import TerrainData


@dataclass
class PluginsData:
    """
    PluginsData stores the configuration for various plugins used in the simulation environment.

    Attributes:
    ----------
    scene_broadcaster : bool
        Flag indicating whether the scene broadcaster plugin is enabled. Default is False.
    user_commands : bool
        Flag indicating whether the user commands plugin is enabled. Default is False.
    fts_broadcaster : bool
        Flag indicating whether the force torque sensor (FTS) broadcaster plugin is enabled. Default is False.
    sensors_render_engine : str
        The rendering engine used for sensors. Default is "ogre2".
    """

    scene_broadcaster: bool = field(default_factory=bool)
    user_commands: bool = field(default_factory=bool)
    fts_broadcaster: bool = field(default_factory=bool)
    sensors_render_engine: str = field(default="ogre2")


@dataclass
class SceneData:
    """
    SceneData contains the configuration and settings for the simulation scene.

    Attributes:
    ----------
    physics_rollouts_num : int
        The number of physics rollouts to perform in the simulation. Default is 0.
    gravity : tuple[float, float, float]
        The gravitational acceleration vector applied in the scene, represented as (x, y, z). Default is (0.0, 0.0, -9.80665).
    gravity_std : tuple[float, float, float]
        The standard deviation for the gravitational acceleration, represented as (x, y, z). Default is (0.0, 0.0, 0.0232).
    robot : RobotData
        The configuration data for the robot present in the scene. Default is an instance of RobotData.
    terrain : TerrainData
        The configuration data for the terrain in the scene. Default is an instance of TerrainData.
    light : LightData
        The configuration data for the lighting in the scene. Default is an instance of LightData.
    objects : list[ObjectData]
        A list of objects present in the scene, represented by their ObjectData configurations. Default is an empty list.
    camera : list[CameraData]
        A list of cameras in the scene, represented by their CameraData configurations. Default is an empty list.
    plugins : PluginsData
        The configuration data for various plugins utilized in the simulation environment. Default is an instance of PluginsData.
    """

    physics_rollouts_num: int = field(default=0)
    gravity: tuple[float, float, float] = field(default=(0.0, 0.0, -9.80665))
    gravity_std: tuple[float, float, float] = field(default=(0.0, 0.0, 0.0232))
    robot: RobotData = field(default_factory=RobotData)
    terrain: TerrainData = field(default_factory=TerrainData)
    light: LightData = field(default_factory=LightData)
    objects: list[ObjectData] = field(default_factory=list)
    camera: list[CameraData] = field(default_factory=list)
    plugins: PluginsData = field(default_factory=PluginsData)
