import numpy as np
from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo


class Sun(model_wrapper.ModelWrapper):
    """
    Sun is a class that represents a directional light source in the simulation, similar to the Sun.

    It can have both visual and light properties, with customizable parameters such as color, direction, and attenuation.

    Attributes:
    ----------
    world : scenario_gazebo.World
        The Gazebo world where the sun model will be inserted.
    name : str
        The name of the sun model. Default is "sun".
    direction : tuple[float, float, float]
        The direction of the sunlight, normalized. Default is (0.5, -0.25, -0.75).
    color : tuple[float, float, float, float]
        The RGBA color values for the light's diffuse color. Default is (1.0, 1.0, 1.0, 1.0).
    distance : float
        The distance from the origin to the sun. Default is 800.0.
    visual : bool
        If True, a visual representation of the sun will be added. Default is True.
    radius : float
        The radius of the visual representation of the sun. Default is 20.0.
    specular : float
        The intensity of the specular reflection. Default is 1.0.
    attenuation_range : float
        The maximum range for the light attenuation. Default is 10000.0.
    attenuation_constant : float
        The constant attenuation factor. Default is 0.9.
    attenuation_linear : float
        The linear attenuation factor. Default is 0.01.
    attenuation_quadratic : float
        The quadratic attenuation factor. Default is 0.001.

    Raises:
    -------
    RuntimeError:
        If the sun model fails to be inserted into the Gazebo world.

    Methods:
    --------
    get_sdf() -> str:
        Generates the SDF string used to describe the sun model in the Gazebo simulation.
    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "sun",
        direction: tuple[float, float, float] = (0.5, -0.25, -0.75),
        color: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0),
        distance: float = 800.0,
        visual: bool = True,
        radius: float = 20.0,
        specular: float = 1.0,
        attenuation_range: float = 10000.0,
        attenuation_constant: float = 0.9,
        attenuation_linear: float = 0.01,
        attenuation_quadratic: float = 0.001,
        **kwargs,
    ):
        # Get a unique model name
        model_name = get_unique_model_name(world, name)

        # Normalize direction
        np_direction = np.array(direction)
        np_direction = np_direction / np.linalg.norm(np_direction)

        # Initial pose
        initial_pose = scenario_core.Pose(
            (
                -np_direction[0] * distance,
                -np_direction[1] * distance,
                -np_direction[2] * distance,
            ),
            (1, 0, 0, 0),
        )

        # Create SDF string for the model
        sdf = self.get_sdf(
            model_name=model_name,
            direction=tuple(np_direction),
            color=color,
            visual=visual,
            radius=radius,
            specular=specular,
            attenuation_range=attenuation_range,
            attenuation_constant=attenuation_constant,
            attenuation_linear=attenuation_linear,
            attenuation_quadratic=attenuation_quadratic,
        )

        # Insert the model
        ok_model = world.to_gazebo().insert_model_from_string(
            sdf, initial_pose, model_name
        )
        if not ok_model:
            raise RuntimeError("Failed to insert " + model_name)

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        model_wrapper.ModelWrapper.__init__(self, model=model)

    @classmethod
    def get_sdf(
        cls,
        model_name: str,
        direction: tuple[float, float, float],
        color: tuple[float, float, float, float],
        visual: bool,
        radius: float,
        specular: float,
        attenuation_range: float,
        attenuation_constant: float,
        attenuation_linear: float,
        attenuation_quadratic: float,
    ) -> str:
        """
        Generates the SDF string for the sun model.

        Args:
        -----
        model_name : str
            The name of the sun model.
        direction : tuple[float, float, float]
            The direction vector for the sunlight.
        color : tuple[float, float, float, float]
            The RGBA color values for the sunlight.
        visual : bool
            Whether to include a visual representation of the sun (a sphere).
        radius : float
            The radius of the visual sphere.
        specular : float
            The specular reflection intensity.
        attenuation_range : float
            The range of the light attenuation.
        attenuation_constant : float
            The constant factor for the light attenuation.
        attenuation_linear : float
            The linear factor for the light attenuation.
        attenuation_quadratic : float
            The quadratic factor for the light attenuation.

        Returns:
        --------
        str:
            The SDF string for the sun model.
        """

        return f'''<sdf version="1.9">
                <model name="{model_name}">
                    <static>true</static>
                    <link name="{model_name}_link">
                        <light type="directional" name="{model_name}_light">
                            <direction>{direction[0]} {direction[1]} {direction[2]}</direction>
                            <attenuation>
                                <range>{attenuation_range}</range>
                                <constant>{attenuation_constant}</constant>
                                <linear>{attenuation_linear}</linear>
                                <quadratic>{attenuation_quadratic}</quadratic>
                            </attenuation>
                            <diffuse>{color[0]} {color[1]} {color[2]} 1</diffuse>
                            <specular>{specular*color[0]} {specular*color[1]} {specular*color[2]} 1</specular>
                            <cast_shadows>true</cast_shadows>
                        </light>
                        {
                        f"""
                        <visual name="{model_name}_visual">
                            <geometry>
                                <sphere>
                                    <radius>{radius}</radius>
                                </sphere>
                            </geometry>
                            <material>
                                <emissive>{color[0]} {color[1]} {color[2]} 1</emissive>
                            </material>
                            <cast_shadows>false</cast_shadows>
                        </visual>
                        """ if visual else ""
                        }
                    </link>
                </model>
            </sdf>'''
