from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from rbs_assets_library import get_model_file
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo


class Model(model_wrapper.ModelWrapper):
    """
    Represents a 3D mesh model in the Gazebo simulation environment.
    This class is responsible for loading and inserting a mesh model
    into the Gazebo world with specified attributes.

    Args:
        world (scenario_gazebo.World): The Gazebo world where the mesh model will be inserted.
        name (str, optional): The name of the mesh model. Defaults to "object".
        position (tuple[float, float, float], optional): The position of the mesh in the world. Defaults to (0, 0, 0).
        orientation (tuple[float, float, float, float], optional): The orientation of the mesh in quaternion format. Defaults to (1, 0, 0, 0).
        gui_only (bool, optional): If True, the visual representation of the mesh will only appear in the GUI. Defaults to False.
        **kwargs: Additional keyword arguments.

    Raises:
        RuntimeError: If the mesh model fails to be inserted into the Gazebo world.

    Methods:
        get_sdf(model_name: str) -> str:
            Generates the SDF string used to describe the mesh model in the Gazebo simulation.
    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "object",
        position: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float, float] = (1, 0, 0, 0),
        gui_only: bool = False,
        **kwargs,
    ):
        # Get a unique model name
        model_name = get_unique_model_name(world, name)

        # Initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        # Create SDF string for the model
        sdf = self.get_sdf(
            model_name=name,
        )

        # Insert the model
        ok_model = world.to_gazebo().insert_model(sdf, initial_pose, model_name)
        if not ok_model:
            raise RuntimeError("Failed to insert " + model_name)

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        super().__init__(model=model)

    @classmethod
    def get_sdf(
        cls,
        model_name: str,
    ) -> str:
        """
        Generates the SDF string for the specified mesh model.

        Args:
            model_name (str): The name of the mesh model to generate the SDF for.

        Returns:
            str: The SDF string that defines the mesh model in Gazebo.
        """
        return get_model_file(model_name)
