import numpy as np
from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo


class Box(model_wrapper.ModelWrapper):
    """
    The Box class represents a 3D box model in the Gazebo simulation environment.
    It includes physical and visual properties such as size, mass, color, and collision properties.

    Attributes:
        world (scenario_gazebo.World): The Gazebo world where the box model will be inserted.
        name (str): The name of the box model. Default is "box".
        position (tuple[float, float, float]): The position of the box in the world. Default is (0, 0, 0).
        orientation (tuple[float, float, float, float]): The orientation of the box in quaternion format. Default is (1, 0, 0, 0).
        size (tuple[float, float, float]): The size of the box (width, length, height). Default is (0.05, 0.05, 0.05).
        mass (float): The mass of the box in kilograms. Default is 0.1.
        static (bool): If True, the box will be immovable and static. Default is False.
        collision (bool): If True, the box will have collision properties. Default is True.
        friction (float): The friction coefficient for the box’s surface. Default is 1.0.
        visual (bool): If True, the box will have a visual representation. Default is True.
        gui_only (bool): If True, the visual representation of the box will only appear in the GUI, not in the simulation physics. Default is False.
        color (tuple[float, float, float, float]): The RGBA color of the box. Default is (0.8, 0.8, 0.8, 1.0).

    Raises:
    RuntimeError:
        If the box model fails to be inserted into the Gazebo world.

    Methods:
    --------
    get_sdf() -> str:
        Generates the SDF string used to describe the box model in the Gazebo simulation.
    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "box",
        position: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float, float] = (1, 0, 0, 0),
        size: tuple[float, float, float] = (0.05, 0.05, 0.05),
        mass: float = 0.1,
        static: bool = False,
        collision: bool = True,
        friction: float = 1.0,
        visual: bool = True,
        gui_only: bool = False,
        color: tuple[float, float, float, float] | None = None,
        **kwargs,
    ):
        # Get a unique model name
        model_name = get_unique_model_name(world, name)

        # Initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        if not color:
            color = tuple(np.random.uniform(0.0, 1.0, (3,)))
            color = (color[0], color[1], color[2], 1.0)

        # Create SDF string for the model
        sdf = self.get_sdf(
            model_name=model_name,
            size=size,
            mass=mass,
            static=static,
            collision=collision,
            friction=friction,
            visual=visual,
            gui_only=gui_only,
            color=color,
        )

        # Insert the model
        ok_model = world.to_gazebo().insert_model_from_string(
            sdf, initial_pose, model_name
        )
        if not ok_model:
            raise RuntimeError(f"Failed to insert {model_name}")

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        super().__init__(model=model)

    @classmethod
    def get_sdf(
        cls,
        model_name: str,
        size: tuple[float, float, float],
        mass: float,
        static: bool,
        collision: bool,
        friction: float,
        visual: bool,
        gui_only: bool,
        color: tuple[float, float, float, float],
    ) -> str:
        """
        Generates the SDF string for the box model.

        Args:
            model_name (str): The name of the box model.
            size (tuple[float, float, float]): The dimensions of the box (width, length, height).
            mass (float): The mass of the box.
            static (bool): If True, the box will be static and immovable.
            collision (bool): If True, the box will have collision properties.
            friction (float): The friction coefficient for the box.
            visual (bool): If True, the box will have a visual representation.
            gui_only (bool): If True, the box's visual representation will only appear in the GUI and will not affect physics.
            color (tuple[float, float, float, float]): The RGBA color of the box.

        Returns:
            The SDF string that defines the box model in Gazebo.
        """
        return f'''<sdf version="1.7">
                <model name="{model_name}">
                    <static>{"true" if static else "false"}</static>
                    <link name="{model_name}_link">
                        {
                        f"""
                        <collision name="{model_name}_collision">
                            <geometry>
                                <box>
                                    <size>{size[0]} {size[1]} {size[2]}</size>
                                </box>
                            </geometry>
                            <surface>
                                <friction>
                                    <ode>
                                        <mu>{friction}</mu>
                                        <mu2>{friction}</mu2>
                                        <fdir1>0 0 0</fdir1>
                                        <slip1>0.0</slip1>
                                        <slip2>0.0</slip2>
                                    </ode>
                                </friction>
                            </surface>
                        </collision>
                        """ if collision else ""
                        }
                        {
                        f"""
                        <visual name="{model_name}_visual">
                            <geometry>
                                <box>
                                    <size>{size[0]} {size[1]} {size[2]}</size>
                                </box>
                            </geometry>
                            <material>
                                <ambient>{color[0]} {color[1]} {color[2]} {color[3]}</ambient>
                                <diffuse>{color[0]} {color[1]} {color[2]} {color[3]}</diffuse>
                                <specular>{color[0]} {color[1]} {color[2]} {color[3]}</specular>
                            </material>
                            <transparency>{1.0 - color[3]}</transparency>
                            {'<visibility_flags>1</visibility_flags> <cast_shadows>false</cast_shadows>' if gui_only else ''}
                        </visual>
                        """ if visual else ""
                        }
                        <inertial>
                            <mass>{mass}</mass>
                            <inertia>
                                <ixx>{(size[1]**2 + size[2]**2) * mass / 12}</ixx>
                                <iyy>{(size[0]**2 + size[2]**2) * mass / 12}</iyy>
                                <izz>{(size[0]**2 + size[1]**2) * mass / 12}</izz>
                                <ixy>0.0</ixy>
                                <ixz>0.0</ixz>
                                <iyz>0.0</iyz>
                            </inertia>
                        </inertial>
                    </link>
                </model>
            </sdf>'''
