import numpy as np
from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo


class Cylinder(model_wrapper.ModelWrapper):
    """
    The Cylinder class represents a cylindrical model in the Gazebo simulation environment.

    Attributes:
        world (scenario_gazebo.World): The Gazebo world where the cylinder model will be inserted.
        name (str): The name of the cylinder model. Default is "cylinder".
        position (tuple[float, float, float]): The position of the cylinder in the world. Default is (0, 0, 0).
        orientation (tuple[float, float, float, float]): The orientation of the cylinder in quaternion format. Default is (1, 0, 0, 0).
        radius (float): The radius of the cylinder. Default is 0.025.
        length (float): The length/height of the cylinder. Default is 0.1.
        mass (float): The mass of the cylinder in kilograms. Default is 0.1.
        static (bool): If True, the cylinder will be immovable. Default is False.
        collision (bool): If True, the cylinder will have collision properties. Default is True.
        friction (float): The friction coefficient for the cylinder's surface. Default is 1.0.
        visual (bool): If True, the cylinder will have a visual representation. Default is True.
        gui_only (bool): If True, the visual representation of the cylinder will only appear in the GUI, not in the simulation physics. Default is False.
        color (tuple[float, float, float, float]): The RGBA color of the cylinder. Default is (0.8, 0.8, 0.8, 1.0).

    Raises:
        RuntimeError: If the cylinder model fails to be inserted into the Gazebo world.

    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "cylinder",
        position: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float, float] = (1, 0, 0, 0),
        radius: float = 0.025,
        length: float = 0.1,
        mass: float = 0.1,
        static: bool = False,
        collision: bool = True,
        friction: float = 1.0,
        visual: bool = True,
        gui_only: bool = False,
        color: tuple[float, float, float, float] | None = None,
        **kwargs,
    ):
        model_name = get_unique_model_name(world, name)

        initial_pose = scenario_core.Pose(position, orientation)

        if not color:
            color = tuple(np.random.uniform(0.0, 1.0, (3,)))
            color = (color[0], color[1], color[2], 1.0)

        sdf = self.get_sdf(
            model_name=model_name,
            radius=radius,
            length=length,
            mass=mass,
            static=static,
            collision=collision,
            friction=friction,
            visual=visual,
            gui_only=gui_only,
            color=color,
        )

        ok_model = world.to_gazebo().insert_model_from_string(
            sdf, initial_pose, model_name
        )
        if not ok_model:
            raise RuntimeError(f"Failed to insert {model_name}")

        model = world.get_model(model_name)

        super().__init__(model=model)

    @classmethod
    def get_sdf(
        cls,
        model_name: str,
        radius: float,
        length: float,
        mass: float,
        static: bool,
        collision: bool,
        friction: float,
        visual: bool,
        gui_only: bool,
        color: tuple[float, float, float, float],
    ) -> str:
        """
        Generates the SDF string for the cylinder model.

        Args:
            model_name (str): The name of the model.
            radius (float): The radius of the cylinder.
            length (float): The length or height of the cylinder.
            mass (float): The mass of the cylinder in kilograms.
            static (bool): If True, the cylinder will remain immovable in the simulation.
            collision (bool): If True, adds collision properties to the cylinder.
            friction (float): The friction coefficient for the cylinder's surface.
            visual (bool): If True, a visual representation of the cylinder will be created.
            gui_only (bool): If True, the visual representation will only appear in the GUI, without impacting the simulation's physics.
            color (tuple[float, float, float, float]): The RGBA color of the cylinder, where each value is between 0 and 1.

        Returns:
            str: The SDF string representing the cylinder model
        """
        inertia_xx_yy = (3 * radius**2 + length**2) * mass / 12

        return f'''<sdf version="1.7">
                <model name="{model_name}">
                    <static>{"true" if static else "false"}</static>
                    <link name="{model_name}_link">
                        {
                        f"""
                        <collision name="{model_name}_collision">
                            <geometry>
                                <cylinder>
                                    <radius>{radius}</radius>
                                    <length>{length}</length>
                                </cylinder>
                            </geometry>
                            <surface>
                                <friction>
                                    <ode>
                                        <mu>{friction}</mu>
                                        <mu2>{friction}</mu2>
                                        <fdir1>0 0 0</fdir1>
                                        <slip1>0.0</slip1>
                                        <slip2>0.0</slip2>
                                    </ode>
                                </friction>
                            </surface>
                        </collision>
                        """ if collision else ""
                        }
                        {
                        f"""
                        <visual name="{model_name}_visual">
                            <geometry>
                                <cylinder>
                                    <radius>{radius}</radius>
                                    <length>{length}</length>
                                </cylinder>
                            </geometry>
                            <material>
                                <ambient>{color[0]} {color[1]} {color[2]} {color[3]}</ambient>
                                <diffuse>{color[0]} {color[1]} {color[2]} {color[3]}</diffuse>
                                <specular>{color[0]} {color[1]} {color[2]} {color[3]}</specular>
                            </material>
                            <transparency>{1.0-color[3]}</transparency>
                            {'<visibility_flags>1</visibility_flags> <cast_shadows>false</cast_shadows>' if gui_only else ''}
                        </visual>
                        """ if visual else ""
                        }
                        <inertial>
                            <mass>{mass}</mass>
                            <inertia>
                                <ixx>{inertia_xx_yy}</ixx>
                                <iyy>{inertia_xx_yy}</iyy>
                                <izz>{(mass*radius**2)/2}</izz>
                                <ixy>0.0</ixy>
                                <ixz>0.0</ixz>
                                <iyz>0.0</iyz>
                            </inertia>
                        </inertial>
                    </link>
                </model>
            </sdf>'''
