from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo


class Plane(model_wrapper.ModelWrapper):
    """
    The Plane class represents a plane model in the Gazebo simulation environment.
    It allows for defining a flat surface with collision and visual properties, as well as its orientation and friction settings.

    Attributes:
        world (scenario_gazebo.World): The Gazebo world where the plane model will be inserted.
        name (str): The name of the plane model. Default is "plane".
        position (tuple[float, float, float]): The position of the plane in the world. Default is (0, 0, 0).
        orientation (tuple[float, float, float, float]): The orientation of the plane in quaternion format. Default is (1, 0, 0, 0).
        size (tuple[float, float]): The size of the plane along the x and y axes. Default is (1.0, 1.0).
        direction (tuple[float, float, float]): The normal vector representing the plane's direction. Default is (0.0, 0.0, 1.0), representing a horizontal plane.
        collision (bool): If True, the plane will have collision properties. Default is True.
        friction (float): The friction coefficient for the plane's surface. Default is 1.0.
        visual (bool): If True, the plane will have a visual representation. Default is True.

    Raises:
        RuntimeError: If the plane model fails to be inserted into the Gazebo world.
    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "plane",
        position: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float, float] = (1, 0, 0, 0),
        size: tuple[float, float] = (1.0, 1.0),
        direction: tuple[float, float, float] = (0.0, 0.0, 1.0),
        collision: bool = True,
        friction: float = 1.0,
        visual: bool = True,
        **kwargs,
    ):
        model_name = get_unique_model_name(world, name)

        initial_pose = scenario_core.Pose(position, orientation)

        sdf = f'''<sdf version="1.7">
            <model name="{model_name}">
                <static>true</static>
                <link name="{model_name}_link">
                    {
                    f"""
                    <collision name="{model_name}_collision">
                        <geometry>
                            <plane>
                                <normal>{direction[0]} {direction[1]} {direction[2]}</normal>
                                <size>{size[0]} {size[1]}</size>
                            </plane>
                        </geometry>
                        <surface>
                            <friction>
                                <ode>
                                    <mu>{friction}</mu>
                                    <mu2>{friction}</mu2>
                                    <fdir1>0 0 0</fdir1>
                                    <slip1>0.0</slip1>
                                    <slip2>0.0</slip2>
                                </ode>
                            </friction>
                        </surface>
                    </collision>
                    """ if collision else ""
                    }
                    {
                    f"""
                    <visual name="{model_name}_visual">
                        <geometry>
                            <plane>
                                <normal>{direction[0]} {direction[1]} {direction[2]}</normal>
                                <size>{size[0]} {size[1]}</size>
                            </plane>
                        </geometry>
                        <material>
                            <ambient>0.8 0.8 0.8 1</ambient>
                            <diffuse>0.8 0.8 0.8 1</diffuse>
                            <specular>0.8 0.8 0.8 1</specular>
                        </material>
                    </visual>
                    """ if visual else ""
                    }
                </link>
            </model>
        </sdf>'''

        ok_model = world.to_gazebo().insert_model_from_string(
            sdf, initial_pose, model_name
        )
        if not ok_model:
            raise RuntimeError(f"Failed to insert {model_name}")

        model = world.get_model(model_name)

        super().__init__(model=model)
