import numpy as np
from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo


class Sphere(model_wrapper.ModelWrapper):
    """
    A Sphere model for Gazebo simulation.

    This class represents a spherical object that can be inserted into a Gazebo world.
    The sphere can be customized by specifying its physical properties, such as radius, mass,
    collision parameters, friction, and visual attributes.

    Attributes:
        world (scenario_gazebo.World): The Gazebo world where the sphere model will be inserted.
        name (str): The name of the sphere model. Defaults to "sphere".
        position (tuple[float, float, float]): The position of the sphere in the world. Defaults to (0, 0, 0).
        orientation (tuple[float, float, float, float]): The orientation of the sphere in quaternion format. Defaults to (1, 0, 0, 0).
        radius (float): The radius of the sphere. Defaults to 0.025.
        mass (float): The mass of the sphere. Defaults to 0.1.
        static (bool): If True, the sphere will be static in the simulation. Defaults to False.
        collision (bool): If True, the sphere will have collision properties. Defaults to True.
        friction (float): The friction coefficient of the sphere's surface. Defaults to 1.0.
        visual (bool): If True, the sphere will have a visual representation. Defaults to True.
        gui_only (bool): If True, the visual will only appear in the GUI. Defaults to False.
        color (tuple[float, float, float, float]): The RGBA color of the sphere. Defaults to (0.8, 0.8, 0.8, 1.0).

    Raises:
        RuntimeError: If the sphere fails to be inserted into the Gazebo world.
    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "sphere",
        position: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float, float] = (1, 0, 0, 0),
        radius: float = 0.025,
        mass: float = 0.1,
        static: bool = False,
        collision: bool = True,
        friction: float = 1.0,
        visual: bool = True,
        gui_only: bool = False,
        color: tuple[float, float, float, float] | None = None,
        **kwargs,
    ):
        # Get a unique model name
        model_name = get_unique_model_name(world, name)

        # Initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        if not color:
            color = tuple(np.random.uniform(0.0, 1.0, (3,)))
            color = (color[0], color[1], color[2], 1.0)

        # Create SDF string for the model
        sdf = self.get_sdf(
            model_name=model_name,
            radius=radius,
            mass=mass,
            static=static,
            collision=collision,
            friction=friction,
            visual=visual,
            gui_only=gui_only,
            color=color,
        )

        # Insert the model
        ok_model = world.to_gazebo().insert_model_from_string(
            sdf, initial_pose, model_name
        )
        if not ok_model:
            raise RuntimeError(f"Failed to insert {model_name}")

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        super().__init__(model=model)

    @classmethod
    def get_sdf(
        cls,
        model_name: str,
        radius: float,
        mass: float,
        static: bool,
        collision: bool,
        friction: float,
        visual: bool,
        gui_only: bool,
        color: tuple[float, float, float, float],
    ) -> str:
        """
        Generates the SDF (Simulation Description Format) string for the sphere model.

        Args:
            model_name (str): The name of the model.
            radius (float): The radius of the sphere.
            mass (float): The mass of the sphere.
            static (bool): Whether the sphere is static in the simulation.
            collision (bool): Whether the sphere should have collision properties.
            friction (float): The friction coefficient for the sphere.
            visual (bool): Whether the sphere should have a visual representation.
            gui_only (bool): Whether the visual representation is only visible in the GUI.
            color (tuple[float, float, float, float]): The RGBA color of the sphere.

        Returns:
            str: The SDF string representing the sphere.
        """
        # Inertia is identical for all axes
        inertia_xx_yy_zz = (mass * radius**2) * 2 / 5

        return f'''<sdf version="1.7">
                <model name="{model_name}">
                    <static>{"true" if static else "false"}</static>
                    <link name="{model_name}_link">
                        {
                        f"""
                        <collision name="{model_name}_collision">
                            <geometry>
                                <sphere>
                                    <radius>{radius}</radius>
                                </sphere>
                            </geometry>
                            <surface>
                                <friction>
                                    <ode>
                                        <mu>{friction}</mu>
                                        <mu2>{friction}</mu2>
                                        <fdir1>0 0 0</fdir1>
                                        <slip1>0.0</slip1>
                                        <slip2>0.0</slip2>
                                    </ode>
                                </friction>
                            </surface>
                        </collision>
                        """ if collision else ""
                        }
                        {
                        f"""
                        <visual name="{model_name}_visual">
                            <geometry>
                                <sphere>
                                    <radius>{radius}</radius>
                                </sphere>
                            </geometry>
                            <material>
                                <ambient>{color[0]} {color[1]} {color[2]} {color[3]}</ambient>
                                <diffuse>{color[0]} {color[1]} {color[2]} {color[3]}</diffuse>
                                <specular>{color[0]} {color[1]} {color[2]} {color[3]}</specular>
                            </material>
                            <transparency>{1.0 - color[3]}</transparency>
                            {'<visibility_flags>1</visibility_flags> <cast_shadows>false</cast_shadows>' if gui_only else ''}
                        </visual>
                        """ if visual else ""
                        }
                        <inertial>
                            <mass>{mass}</mass>
                            <inertia>
                                <ixx>{inertia_xx_yy_zz}</ixx>
                                <iyy>{inertia_xx_yy_zz}</iyy>
                                <izz>{inertia_xx_yy_zz}</izz>
                                <ixy>0.0</ixy>
                                <ixz>0.0</ixz>
                                <iyz>0.0</iyz>
                            </inertia>
                        </inertial>
                    </link>
                </model>
            </sdf>'''
