from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from numpy.random import RandomState
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo

from env_manager.models.utils import ModelCollectionRandomizer


class RandomObject(model_wrapper.ModelWrapper):
    """
    Represents a randomly selected 3D object model in the Gazebo simulation environment.
    This class allows for the insertion of various models based on a collection of model paths,
    utilizing a randomization strategy for the chosen model.

    Args:
        world (scenario_gazebo.World): The Gazebo world where the random object model will be inserted.
        name (str, optional): The name of the random object model. Defaults to "object".
        position (tuple[float, float, float], optional): The position of the object in the world. Defaults to (0, 0, 0).
        orientation (tuple[float, float, float, float], optional): The orientation of the object in quaternion format. Defaults to (1, 0, 0, 0).
        model_paths (str | None, optional): Paths to the model files. Must be set; raises an error if None.
        owner (str, optional): The owner of the model collection. Defaults to "GoogleResearch".
        collection (str, optional): The collection of models to choose from. Defaults to "Google Scanned Objects".
        server (str, optional): The server URL for the model collection. Defaults to "https://fuel.ignitionrobotics.org".
        server_version (str, optional): The version of the server to use. Defaults to "1.0".
        unique_cache (bool, optional): If True, enables caching of unique models. Defaults to False.
        reset_collection (bool, optional): If True, resets the model collection for new selections. Defaults to False.
        np_random (RandomState | None, optional): An instance of RandomState for random number generation. Defaults to None.
        **kwargs: Additional keyword arguments.

    Raises:
        RuntimeError: If the model path is not set or if the random object model fails to be inserted into the Gazebo world.
    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "object",
        position: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float, float] = (1, 0, 0, 0),
        model_paths: str | None = None,
        owner: str = "GoogleResearch",
        collection: str = "Google Scanned Objects",
        server: str = "https://fuel.ignitionrobotics.org",
        server_version: str = "1.0",
        unique_cache: bool = False,
        reset_collection: bool = False,
        np_random: RandomState | None = None,
        **kwargs,
    ):
        # Get a unique model name
        if model_paths is None:
            raise RuntimeError("Set model path for continue")
        model_name = get_unique_model_name(world, name)

        # Initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        model_collection_randomizer = ModelCollectionRandomizer(
            model_paths=model_paths,
            owner=owner,
            collection=collection,
            server=server,
            server_version=server_version,
            unique_cache=unique_cache,
            reset_collection=reset_collection,
            np_random=np_random,
        )

        # Note: using default arguments here
        modified_sdf_file = model_collection_randomizer.random_model()

        # Insert the model
        ok_model = world.to_gazebo().insert_model(
            modified_sdf_file, initial_pose, model_name
        )
        if not ok_model:
            raise RuntimeError("Failed to insert " + model_name)

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        model_wrapper.ModelWrapper.__init__(self, model=model)
