import numpy as np
from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from numpy.random import RandomState
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo

from . import Box, Cylinder, Sphere


class RandomPrimitive(model_wrapper.ModelWrapper):
    """
    Represents a randomly generated primitive shape (box, cylinder, or sphere) in the Gazebo simulation environment.
    This class allows for the insertion of various primitive models based on the user's specifications or randomly chosen.

    Args:
        world (scenario_gazebo.World): The Gazebo world where the primitive model will be inserted.
        name (str, optional): The name of the primitive model. Defaults to "primitive".
        use_specific_primitive ((str | None), optional): If specified, the exact type of primitive to create ('box', 'cylinder', or 'sphere'). Defaults to None, which will randomly select a primitive type.
        position (tuple[float, float, float], optional): The position of the primitive in the world. Defaults to (0, 0, 0).
        orientation (tuple[float, float, float, float], optional): The orientation of the primitive in quaternion format. Defaults to (1, 0, 0, 0).
        static (bool, optional): If True, the primitive will be static and immovable. Defaults to False.
        collision (bool, optional): If True, the primitive will have collision properties. Defaults to True.
        visual (bool, optional): If True, the primitive will have a visual representation. Defaults to True.
        gui_only (bool, optional): If True, the visual representation will only appear in the GUI and not in the simulation physics. Defaults to False.
        np_random (RandomState | None, optional): An instance of RandomState for random number generation. If None, a default random generator will be used. Defaults to None.
        **kwargs: Additional keyword arguments.

    Raises:
        RuntimeError: If the primitive model fails to be inserted into the Gazebo world.
        TypeError: If the specified primitive type is not supported.
    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "primitive",
        use_specific_primitive: (str | None) = None,
        position: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float, float] = (1, 0, 0, 0),
        static: bool = False,
        collision: bool = True,
        visual: bool = True,
        gui_only: bool = False,
        np_random: RandomState | None = None,
        **kwargs,
    ):
        if np_random is None:
            np_random = np.random.default_rng()

        # Get a unique model name
        model_name = get_unique_model_name(world, name)

        # Initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        # Create SDF string for the model
        sdf = self.get_sdf(
            model_name=model_name,
            use_specific_primitive=use_specific_primitive,
            static=static,
            collision=collision,
            visual=visual,
            gui_only=gui_only,
            np_random=np_random,
        )

        # Insert the model
        ok_model = world.to_gazebo().insert_model_from_string(
            sdf, initial_pose, model_name
        )
        if not ok_model:
            raise RuntimeError("Failed to insert " + model_name)

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        model_wrapper.ModelWrapper.__init__(self, model=model)

    @classmethod
    def get_sdf(
        cls,
        model_name: str,
        use_specific_primitive: (str | None),
        static: bool,
        collision: bool,
        visual: bool,
        gui_only: bool,
        np_random: RandomState,
    ) -> str:
        """
        Generates the SDF (Simulation Description Format) string for the specified primitive model.

        This method can create the SDF representation for a box, cylinder, or sphere based on the provided parameters.
        If a specific primitive type is not provided, one will be randomly selected.

        Args:
            model_name (str): The name of the model being generated.
            use_specific_primitive ((str | None)): The specific type of primitive to create ('box', 'cylinder', or 'sphere'). If None, a random primitive will be chosen.
            static (bool): If True, the primitive will be static and immovable.
            collision (bool): If True, the primitive will have collision properties.
            visual (bool): If True, the primitive will have a visual representation.
            gui_only (bool): If True, the visual representation will only appear in the GUI and will not affect physics.
            np_random (RandomState): An instance of RandomState for random number generation.

        Returns:
            str: The SDF string that defines the specified primitive model, including its physical and visual properties.

        Raises:
            TypeError: If the specified primitive type is not supported.

        """
        if use_specific_primitive is not None:
            primitive = use_specific_primitive
        else:
            primitive = np_random.choice(["box", "cylinder", "sphere"])

        mass = np_random.uniform(0.05, 0.25)
        friction = np_random.uniform(0.75, 1.5)
        color = tuple(np_random.uniform(0.0, 1.0, (3,)))
        color: tuple[float, float, float, float] = (color[0], color[1], color[2], 1.0)

        if "box" == primitive:
            return Box.get_sdf(
                model_name=model_name,
                size=tuple(np_random.uniform(0.04, 0.06, (3,))),
                mass=mass,
                static=static,
                collision=collision,
                friction=friction,
                visual=visual,
                gui_only=gui_only,
                color=color,
            )
        elif "cylinder" == primitive:
            return Cylinder.get_sdf(
                model_name=model_name,
                radius=np_random.uniform(0.01, 0.0375),
                length=np_random.uniform(0.025, 0.05),
                mass=mass,
                static=static,
                collision=collision,
                friction=friction,
                visual=visual,
                gui_only=gui_only,
                color=color,
            )
        elif "sphere" == primitive:
            return Sphere.get_sdf(
                model_name=model_name,
                radius=np_random.uniform(0.01, 0.0375),
                mass=mass,
                static=static,
                collision=collision,
                friction=friction,
                visual=visual,
                gui_only=gui_only,
                color=color,
            )
        else:
            raise TypeError(
                f"Type '{use_specific_primitive}' in not a supported primitive. Pleasure use 'box', 'cylinder' or 'sphere."
            )
