from gym_gz.utils.scenario import get_unique_model_name
from rclpy.node import Node
from robot_builder.parser.urdf import URDF_parser
from robot_builder.elements.robot import Robot

from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo

from .robot import RobotWrapper


class RbsArm(RobotWrapper):
    """
    A class representing a robotic arm built using the `robot_builder` library.

    This class is responsible for creating a robotic arm model within a Gazebo simulation environment.
    It allows for the manipulation of joint positions for both the arm and the gripper.

    Attributes:
        - DEFAULT_ARM_JOINT_POSITIONS (list[float]): The default joint positions for the arm.
        - OPEN_GRIPPER_JOINT_POSITIONS (list[float]): The joint positions for the gripper in the open state.
        - CLOSED_GRIPPER_JOINT_POSITIONS (list[float]): The joint positions for the gripper in the closed state.
        - DEFAULT_GRIPPER_JOINT_POSITIONS (list[float]): The default joint positions for the gripper.
    
    Args:
        - world (scenario_gazebo.World): The Gazebo world where the robot model will be inserted.
        - node (Node): The ROS2 node associated with the robotic arm.
        - urdf_string (str): The URDF string defining the robot.
        - name (str, optional): The name of the robotic arm. Defaults to "rbs_arm".
        - position (tuple[float, float, float], optional): The initial position of the robot in the world. Defaults to (0.0, 0.0, 0.0).
        - orientation (tuple[float, float, float, float], optional): The initial orientation of the robot in quaternion format. Defaults to (1.0, 0, 0, 0).
        - initial_arm_joint_positions (list[float] | float, optional): The initial joint positions for the arm. Defaults to `DEFAULT_ARM_JOINT_POSITIONS`.
        - initial_gripper_joint_positions (list[float] | float, optional): The initial joint positions for the gripper. Defaults to `DEFAULT_GRIPPER_JOINT_POSITIONS`.

    Raises:
        - RuntimeError: If the robot model fails to be inserted into the Gazebo world.
    """

    DEFAULT_ARM_JOINT_POSITIONS: list[float] = [
        0.0,
        0.5,
        3.14159,
        1.5,
        0.0,
        1.4,
        0.0,
    ]

    OPEN_GRIPPER_JOINT_POSITIONS: list[float] = [
        0.064,
    ]

    CLOSED_GRIPPER_JOINT_POSITIONS: list[float] = [
        0.0,
    ]

    DEFAULT_GRIPPER_JOINT_POSITIONS: list[float] = CLOSED_GRIPPER_JOINT_POSITIONS

    def __init__(
        self,
        world: scenario_gazebo.World,
        node: Node,
        urdf_string: str,
        name: str = "rbs_arm",
        position: tuple[float, float, float] = (0.0, 0.0, 0.0),
        orientation: tuple[float, float, float, float] = (1.0, 0, 0, 0),
        initial_arm_joint_positions: list[float] | float = DEFAULT_ARM_JOINT_POSITIONS,
        initial_gripper_joint_positions: list[float]
        | float = DEFAULT_GRIPPER_JOINT_POSITIONS,
    ):
        # Get a unique model name
        model_name = get_unique_model_name(world, name)

        # Setup initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        # Insert the model
        ok_model = world.insert_model_from_string(urdf_string, initial_pose, model_name)
        if not ok_model:
            raise RuntimeError("Failed to insert " + model_name)

        # Get the model
        model = world.get_model(model_name)

        # Parse robot to get metadata
        self._robot: Robot = URDF_parser.load_string(urdf_string)

        self.__initial_gripper_joint_positions = (
            [float(initial_gripper_joint_positions)]
            * len(self._robot.gripper_actuated_joint_names)
            if isinstance(initial_gripper_joint_positions, (int, float))
            else initial_gripper_joint_positions
        )

        self.__initial_arm_joint_positions = (
            [float(initial_arm_joint_positions)] * len(self._robot.actuated_joint_names)
            if isinstance(initial_arm_joint_positions, (int, float))
            else initial_arm_joint_positions
        )

        # Set initial joint configuration
        self.set_initial_joint_positions(model)
        super().__init__(model=model)

    @property
    def robot(self) -> Robot:
        """Returns the robot metadata parsed from the URDF string.

        Returns:
            Robot: The robot instance containing metadata.
        """
        return self._robot

    @property
    def initial_arm_joint_positions(self) -> list[float]:
        """Returns the initial joint positions for the arm.

        Returns:
            list[float]: The initial joint positions for the arm.
        """
        return self.__initial_arm_joint_positions

    @property
    def initial_gripper_joint_positions(self) -> list[float]:
        """Returns the initial joint positions for the gripper.

        Returns:
            list[float]: The initial joint positions for the gripper.
        """
        return self.__initial_gripper_joint_positions

    def set_initial_joint_positions(self, model):
        """Sets the initial positions for the robot's joints.

        This method resets the joint positions of both the arm and gripper to their specified initial values.

        Args:
            model: The model representation of the robot within the Gazebo environment.

        Raises:
            RuntimeError: If resetting the joint positions fails for any of the joints.
        """
        model = model.to_gazebo()

        joint_position_data = [
            (self.__initial_arm_joint_positions, self._robot.actuated_joint_names),
        ]
        if self._robot.gripper_actuated_joint_names:
            joint_position_data.append(
                (self.__initial_gripper_joint_positions, self._robot.gripper_actuated_joint_names)
            )

        for positions, joint_names in joint_position_data:
            print(f"Setting joint positions for {joint_names}: {positions}")
            if not model.reset_joint_positions(positions, joint_names):
                raise RuntimeError(f"Failed to set initial positions of {joint_names}'s joints")
