from gym_gz.scenario import model_wrapper
from robot_builder.elements.robot import Robot
from scenario import gazebo as scenario_gazebo
from abc import abstractmethod


class RobotWrapper(model_wrapper.ModelWrapper):
    """
    An abstract base class for robot models in a Gazebo simulation.

    This class extends the ModelWrapper from gym_gz and provides a structure for creating
    robot models with specific configurations, including joint positions for arms and grippers.

    Args:
        world (scenario_gazebo.World, optional): The Gazebo world where the robot model will be inserted.
        urdf_string (str, optional): The URDF (Unified Robot Description Format) string defining the robot.
        name (str, optional): The name of the robot model. Defaults to None.
        position (tuple[float, float, float], optional): The initial position of the robot in the world. Defaults to None.
        orientation (tuple[float, float, float, float], optional): The initial orientation of the robot in quaternion format. Defaults to None.
        initial_arm_joint_positions (list[float] | float, optional): The initial joint positions for the arm. Defaults to an empty list.
        initial_gripper_joint_positions (list[float] | float, optional): The initial joint positions for the gripper. Defaults to an empty list.
        model (model_wrapper.ModelWrapper, optional): An existing model instance to initialize the wrapper. Must be provided.
        **kwargs: Additional keyword arguments.

    Raises:
        ValueError: If the model parameter is not provided.
    """
    def __init__(
        self,
        world: scenario_gazebo.World | None = None,
        urdf_string: str | None = None,
        name: str | None = None,
        position: tuple[float, float, float] | None = None,
        orientation: tuple[float, float, float, float] | None = None,
        initial_arm_joint_positions: list[float] | float = [],
        initial_gripper_joint_positions: list[float] | float = [],
        model: model_wrapper.ModelWrapper | None = None,
        **kwargs,
    ):
        if model is not None:
            super().__init__(model=model)
        else:
            raise ValueError("Model should be defined for the parent class")

    @property
    @abstractmethod
    def robot(self) -> Robot:
        """Returns the robot instance containing metadata.

        This property must be implemented by subclasses to return the specific robot metadata.

        Returns:
            Robot: The robot instance.
        """
        pass


    @property
    @abstractmethod
    def initial_gripper_joint_positions(self) -> list[float]:
        """Returns the initial joint positions for the gripper.

        This property must be implemented by subclasses to provide the initial positions of the gripper joints.

        Returns:
            list[float]: The initial joint positions for the gripper.
        """
        pass

    @property
    @abstractmethod
    def initial_arm_joint_positions(self) -> list[float]:
        """Returns the initial joint positions for the arm.

        This property must be implemented by subclasses to provide the initial positions of the arm joints.

        Returns:
            list[float]: The initial joint positions for the arm.
        """
        pass

    @abstractmethod
    def set_initial_joint_positions(self, model):
        """Sets the initial positions for the robot's joints.

        This method must be implemented by subclasses to reset the joint positions of the robot
        to their specified initial values.

        Args:
            model: The model representation of the robot within the Gazebo environment.

        Raises:
            RuntimeError: If resetting the joint positions fails.
        """
        pass
