from gym_gz.scenario import model_wrapper
from gym_gz.utils.scenario import get_unique_model_name
from scenario import core as scenario_core
from scenario import gazebo as scenario_gazebo


class Ground(model_wrapper.ModelWrapper):
    """
    Represents a ground model in a Gazebo simulation.

    This class extends the ModelWrapper from gym_gz to define a ground model that can be
    inserted into the Gazebo world. The ground is defined by its size, position, orientation,
    and friction properties.

    Args:
        world (scenario_gazebo.World): The Gazebo world where the ground will be inserted.
        name (str, optional): The name of the ground model. Defaults to "ground".
        position (tuple[float, float, float], optional): The initial position of the ground. Defaults to (0, 0, 0).
        orientation (tuple[float, float, float, float], optional): The initial orientation of the ground in quaternion. Defaults to (1, 0, 0, 0).
        size (tuple[float, float], optional): The size of the ground plane. Defaults to (1.5, 1.5).
        collision_thickness (float, optional): The thickness of the collision surface. Defaults to 0.05.
        friction (float, optional): The friction coefficient for the ground surface. Defaults to 5.0.
        ambient (tuple[float, float, float, float], optional): The ambient color of the material. Defaults to (0.8, 0.8, 0.8, 1.0).
        specular (tuple[float, float, float, float], optional): The specular color of the material. Defaults to (0.8, 0.8, 0.8, 1.0).
        diffuse (tuple[float, float, float, float], optional): The diffuse color of the material. Defaults to (0.8, 0.8, 0.8, 1.0).
        **kwargs: Additional keyword arguments for future extensions.

    Raises:
        RuntimeError: If the ground model fails to insert into the Gazebo world.
    """

    def __init__(
        self,
        world: scenario_gazebo.World,
        name: str = "ground",
        position: tuple[float, float, float] = (0, 0, 0),
        orientation: tuple[float, float, float, float] = (1, 0, 0, 0),
        size: tuple[float, float] = (1.5, 1.5),
        collision_thickness=0.05,
        friction: float = 5.0,
        ambient: tuple[float, float, float, float] = (0.8, 0.8, 0.8, 1.0),
        specular: tuple[float, float, float, float] = (0.8, 0.8, 0.8, 1.0),
        diffuse: tuple[float, float, float, float] = (0.8, 0.8, 0.8, 1.0),
        **kwargs,
    ):
        # Get a unique model name
        model_name = get_unique_model_name(world, name)

        # Initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        # Create SDF string for the model with the provided material properties
        sdf = f"""<sdf version="1.9">
            <model name="{model_name}">
                <static>true</static>
                <link name="{model_name}_link">
                    <collision name="{model_name}_collision">
                        <geometry>
                            <plane>
                                <normal>0 0 1</normal>
                                <size>{size[0]} {size[1]}</size>
                            </plane>
                        </geometry>
                        <surface>
                            <friction>
                                <ode>
                                    <mu>{friction}</mu>
                                    <mu2>{friction}</mu2>
                                    <fdir1>0 0 0</fdir1>
                                    <slip1>0.0</slip1>
                                    <slip2>0.0</slip2>
                                </ode>
                            </friction>
                        </surface>
                    </collision>
                    <visual name="{model_name}_visual">
                        <geometry>
                            <plane>
                                <normal>0 0 1</normal>
                                <size>{size[0]} {size[1]}</size>
                            </plane>
                        </geometry>
                        <material>
                            <ambient>{' '.join(map(str, ambient))}</ambient>
                            <diffuse>{' '.join(map(str, diffuse))}</diffuse>
                            <specular>{' '.join(map(str, specular))}</specular>
                        </material>
                    </visual>
                </link>
            </model>
        </sdf>"""

        # Insert the model
        ok_model = world.to_gazebo().insert_model_from_string(
            sdf, initial_pose, model_name
        )
        if not ok_model:
            raise RuntimeError("Failed to insert " + model_name)

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        model_wrapper.ModelWrapper.__init__(self, model=model)
