import glob
import os

import numpy as np
import trimesh
from gym_gz.utils import logger
from numpy.random import RandomState
from pcg_gazebo.parsers import parse_sdf
from pcg_gazebo.parsers.sdf import create_sdf_element
from scenario import gazebo as scenario_gazebo

# Note: only models with mesh geometry are supported


class ModelCollectionRandomizer:
    """
    A class to randomize and manage paths for model collections in simulation environments.

    This class allows for the selection and management of model paths, enabling features
    like caching, blacklisting, and random sampling from a specified collection.

    Attributes:
        _class_model_paths (list[str] | None): Class-level cache for model paths.
        __sdf_base_name (str): The base name for the SDF model file.
        __configured_sdf_base_name (str): The base name for the modified SDF model file.
        __blacklisted_base_name (str): The base name used for blacklisted models.
        __collision_mesh_dir (str): Directory where collision meshes are stored.
        __collision_mesh_file_type (str): The file type of the collision meshes.
        __original_scale_base_name (str): The base name for original scale metadata files.
        _unique_cache (bool): Whether to use a unique cache for model paths.
        _enable_blacklisting (bool): Whether to enable blacklisting of unusable models.
        _model_paths (list[str] | None): Instance-specific model paths if unique cache is enabled.
        np_random (RandomState): Numpy random state for random sampling.
    """

    _class_model_paths = None
    __sdf_base_name = "model.sdf"
    __configured_sdf_base_name = "model_modified.sdf"
    __blacklisted_base_name = "BLACKLISTED"
    __collision_mesh_dir = "meshes/collision/"
    __collision_mesh_file_type = "stl"
    __original_scale_base_name = "original_scale.txt"

    def __init__(
        self,
        model_paths=None,
        owner="GoogleResearch",
        collection="Google Scanned Objects",
        server="https://fuel.ignitionrobotics.org",
        server_version="1.0",
        unique_cache=False,
        reset_collection=False,
        enable_blacklisting=True,
        np_random: RandomState | None = None,
    ):
        """
        Initializes the ModelCollectionRandomizer.

        Args:
            model_paths (list[str] | None): List of model paths to initialize with.
            owner (str): The owner of the model collection (default is "GoogleResearch").
            collection (str): The name of the model collection (default is "Google Scanned Objects").
            server (str): The server URL for the model collection (default is "https://fuel.ignitionrobotics.org").
            server_version (str): The server version to use (default is "1.0").
            unique_cache (bool): Whether to use a unique cache for the instance (default is False).
            reset_collection (bool): Whether to reset the class-level model paths (default is False).
            enable_blacklisting (bool): Whether to enable blacklisting of models (default is True).
            np_random (RandomState | None): Numpy random state for random operations (default is None).

        Raises:
            ValueError: If an invalid model path is provided.
        """

        # If enabled, the newly created objects of this class will use its own individual cache
        # for model paths and must discover/download them on its own
        self._unique_cache = unique_cache

        # Flag that determines if models that cannot be used are blacklisted
        self._enable_blacklisting = enable_blacklisting

        # If enabled, the cache of the class used to store model paths among instances will be reset
        if reset_collection and not self._unique_cache:
            self._class_model_paths = None

        # Get file path to all models from
        # a) `model_paths` arg
        # b) local cache owner (if `owner` has some models, i.e `collection` is already downloaded)
        # c) Fuel collection (if `owner` has no models in local cache)
        if model_paths is not None:
            # Use arg
            if self._unique_cache:
                self._model_paths = model_paths
            else:
                self._class_model_paths = model_paths
        else:
            # Use local cache or Fuel
            if self._unique_cache:
                self._model_paths = self.get_collection_paths(
                    owner=owner,
                    collection=collection,
                    server=server,
                    server_version=server_version,
                )
            elif self._class_model_paths is None:
                # Executed only once, unless the paths are reset with `reset_collection` arg
                self._class_model_paths = self.get_collection_paths(
                    owner=owner,
                    collection=collection,
                    server=server,
                    server_version=server_version,
                )

        # Initialise rng with (with seed is desired)
        if np_random is not None:
            self.np_random = np_random
        else:
            self.np_random = np.random.default_rng()

    @classmethod
    def get_collection_paths(
        cls,
        owner="GoogleResearch",
        collection="Google Scanned Objects",
        server="https://fuel.ignitionrobotics.org",
        server_version="1.0",
        model_name: str = "",
    ) -> list[str]:
        """
        Retrieves model paths from the local cache or downloads them from a specified server.

        The method first checks the local cache for models belonging to the specified owner.
        If no models are found, it attempts to download the models from the specified Fuel server.

        Args:
            cls: The class reference.
            owner (str): The owner of the model collection (default is "GoogleResearch").
            collection (str): The name of the model collection (default is "Google Scanned Objects").
            server (str): The server URL for the model collection (default is "https://fuel.ignitionrobotics.org").
            server_version (str): The server version to use (default is "1.0").
            model_name (str): The name of the specific model to fetch (default is an empty string).

        Returns:
            list[str]: A list of paths to the retrieved models.

        Raises:
            RuntimeError: If the download command fails or if no models are found after the download.
        """

        # First check the local cache (for performance)
        # Note: This unfortunately does not check if models belong to the specified collection
        # TODO: Make sure models belong to the collection if sampled from local cache
        model_paths = scenario_gazebo.get_local_cache_model_paths(
            owner=owner, name=model_name
        )
        if len(model_paths) > 0:
            return model_paths

        # Else download the models from Fuel and then try again
        if collection:
            download_uri = "%s/%s/%s/collections/%s" % (
                server,
                server_version,
                owner,
                collection,
            )
        elif model_name:
            download_uri = "%s/%s/%s/models/%s" % (
                server,
                server_version,
                owner,
                model_name,
            )
        download_command = 'ign fuel download -v 3 -t model -j %s -u "%s"' % (
            os.cpu_count(),
            download_uri,
        )
        os.system(download_command)

        model_paths = scenario_gazebo.get_local_cache_model_paths(
            owner=owner, name=model_name
        )
        if 0 == len(model_paths):
            logger.error(
                'URI "%s" is not valid and does not contain any models that are \
                          owned by the owner of the collection'
                % download_uri
            )
            pass

        return model_paths

    def random_model(
        self,
        min_scale=0.125,
        max_scale=0.175,
        min_mass=0.05,
        max_mass=0.25,
        min_friction=0.75,
        max_friction=1.5,
        decimation_fraction_of_visual=0.25,
        decimation_min_faces=40,
        decimation_max_faces=200,
        max_faces=40000,
        max_vertices=None,
        component_min_faces_fraction=0.1,
        component_max_volume_fraction=0.35,
        fix_mtl_texture_paths=True,
        skip_blacklisted=True,
        return_sdf_path=True,
    ) -> str:
        """
        Selects and configures a random model from the collection.

        The method attempts to find a valid model, applying randomization
        and returning the path to the configured SDF file or the model directory.

        Args:
            min_scale (float): Minimum scale for the model.
            max_scale (float): Maximum scale for the model.
            min_mass (float): Minimum mass for the model.
            max_mass (float): Maximum mass for the model.
            min_friction (float): Minimum friction coefficient.
            max_friction (float): Maximum friction coefficient.
            decimation_fraction_of_visual (float): Fraction of visual decimation.
            decimation_min_faces (int): Minimum number of faces for decimation.
            decimation_max_faces (int): Maximum number of faces for decimation.
            max_faces (int): Maximum faces for the model.
            max_vertices (int, optional): Maximum vertices for the model.
            component_min_faces_fraction (float): Minimum face fraction for components.
            component_max_volume_fraction (float): Maximum volume fraction for components.
            fix_mtl_texture_paths (bool): Whether to fix MTL texture paths.
            skip_blacklisted (bool): Whether to skip blacklisted models.
            return_sdf_path (bool): If True, return the configured SDF file path.

        Returns:
            str: Path to the configured SDF file or model directory.

        Raises:
            RuntimeError: If a valid model cannot be found after multiple attempts.
        """

        # Loop until a model is found, checked for validity, configured and returned
        # If any of these steps fail, sample another model and try again
        # Note: Due to this behaviour, the function could stall if all models are invalid
        # TODO: Add a simple timeout to random sampling of valid model (# of attempts or time-based)
        while True:
            # Get path to a random model from the collection
            model_path = self.get_random_model_path()

            # Check if the model is already blacklisted and skip if desired
            if skip_blacklisted and self.is_blacklisted(model_path):
                continue

            # Check is the model is already configured
            if self.is_configured(model_path):
                # If so, break the loop
                break

            # Process the model and break loop only if it is valid
            if self.process_model(
                model_path,
                decimation_fraction_of_visual=decimation_fraction_of_visual,
                decimation_min_faces=decimation_min_faces,
                decimation_max_faces=decimation_max_faces,
                max_faces=max_faces,
                max_vertices=max_vertices,
                component_min_faces_fraction=component_min_faces_fraction,
                component_max_volume_fraction=component_max_volume_fraction,
                fix_mtl_texture_paths=fix_mtl_texture_paths,
            ):
                break

        # Apply randomization
        self.randomize_configured_model(
            model_path,
            min_scale=min_scale,
            max_scale=max_scale,
            min_friction=min_friction,
            max_friction=max_friction,
            min_mass=min_mass,
            max_mass=max_mass,
        )

        if return_sdf_path:
            # Return path to the configured SDF file
            return self.get_configured_sdf_path(model_path)
        else:
            # Return path to the model directory
            return model_path

    def process_all_models(
        self,
        decimation_fraction_of_visual=0.025,
        decimation_min_faces=8,
        decimation_max_faces=400,
        max_faces=40000,
        max_vertices=None,
        component_min_faces_fraction=0.1,
        component_max_volume_fraction=0.35,
        fix_mtl_texture_paths=True,
    ):
        """
        Processes all models in the collection, applying configuration and decimation.

        This method iterates over each model in the collection, applying visual decimation
        and checking for validity. If a model cannot be processed successfully, it is
        blacklisted. The method tracks and prints the number of processed models and
        the count of blacklisted models at the end of execution.

        Args:
            decimation_fraction_of_visual (float): Fraction of visual decimation to apply.
                Default is 0.025.
            decimation_min_faces (int): Minimum number of faces for decimation. Default is 8.
            decimation_max_faces (int): Maximum number of faces for decimation. Default is 400.
            max_faces (int): Maximum faces allowed for the model. Default is 40000.
            max_vertices (int, optional): Maximum vertices allowed for the model. Default is None.
            component_min_faces_fraction (float): Minimum face fraction for components during
                processing. Default is 0.1.
            component_max_volume_fraction (float): Maximum volume fraction for components during
                processing. Default is 0.35.
            fix_mtl_texture_paths (bool): Whether to fix MTL texture paths for the models.
                Default is True.

        Returns:
            None: This method modifies the models in place and does not return a value.

        Prints:
            - The status of each processed model along with its index.
            - The total number of blacklisted models after processing all models.

        Raises:
            None: The method does not raise exceptions but may log issues if models cannot be
            processed.
        """
        if self._unique_cache:
            model_paths = self._model_paths
        else:
            model_paths = self._class_model_paths

        blacklist_model_counter = 0
        for i in range(len(model_paths)):
            if not self.process_model(
                model_paths[i],
                decimation_fraction_of_visual=decimation_fraction_of_visual,
                decimation_min_faces=decimation_min_faces,
                decimation_max_faces=decimation_max_faces,
                max_faces=max_faces,
                max_vertices=max_vertices,
                component_min_faces_fraction=component_min_faces_fraction,
                component_max_volume_fraction=component_max_volume_fraction,
                fix_mtl_texture_paths=fix_mtl_texture_paths,
            ):
                blacklist_model_counter += 1

            print('Processed model %i/%i "%s"' % (i, len(model_paths), model_paths[i]))

        print("Number of blacklisted models: %i" % blacklist_model_counter)

    def process_model(
        self,
        model_path,
        decimation_fraction_of_visual=0.25,
        decimation_min_faces=40,
        decimation_max_faces=200,
        max_faces=40000,
        max_vertices=None,
        component_min_faces_fraction=0.1,
        component_max_volume_fraction=0.35,
        fix_mtl_texture_paths=True,
    ) -> bool:
        """
        Processes a specified model to configure it for simulation, applying geometry
        decimation and checking for various validity criteria.

        This method parses the SDF (Simulation Description Format) of the specified
        model, processes its components, checks for geometry validity, and updates
        inertial properties as needed. The processed model is then saved back to
        an SDF file.

        Args:
            model_path (str): Path to the model to be processed.
            decimation_fraction_of_visual (float): Fraction of visual geometry to retain
                during decimation. Default is 0.25.
            decimation_min_faces (int): Minimum number of faces allowed in the decimated
                visual geometry. Default is 40.
            decimation_max_faces (int): Maximum number of faces allowed in the decimated
                visual geometry. Default is 200.
            max_faces (int): Maximum number of faces for the entire model. Default is 40000.
            max_vertices (Optional[int]): Maximum number of vertices for the model. Default is None.
            component_min_faces_fraction (float): Minimum face fraction for components to
                be considered valid during processing. Default is 0.1.
            component_max_volume_fraction (float): Maximum volume fraction for components
                to be considered valid during processing. Default is 0.35.
            fix_mtl_texture_paths (bool): Whether to fix texture paths in MTL files for
                mesh formats. Default is True.

        Returns:
            bool: Returns True if the model was processed successfully; otherwise,
            returns False if any validity checks fail or if processing issues are encountered.

        Raises:
            None: The method does not raise exceptions but may return False if processing
            fails due to model invalidity.

        Notes:
            - The model is blacklisted if it fails to meet the specified criteria for
            geometry and inertial properties.
            - The method updates the SDF of the model with processed geometries and
            inertial properties.
        """

        # Parse the SDF of the model
        sdf = parse_sdf(self.get_sdf_path(model_path))

        # Process the model(s) contained in the SDF
        for model in sdf.models:
            # Process the link(s) of each model
            for link in model.links:
                # Get rid of the existing collisions prior to simplifying it
                link.collisions.clear()

                # Values for the total inertial properties of current link
                # These values will be updated for each body that the link contains
                total_mass = 0.0
                total_inertia = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
                common_centre_of_mass = [0.0, 0.0, 0.0]

                # Go through the visuals and process them
                for visual in link.visuals:
                    # Get path to the mesh of the link's visual
                    mesh_path = self.get_mesh_path(model_path, visual)

                    # If desired, fix texture path in 'mtl' files for '.obj' mesh format
                    if fix_mtl_texture_paths:
                        self.fix_mtl_texture_paths(
                            model_path, mesh_path, model.attributes["name"]
                        )

                    # Load the mesh (without materials)
                    mesh = trimesh.load(mesh_path, force="mesh", skip_materials=True)

                    # Check if model has too much geometry (blacklist if needed)
                    if not self.check_excessive_geometry(
                        mesh, model_path, max_faces=max_faces, max_vertices=max_vertices
                    ):
                        return False

                    # Check if model has disconnected geometry/components (blacklist if needed)
                    if not self.check_disconnected_components(
                        mesh,
                        model_path,
                        component_min_faces_fraction=component_min_faces_fraction,
                        component_max_volume_fraction=component_max_volume_fraction,
                    ):
                        return False

                    # Compute inertial properties for this mesh
                    (
                        total_mass,
                        total_inertia,
                        common_centre_of_mass,
                    ) = self.sum_inertial_properties(
                        mesh, total_mass, total_inertia, common_centre_of_mass
                    )

                    # Add decimated collision geometry to the SDF
                    self.add_collision(
                        mesh,
                        link,
                        model_path,
                        fraction_of_visual=decimation_fraction_of_visual,
                        min_faces=decimation_min_faces,
                        max_faces=decimation_max_faces,
                    )

                    # Write original scale (size) into the SDF
                    # This is used for later reference during randomization (for scale limits)
                    self.write_original_scale(mesh, model_path)

                # Make sure the link has valid inertial properties (blacklist if needed)
                if not self.check_inertial_properties(
                    model_path, total_mass, total_inertia
                ):
                    return False

                # Write inertial properties to the SDF of the link
                self.write_inertial_properties(
                    link, total_mass, total_inertia, common_centre_of_mass
                )

        # Write the configured SDF into a file
        sdf.export_xml(self.get_configured_sdf_path(model_path))
        return True

    def add_collision(
        self,
        mesh,
        link,
        model_path,
        fraction_of_visual=0.05,
        min_faces=8,
        max_faces=750,
        friction=1.0,
    ):
        """
        Adds collision geometry to a specified link in a model, using a simplified
        version of the provided mesh.

        This method creates a collision mesh by decimating the visual geometry of
        the model, generating an SDF (Simulation Description Format) element for
        the collision, and adding it to the specified link.

        Args:
            mesh (trimesh.Trimesh): The original mesh geometry from which the
                collision geometry will be created.
            link (Link): The link to which the collision geometry will be added.
            model_path (str): The path to the model where the collision mesh will
                be stored.
            fraction_of_visual (float): The fraction of the visual mesh to retain
                for the collision mesh. Default is 0.05.
            min_faces (int): The minimum number of faces allowed in the collision
                mesh after decimation. Default is 8.
            max_faces (int): The maximum number of faces allowed in the collision
                mesh after decimation. Default is 750.
            friction (float): The friction coefficient to be applied to the collision
                surface. Default is 1.0.

        Returns:
            None: This method does not return a value.

        Notes:
            - The collision mesh is generated through quadratic decimation of the
            original visual mesh based on the specified parameters.
            - The created collision mesh is stored in the same directory as the
            model, and its path is relative to the model path.
            - The method updates the link's SDF with the collision geometry and
            surface properties.
        """

        # Determine name of path to the collision geometry
        collision_name = (
            link.attributes["name"] + "_collision_" + str(len(link.collisions))
        )
        collision_mesh_path = self.get_collision_mesh_path(model_path, collision_name)

        # Determine number of faces to keep after the decimation
        face_count = min(
            max(fraction_of_visual * len(mesh.faces), min_faces), max_faces
        )

        # Simplify mesh via decimation
        collision_mesh = mesh.simplify_quadratic_decimation(face_count)

        # Export the collision mesh to the appropriate location
        os.makedirs(os.path.dirname(collision_mesh_path), exist_ok=True)
        collision_mesh.export(
            collision_mesh_path, file_type=self.__collision_mesh_file_type
        )

        # Create collision SDF element
        collision = create_sdf_element("collision")

        # Add collision geometry to the SDF
        collision.geometry.mesh = create_sdf_element("mesh")
        collision.geometry.mesh.uri = os.path.relpath(
            collision_mesh_path, start=model_path
        )

        # Add surface friction to the SDF of collision (default to 1 and randomize later)
        collision.surface = create_sdf_element("surface")
        collision.surface.friction = create_sdf_element("friction", "surface")
        collision.surface.friction.ode = create_sdf_element("ode", "collision")
        collision.surface.friction.ode.mu = friction
        collision.surface.friction.ode.mu2 = friction

        # Add it to the SDF of the link
        collision_name = os.path.basename(collision_mesh_path).split(".")[0]
        link.add_collision(collision_name, collision)

    def sum_inertial_properties(
        self, mesh, total_mass, total_inertia, common_centre_of_mass, density=1.0
    ) -> tuple[float, float, float]:
        # Arbitrary density is used here
        # The mass will be randomized once it is fully computed for a link
        mesh.density = density

        # Tmp variable to store the mass of all previous geometry, used to determine centre of mass
        mass_of_others = total_mass

        # For each additional mesh, simply add the mass and inertia
        total_mass += mesh.mass
        total_inertia += mesh.moment_inertia

        # Compute a common centre of mass between all previous geometry and the new mesh
        common_centre_of_mass = [
            mass_of_others * common_centre_of_mass[0] + mesh.mass * mesh.center_mass[0],
            mass_of_others * common_centre_of_mass[1] + mesh.mass * mesh.center_mass[1],
            mass_of_others * common_centre_of_mass[2] + mesh.mass * mesh.center_mass[2],
        ] / total_mass

        return total_mass, total_inertia, common_centre_of_mass

    def randomize_configured_model(
        self,
        model_path,
        min_scale=0.05,
        max_scale=0.25,
        min_mass=0.1,
        max_mass=3.0,
        min_friction=0.75,
        max_friction=1.5,
    ):
        """
        Randomizes the scale, mass, and friction properties of a configured model
        in the specified SDF file.

        This method modifies the properties of each link in the model, applying
        random values within specified ranges for scale, mass, and friction.
        The randomized values overwrite the original settings in the configured
        SDF file.

        Args:
            model_path (str): The path to the model directory containing the
                configured SDF file.
            min_scale (float): The minimum scale factor for randomization.
                Default is 0.05.
            max_scale (float): The maximum scale factor for randomization.
                Default is 0.25.
            min_mass (float): The minimum mass for randomization. Default is 0.1.
            max_mass (float): The maximum mass for randomization. Default is 3.0.
            min_friction (float): The minimum friction coefficient for randomization.
                Default is 0.75.
            max_friction (float): The maximum friction coefficient for randomization.
                Default is 1.5.

        Returns:
            None: This method does not return a value.

        Notes:
            - The configured SDF file is updated in place, meaning the original
            properties are overwritten with the new randomized values.
            - The randomization is applied to each link in the model, ensuring
            a diverse range of physical properties.
        """

        # Get path to the configured SDF file
        configured_sdf_path = self.get_configured_sdf_path(model_path)

        # Parse the configured SDF that needs to be randomized
        sdf = parse_sdf(configured_sdf_path)

        # Process the model(s) contained in the SDF
        for model in sdf.models:
            # Process the link(s) of each model
            for link in model.links:
                # Randomize scale of the link
                self.randomize_scale(
                    model_path, link, min_scale=min_scale, max_scale=max_scale
                )

                # Randomize inertial properties of the link
                self.randomize_inertial(link, min_mass=min_mass, max_mass=max_mass)

                # Randomize friction of the link
                self.randomize_friction(
                    link, min_friction=min_friction, max_friction=max_friction
                )

        # Overwrite the configured SDF file with randomized values
        sdf.export_xml(configured_sdf_path)

    def randomize_scale(self, model_path, link, min_scale=0.05, max_scale=0.25):
        """
        Randomizes the scale of a link's geometry within a specified range.

        This method modifies the scale of the visual and collision geometry
        of a link in the model. The scale is randomized using a uniform
        distribution within the given minimum and maximum scale values.
        Additionally, the link's inertial properties (mass and inertia)
        are recalculated based on the new scale.

        Args:
            model_path (str): The path to the model directory, used to
                read the original scale of the mesh.
            link (Link): The link object containing the visual and collision
                geometries to be scaled.
            min_scale (float): The minimum scale factor for randomization.
                Default is 0.05.
            max_scale (float): The maximum scale factor for randomization.
                Default is 0.25.

        Returns:
            bool: Returns `False` if the link has more than one visual,
                indicating that scaling is not supported. Returns `None`
                if the scaling is successful.

        Notes:
            - This method currently supports only links that contain a
            single mesh geometry.
            - The mass of the link is scaled by the cube of the scale factor,
            while the inertial properties are scaled by the fifth power of
            the scale factor.
            - The scale values are applied uniformly to all dimensions (x, y, z).
        """

        # Note: This function currently supports only scaling of links with single mesh geometry
        if len(link.visuals) > 1:
            return False

        # Get a random scale for the size of mesh
        random_scale = self.np_random.uniform(min_scale, max_scale)

        # Determine a scale factor that will result in such scale for the size of mesh
        original_mesh_scale = self.read_original_scale(model_path)
        scale_factor = random_scale / original_mesh_scale

        # Determine scale factor for inertial properties based on random scale and current scale
        current_scale = link.visuals[0].geometry.mesh.scale.value[0]
        inertial_scale_factor = scale_factor / current_scale

        # Write scale factor into SDF for visual and collision geometry
        link.visuals[0].geometry.mesh.scale = [scale_factor] * 3
        link.collisions[0].geometry.mesh.scale = [scale_factor] * 3

        # Recompute inertial properties according to the scale
        link.inertial.pose.x *= inertial_scale_factor
        link.inertial.pose.y *= inertial_scale_factor
        link.inertial.pose.z *= inertial_scale_factor

        # Mass is scaled n^3
        link.mass = link.mass.value * inertial_scale_factor**3

        # Inertia is scaled n^5
        inertial_scale_factor_n5 = inertial_scale_factor**5
        link.inertia.ixx = link.inertia.ixx.value * inertial_scale_factor_n5
        link.inertia.iyy = link.inertia.iyy.value * inertial_scale_factor_n5
        link.inertia.izz = link.inertia.izz.value * inertial_scale_factor_n5
        link.inertia.ixy = link.inertia.ixy.value * inertial_scale_factor_n5
        link.inertia.ixz = link.inertia.ixz.value * inertial_scale_factor_n5
        link.inertia.iyz = link.inertia.iyz.value * inertial_scale_factor_n5

    def randomize_inertial(
        self, link, min_mass=0.1, max_mass=3.0
    ) -> tuple[float, float]:
        """
        Randomizes the mass and updates the inertial properties of a link.

        This method assigns a random mass to the link within the specified
        range and recalculates its inertial properties based on the new mass.

        Args:
            link (Link): The link object whose mass and inertial properties
                will be randomized.
            min_mass (float): The minimum mass for randomization. Default is 0.1.
            max_mass (float): The maximum mass for randomization. Default is 3.0.

        Returns:
            tuple[float, float]: A tuple containing the new mass and the
                mass scale factor used to update the inertial properties.

        Notes:
            - The method modifies the link's mass and updates its
            inertial properties (ixx, iyy, izz, ixy, ixz, iyz) based on the
            ratio of the new mass to the original mass.
        """

        random_mass = self.np_random.uniform(min_mass, max_mass)
        mass_scale_factor = random_mass / link.mass.value

        link.mass = random_mass
        link.inertia.ixx = link.inertia.ixx.value * mass_scale_factor
        link.inertia.iyy = link.inertia.iyy.value * mass_scale_factor
        link.inertia.izz = link.inertia.izz.value * mass_scale_factor
        link.inertia.ixy = link.inertia.ixy.value * mass_scale_factor
        link.inertia.ixz = link.inertia.ixz.value * mass_scale_factor
        link.inertia.iyz = link.inertia.iyz.value * mass_scale_factor

    def randomize_friction(self, link, min_friction=0.75, max_friction=1.5):
        """
        Randomizes the friction coefficients of a link's collision surfaces.

        This method assigns random friction values to each collision surface
        of the link within the specified range.

        Args:
            link (Link): The link object whose collision surfaces' friction
                coefficients will be randomized.
            min_friction (float): The minimum friction coefficient for
                randomization. Default is 0.75.
            max_friction (float): The maximum friction coefficient for
                randomization. Default is 1.5.

        Notes:
            - The friction coefficients are applied to both the
            'mu' and 'mu2' attributes of the collision surface's friction
            properties.
        """

        for collision in link.collisions:
            random_friction = self.np_random.uniform(min_friction, max_friction)

            collision.surface.friction.ode.mu = random_friction
            collision.surface.friction.ode.mu2 = random_friction

    def write_inertial_properties(self, link, mass, inertia, centre_of_mass):
        """
        Writes the specified mass and inertial properties to a link.

        This method updates the link's mass, inertia tensor, and
        centre of mass based on the provided values.

        Args:
            link (Link): The link object to be updated with new inertial properties.
            mass (float): The mass to set for the link.
            inertia (list[list[float]]): A 3x3 list representing the
                inertia tensor of the link.
            centre_of_mass (list[float]): A list containing the x, y, z
                coordinates of the link's centre of mass.

        Notes:
            - The method directly modifies the link's mass and inertia properties,
            as well as its inertial pose, which is set to the specified centre of mass.
        """

        link.mass = mass

        link.inertia.ixx = inertia[0][0]
        link.inertia.iyy = inertia[1][1]
        link.inertia.izz = inertia[2][2]
        link.inertia.ixy = inertia[0][1]
        link.inertia.ixz = inertia[0][2]
        link.inertia.iyz = inertia[1][2]

        link.inertial.pose = [
            centre_of_mass[0],
            centre_of_mass[1],
            centre_of_mass[2],
            0.0,
            0.0,
            0.0,
        ]

    def write_original_scale(self, mesh, model_path):
        """
        Writes the original scale of the mesh to a file.

        This method records the scale of the provided mesh into a text file,
        which can be used later for reference during randomization or scaling
        operations.

        Args:
            mesh (Mesh): The mesh object whose original scale is to be recorded.
            model_path (str): The path to the model directory where the
                original scale file will be saved.

        Notes:
            - The scale is written as a string representation of the mesh's
            scale property.
            - The file is created or overwritten at the location returned
            by the `get_original_scale_path` method.
        """

        file = open(self.get_original_scale_path(model_path), "w")
        file.write(str(mesh.scale))
        file.close()

    def read_original_scale(self, model_path) -> float:
        """
        Reads the original scale of a model from a file.

        This method retrieves the original scale value that was previously
        saved for a specific model. The scale is expected to be stored as
        a string in a file, and this method converts it back to a float.

        Args:
            model_path (str): The path to the model directory from which
                to read the original scale.

        Returns:
            float: The original scale of the model.

        Raises:
            ValueError: If the contents of the scale file cannot be converted to a float.
        """

        file = open(self.get_original_scale_path(model_path), "r")
        original_scale = file.read()
        file.close()

        return float(original_scale)

    def check_excessive_geometry(
        self, mesh, model_path, max_faces=40000, max_vertices=None
    ) -> bool:
        """
        Checks if the mesh exceeds the specified geometry limits.

        This method evaluates the number of faces and vertices in the
        given mesh. If the mesh exceeds the defined limits for faces
        or vertices, it blacklists the model and returns False.

        Args:
            mesh (Mesh): The mesh object to be checked for excessive geometry.
            model_path (str): The path to the model directory for logging purposes.
            max_faces (int, optional): The maximum allowed number of faces.
                Defaults to 40000.
            max_vertices (int, optional): The maximum allowed number of vertices.
                Defaults to None.

        Returns:
            bool: True if the mesh does not exceed the limits; otherwise, False.
        """

        if max_faces is not None:
            num_faces = len(mesh.faces)
            if num_faces > max_faces:
                self.blacklist_model(
                    model_path, reason="Excessive geometry (%d faces)" % num_faces
                )
                return False

        if max_vertices is not None:
            num_vertices = len(mesh.vertices)
            if num_vertices > max_vertices:
                self.blacklist_model(
                    model_path, reason="Excessive geometry (%d vertices)" % num_vertices
                )
                return False

        return True

    def check_disconnected_components(
        self,
        mesh,
        model_path,
        component_min_faces_fraction=0.05,
        component_max_volume_fraction=0.1,
    ) -> bool:
        """
        Checks for disconnected components within the mesh.

        This method identifies connected components within the mesh. If
        there are multiple components, it evaluates their relative volumes.
        If any component exceeds the specified volume fraction threshold,
        the model is blacklisted.

        Args:
            mesh (Mesh): The mesh object to be evaluated for disconnected components.
            model_path (str): The path to the model directory for logging purposes.
            component_min_faces_fraction (float, optional): The minimum fraction
                of faces required for a component to be considered. Defaults to 0.05.
            component_max_volume_fraction (float, optional): The maximum allowed
                volume fraction for a component. Defaults to 0.1.

        Returns:
            bool: True if the mesh has no significant disconnected components;
                otherwise, False.
        """

        # Get a list of all connected components inside the mesh
        # Consider components only with `component_min_faces_fraction` percent faces
        min_faces = round(component_min_faces_fraction * len(mesh.faces))
        connected_components = trimesh.graph.connected_components(
            mesh.face_adjacency, min_len=min_faces
        )

        # If more than 1 objects were detected, consider also relative volume of the meshes
        if len(connected_components) > 1:
            total_volume = mesh.volume

            large_component_counter = 0
            for component in connected_components:
                submesh = mesh.copy()
                mask = np.zeros(len(mesh.faces), dtype=np.bool)
                mask[component] = True
                submesh.update_faces(mask)

                volume_fraction = submesh.volume / total_volume
                if volume_fraction > component_max_volume_fraction:
                    large_component_counter += 1

                if large_component_counter > 1:
                    self.blacklist_model(
                        model_path,
                        reason="Disconnected components (%d instances)"
                        % len(connected_components),
                    )
                    return False

        return True

    def check_inertial_properties(self, model_path, mass, inertia) -> bool:
        """
        Checks the validity of inertial properties for a model.

        This method evaluates whether the mass and the principal moments of 
        inertia are valid. If any of these values are below a specified threshold, 
        the model is blacklisted.

        Args:
            model_path (str): The path to the model directory for logging purposes.
            mass (float): The mass of the model to be checked.
            inertia (list of list of float): A 2D list representing the inertia matrix.

        Returns:
            bool: True if the inertial properties are valid; otherwise, False.
        """
        if (
            mass < 1e-10
            or inertia[0][0] < 1e-10
            or inertia[1][1] < 1e-10
            or inertia[2][2] < 1e-10
        ):
            self.blacklist_model(model_path, reason="Invalid inertial properties")
            return False

        return True

    def get_random_model_path(self) -> str:
        """
        Retrieves a random model path from the available models.

        This method selects and returns a random model path. The selection 
        depends on whether a unique cache is being used or not.

        Returns:
            str: A randomly selected model path.
        """
        if self._unique_cache:
            return self.np_random.choice(self._model_paths)
        else:
            return self.np_random.choice(self._class_model_paths)

    def get_collision_mesh_path(self, model_path, collision_name) -> str:
        """
        Constructs the path for the collision mesh file.

        This method generates and returns the file path for a collision 
        mesh based on the model path and the name of the collision.

        Args:
            model_path (str): The path to the model directory.
            collision_name (str): The name of the collision geometry.

        Returns:
            str: The full path to the collision mesh file.
        """
        return os.path.join(
            model_path,
            self.__collision_mesh_dir,
            collision_name + "." + self.__collision_mesh_file_type,
        )

    def get_sdf_path(self, model_path) -> str:
        """
        Constructs the path for the SDF (Simulation Description Format) file.

        This method generates and returns the file path for the SDF file 
        corresponding to the given model path.

        Args:
            model_path (str): The path to the model directory.

        Returns:
            str: The full path to the SDF file.
        """
        return os.path.join(model_path, self.__sdf_base_name)

    def get_configured_sdf_path(self, model_path) -> str:
        """
        Constructs the path for the configured SDF file.

        This method generates and returns the file path for the configured 
        SDF file that includes any modifications made to the original model.

        Args:
            model_path (str): The path to the model directory.

        Returns:
            str: The full path to the configured SDF file.
        """
        return os.path.join(model_path, self.__configured_sdf_base_name)

    def get_blacklisted_path(self, model_path) -> str:
        """
        Constructs the path for the blacklisted model file.

        This method generates and returns the file path where information 
        about blacklisted models is stored.

        Args:
            model_path (str): The path to the model directory.

        Returns:
            str: The full path to the blacklisted model file.
        """
        return os.path.join(model_path, self.__blacklisted_base_name)

    def get_mesh_path(self, model_path, visual_or_collision) -> str:
        """
        Constructs the path for the mesh associated with a visual or collision.

        This method retrieves the URI from the specified visual or collision 
        object and constructs the full path to the corresponding mesh file.

        Args:
            model_path (str): The path to the model directory.
            visual_or_collision (object): An object containing the geometry 
                information for either visual or collision.

        Returns:
            str: The full path to the mesh file.
        
        Notes:
            This method may require adjustments for specific collections or models.
        """
        # TODO: This might need fixing for certain collections/models
        mesh_uri = visual_or_collision.geometry.mesh.uri.value
        return os.path.join(model_path, mesh_uri)

    def get_original_scale_path(self, model_path) -> str:
        """
        Constructs the path for the original scale file.

        This method generates and returns the file path for the file 
        that stores the original scale of the model.

        Args:
            model_path (str): The path to the model directory.

        Returns:
            str: The full path to the original scale file.
        """
        return os.path.join(model_path, self.__original_scale_base_name)

    def blacklist_model(self, model_path, reason="Unknown"):
        """
        Blacklists a model by writing its path and the reason to a file.

        If blacklisting is enabled, this method writes the specified reason 
        for blacklisting the model to a designated file. It also logs the 
        blacklisting action.

        Args:
            model_path (str): The path to the model directory.
            reason (str): The reason for blacklisting the model. Default is "Unknown".
        """
        if self._enable_blacklisting:
            bl_file = open(self.get_blacklisted_path(model_path), "w")
            bl_file.write(reason)
            bl_file.close()
        logger.warn(
            '%s model "%s". Reason: %s.'
            % (
                "Blacklisting" if self._enable_blacklisting else "Skipping",
                model_path,
                reason,
            )
        )

    def is_blacklisted(self, model_path) -> bool:
        """
        Checks if a model is blacklisted.

        This method checks if the blacklisted file for a given model exists, 
        indicating that the model has been blacklisted.

        Args:
            model_path (str): The path to the model directory.

        Returns:
            bool: True if the model is blacklisted; otherwise, False.
        """
        return os.path.isfile(self.get_blacklisted_path(model_path))

    def is_configured(self, model_path) -> bool:
        """
        Checks if a model is configured.

        This method checks if the configured SDF file for a given model 
        exists, indicating that the model has been processed and configured.

        Args:
            model_path (str): The path to the model directory.

        Returns:
            bool: True if the model is configured; otherwise, False.
        """
        return os.path.isfile(self.get_configured_sdf_path(model_path))

    def fix_mtl_texture_paths(self, model_path, mesh_path, model_name):
        """
        Fixes the texture paths in the associated MTL file of an OBJ model.

        This method modifies the texture paths in the MTL file associated 
        with a given OBJ file to ensure they are correctly linked to 
        the model's textures. It also ensures that the texture files are 
        uniquely named by prepending the model's name to avoid conflicts.

        Args:
            model_path (str): The path to the model directory where textures are located.
            mesh_path (str): The path to the OBJ file for which the MTL file needs fixing.
            model_name (str): The name of the model, used to make texture file names unique.

        Notes:
            - This method scans the model directory for texture files, identifies 
              the relevant MTL file, and updates any texture paths in that MTL file.
            - If a texture file's name does not include the model's name, the method 
              renames it to include the model's name.
            - The method only processes MTL files linked to OBJ files.
        """
        # The `.obj` files use mtl
        if mesh_path.endswith(".obj"):
            # Find all textures located in the model path, used later to relative linking
            texture_files = glob.glob(os.path.join(model_path, "**", "textures", "*.*"))

            # Find location of mtl file, if any
            mtllib_file = None
            with open(mesh_path, "r") as file:
                for line in file:
                    if "mtllib" in line:
                        mtllib_file = line.split(" ")[-1].strip()
                        break

            if mtllib_file is not None:
                mtllib_file = os.path.join(os.path.dirname(mesh_path), mtllib_file)

                fin = open(mtllib_file, "r")
                data = fin.read()
                for line in data.splitlines():
                    if "map_" in line:
                        # Find the name of the texture/map in the mtl
                        map_file = line.split(" ")[-1].strip()

                        # Find the first match of the texture/map file
                        for texture_file in texture_files:
                            if os.path.basename(
                                texture_file
                            ) == map_file or os.path.basename(
                                texture_file
                            ) == os.path.basename(map_file):
                                # Make the file unique to the model (unless it already is)
                                if model_name in texture_file:
                                    new_texture_file_name = texture_file
                                else:
                                    new_texture_file_name = texture_file.replace(
                                        map_file, model_name + "_" + map_file
                                    )

                                os.rename(texture_file, new_texture_file_name)

                                # Apply the correct relative path
                                data = data.replace(
                                    map_file,
                                    os.path.relpath(
                                        new_texture_file_name,
                                        start=os.path.dirname(mesh_path),
                                    ),
                                )
                                break
                fin.close()

                # Write in the correct data
                fout = open(mtllib_file, "w")
                fout.write(data)
                fout.close()
