from dataclasses import asdict

import numpy as np
import rclpy
from geometry_msgs.msg import Point as RosPoint
from geometry_msgs.msg import Quaternion as RosQuat
from gym_gz.scenario.model_wrapper import ModelWrapper
from rbs_assets_library import get_model_meshes_info
from rcl_interfaces.srv import GetParameters
from rclpy.node import Node, Parameter
from scenario.bindings.gazebo import GazeboSimulator, PhysicsEngine_dart, World
from scipy.spatial import distance
from scipy.spatial.transform import Rotation
from visualization_msgs.msg import Marker, MarkerArray

from env_manager.models.terrains.random_ground import RandomGround

from ..models import Box, Camera, Cylinder, Ground, Mesh, Model, Plane, Sphere, Sun
from ..models.configs import CameraData, ObjectData, SceneData
from ..models.robots import get_robot_model_class
from ..utils import Tf2Broadcaster, Tf2DynamicBroadcaster
from ..utils.conversions import quat_to_wxyz
from ..utils.gazebo import (
    transform_move_to_model_pose,
    transform_move_to_model_position,
)
from ..utils.types import Point, Pose


# TODO: Split scene randomizer and scene
class Scene:
    """
    Manages the simulation scene during runtime, including object and environment randomization.

    The Scene class initializes and manages various components in a Gazebo simulation
    environment, such as robots, cameras, lights, and terrain. It also provides methods
    for scene setup, parameter retrieval, and scene initialization.

    Attributes:
        gazebo (GazeboSimulator): An instance of the Gazebo simulator.
        robot (RobotData): The robot data configuration.
        cameras (list[CameraData]): List of camera configurations.
        objects (list[ObjectData]): List of object configurations in the scene.
        node (Node): The ROS 2 node for communication.
        light (LightData): The light configuration for the scene.
        terrain (TerrainData): The terrain configuration for the scene.
        world (World): The Gazebo world instance.
    """

    def __init__(
        self,
        node: Node,
        gazebo: GazeboSimulator,
        scene: SceneData,
        robot_urdf_string: str,
    ) -> None:
        """
        Initializes the Scene object with the necessary components and parameters.

        Args:
            - node (Node): The ROS 2 node for managing communication and parameters.
            - gazebo (GazeboSimulator): An instance of the Gazebo simulator to manage the simulation.
            - scene (SceneData): A data object containing configuration for the scene.
            - robot_urdf_string (str): The URDF string of the robot model.
        """
        self.gazebo = gazebo
        self.robot = scene.robot
        self.cameras = scene.camera
        self.objects = scene.objects
        self.node = node
        self.light = scene.light
        self.terrain = scene.terrain
        self.__scene_initialized = False
        self.__plugin_scene_broadcaster = True
        self.__plugin_user_commands = True
        self.__fts_enable = True
        self.__plugint_sensor_render_engine = "ogre2"
        self.world: World = gazebo.get_world()

        self.__objects_unique_names = {}
        self.object_models: list[ModelWrapper] = []
        self.__object_positions: dict[str, Point] = {}
        self.__objects_workspace_centre: Point = (0.0, 0.0, 0.0)

        if not robot_urdf_string:
            self.param_client = node.create_client(
                GetParameters, "/robot_state_publisher/get_parameters"
            )
            self.get_parameter_sync()
            # scene.robot.urdf_string = (
            #     node.get_parameter("robot_description")
            #     .get_parameter_value()
            #     .string_value
            # )
        else:
            scene.robot.urdf_string = robot_urdf_string

        self.tf2_broadcaster = Tf2Broadcaster(node=node)
        self.tf2_broadcaster_dyn = Tf2DynamicBroadcaster(node=node)

        self.__workspace_center = (0.0, 0.0, 0.0)
        self.__workspace_dimensions = (1.5, 1.5, 1.5)

        half_width = self.__workspace_dimensions[0] / 2
        half_height = self.__workspace_dimensions[1] / 2
        half_depth = self.__workspace_dimensions[2] / 2

        self.__workspace_min_bound = (
            self.__workspace_center[0] - half_width,
            self.__workspace_center[1] - half_height,
            self.__workspace_center[2] - half_depth,
        )
        self.__workspace_max_bound = (
            self.__workspace_center[0] + half_width,
            self.__workspace_center[1] + half_height,
            self.__workspace_center[2] + half_depth,
        )

        # self.marker_array = MarkerArray()

    def get_parameter_sync(self):
        """
        Retrieves the `robot_description` parameter synchronously, waiting for the asynchronous call to complete.
        """
        while not self.param_client.wait_for_service(timeout_sec=5.0):
            self.node.get_logger().info(
                "Service /robot_state_publisher/get_parameters is unavailable, waiting ..."
            )

        request = self.create_get_parameters_request()
        future = self.param_client.call_async(request)

        rclpy.spin_until_future_complete(self.node, future)

        if future.result() is None:
            raise RuntimeError("Failed to retrieve the robot_description parameter")

        response = future.result()
        if not response.values or not response.values[0].string_value:
            raise RuntimeError("The robot_description parameter is missing or empty")

        value = response.values[0]
        self.node.get_logger().info(
            "Succesfully got parameter response from robot_state_publisher"
        )
        param = Parameter(
            "robot_description", Parameter.Type.STRING, value.string_value
        )
        self.robot.urdf_string = value.string_value
        self.node.set_parameters([param])

    def create_get_parameters_request(self) -> GetParameters.Request:
        """
        Creates a request to get the parameters from the robot state publisher.

        Returns:
            GetParameters.Request: A request object containing the parameter name to retrieve.
        """
        return GetParameters.Request(names=["robot_description"])

    def init_scene(self):
        """
        Initializes the simulation scene by adding the terrain, lights, robots, and objects.

        This method pauses the Gazebo simulation, sets up the physics engine, and
        configures the environment and components as defined in the SceneData.

        Note:
            This method must be called after the Scene object has been initialized
            and before running the simulation.
        """

        self.gazebo_paused_run()

        self.init_world_plugins()

        self.world.set_physics_engine(PhysicsEngine_dart)

        self.add_terrain()
        self.add_light()
        self.add_robot()
        for obj in self.objects:
            self.add_object(obj)
        for camera in self.cameras:
            self.add_camera(camera)

        self.__scene_initialized = True

    def reset_scene(self):
        """
        Resets the scene to its initial state by resetting the robot joint positions
        and randomizing the positions of objects and cameras.

        This method performs the following actions:
        - Resets the joint positions of the robot.
        - Clears previously stored object positions.
        - For each object in the scene, if randomization is enabled,
          randomizes its position; otherwise, resets it to its original pose.
        - Randomizes the pose of enabled cameras if their pose has expired.

        Note:
            This method assumes that the objects and cameras have been properly initialized
            and that their properties (like randomization and pose expiration) are set correctly.
        """
        self.reset_robot_joint_positions()

        object_randomized = set()
        self.__object_positions.clear()
        for object in self.objects:
            if object.name in object_randomized:
                continue
            # TODO: Add color randomization (something that affects the runtime during the second spawn of an object)
            # if object.randomize.random_color:
            #     self.randomize_object_color(object)
            if object.randomize.random_pose:
                self.randomize_object_position(object)
            else:
                self.reset_objects_pose(object)
            object_randomized.add(object.name)

        for camera in self.cameras:
            if camera.enable and self.is_camera_pose_expired(camera):
                self.randomize_camera_pose(camera=camera)

    def init_world_plugins(self):
        """
        Initializes and inserts world plugins into the Gazebo simulator for various functionalities.

        This method configures the following plugins based on the internal settings:
        - **SceneBroadcaster**: Inserts the scene broadcaster plugin to enable GUI clients to receive
          updates about the scene. This is only done if the scene broadcaster is enabled.
        - **UserCommands**: Inserts the user commands plugin to allow user interactions with the scene.
        - **Sensors**: Inserts the sensors plugin if any cameras are enabled. The rendering engine for
          sensors can be specified.
        - **ForceTorque**: Inserts the Force Torque Sensor plugin if enabled.

        The method pauses the Gazebo simulation after inserting plugins to ensure that changes take effect.

        Note:
            If plugins are already active, this method does not reinsert them.
        """
        # SceneBroadcaster
        if self.__plugin_scene_broadcaster:
            if not self.gazebo.scene_broadcaster_active(self.world.to_gazebo().name()):
                self.node.get_logger().info(
                    "Inserting world plugins for broadcasting scene to GUI clients..."
                )
                self.world.to_gazebo().insert_world_plugin(
                    "gz-sim-scene-broadcaster-system",
                    "gz::sim::systems::SceneBroadcaster",
                )

                self.gazebo_paused_run()

        # UserCommands
        if self.__plugin_user_commands:
            self.node.get_logger().info(
                "Inserting world plugins to enable user commands..."
            )
            self.world.to_gazebo().insert_world_plugin(
                "gz-sim-user-commands-system",
                "gz::sim::systems::UserCommands",
            )

            self.gazebo_paused_run()

        # Sensors
        self._camera_enable = False
        for camera in self.cameras:
            if camera.enable:
                self._camera_enable = True

        if self._camera_enable:
            self.node.get_logger().info(
                f"Inserting world plugins for sensors with {self.__plugint_sensor_render_engine} rendering engine..."
            )
            self.world.to_gazebo().insert_world_plugin(
                "libgz-sim-sensors-system.so",
                "gz::sim::systems::Sensors",
                "<sdf version='1.9'>"
                f"<render_engine>{self.__plugint_sensor_render_engine}</render_engine>"
                "</sdf>",
            )

        if self.__fts_enable:
            self.world.to_gazebo().insert_world_plugin(
                "gz-sim-forcetorque-system",
                "gz::sim::systems::ForceTorque",
            )

            self.gazebo_paused_run()

    def add_robot(self):
        """
        Configure and insert the robot into the simulation.

        This method performs the following actions:
        - Retrieves the robot model class based on the robot's name.
        - Instantiates the robot model with specified parameters including position, orientation,
          URDF string, and initial joint positions for both the arm and gripper.
        - Ensures that the instantiated robot's name is unique and returns the actual name used.
        - Enables contact detection for all gripper links (fingers) of the robot.
        - Pauses the Gazebo simulation to apply the changes.
        - Resets the robot's joints to their default positions.

        Note:
            The method assumes that the robot name is valid and that the associated model class
            can be retrieved. If the robot's URDF string or joint positions are not correctly specified,
            this may lead to errors during robot instantiation.

        Raises:
            RuntimeError: If the robot model class cannot be retrieved for the specified robot name.
        """

        robot_model_class = get_robot_model_class(self.robot.name)

        # Instantiate robot class based on the selected model
        self.robot_wrapper = robot_model_class(
            world=self.world,
            name=self.robot.name,
            node=self.node,
            position=self.robot.spawn_position,
            orientation=quat_to_wxyz(self.robot.spawn_quat_xyzw),
            urdf_string=self.robot.urdf_string,
            initial_arm_joint_positions=self.robot.joint_positions,
            initial_gripper_joint_positions=self.robot.gripper_joint_positions,
        )

        # The desired name is passed as arg on creation, however, a different name might be selected to be unique
        # Therefore, return the name back to the task
        self.robot_name = self.robot_wrapper.name()

        # Enable contact detection for all gripper links (fingers)
        robot_gazebo = self.robot_wrapper.to_gazebo()
        robot_gazebo.enable_contacts()

        self.gazebo_paused_run()

        # Reset robot joints to the defaults
        self.reset_robot_joint_positions()

    def add_camera(self, camera: CameraData):
        """
        Configure and insert a camera into the simulation, placing it with respect to the robot.

        The method performs the following steps:
        - Determines the appropriate position and orientation of the camera based on whether it is
          relative to the world or to the robot.
        - Creates an instance of the `Camera` class with the specified parameters, including
          position, orientation, type, dimensions, image formatting, update rate, field of view,
          clipping properties, noise characteristics, and ROS 2 bridge publishing options.
        - Pauses the Gazebo simulation to apply changes.
        - Attaches the camera to the robot if the camera's reference is not the world.
        - Broadcasts the transformation (TF) of the camera relative to the robot.

        Args:
            - camera (CameraData): The camera data containing configuration options including:

        Raises:
            - Exception: If the camera cannot be attached to the robot.

        Notes:
            The method expects the camera data to be correctly configured. If the camera's
            `relative_to` field does not match any existing model in the world, it will raise an
            exception when attempting to attach the camera to the robot.
        """

        if self.world.to_gazebo().name() == camera.relative_to:
            camera_position = camera.spawn_position
            camera_quat_wxyz = quat_to_wxyz(camera.spawn_quat_xyzw)
        else:
            # Transform the pose of camera to be with respect to robot - but still represented in world reference frame for insertion into the world
            camera_position, camera_quat_wxyz = transform_move_to_model_pose(
                world=self.world,
                position=camera.spawn_position,
                quat=quat_to_wxyz(camera.spawn_quat_xyzw),
                target_model=self.robot_wrapper,
                target_link=camera.relative_to,
                xyzw=False,
            )

        # Create camera
        self.camera_wrapper = Camera(
            name=camera.name,
            world=self.world,
            position=camera_position,
            orientation=camera_quat_wxyz,
            camera_type=camera.type,
            width=camera.width,
            height=camera.height,
            image_format=camera.image_format,
            update_rate=camera.update_rate,
            horizontal_fov=camera.horizontal_fov,
            vertical_fov=camera.vertical_fov,
            clip_color=camera.clip_color,
            clip_depth=camera.clip_depth,
            noise_mean=camera.noise_mean,
            noise_stddev=camera.noise_stddev,
            ros2_bridge_color=camera.publish_color,
            ros2_bridge_depth=camera.publish_depth,
            ros2_bridge_points=camera.publish_points,
        )

        self.gazebo_paused_run()

        # Attach to robot if needed
        if self.world.to_gazebo().name() != camera.relative_to:
            if not self.robot_wrapper.to_gazebo().attach_link(
                camera.relative_to,
                self.camera_wrapper.name(),
                self.camera_wrapper.link_name,
            ):
                raise Exception("Cannot attach camera link to robot")

        self.__is_camera_attached = True

        self.gazebo_paused_run()

        # Broadcast tf
        self.tf2_broadcaster.broadcast_tf(
            parent_frame_id=camera.relative_to,
            child_frame_id=self.camera_wrapper.frame_id,
            translation=camera.spawn_position,
            rotation=camera.spawn_quat_xyzw,
            xyzw=True,
        )

    def add_object(self, obj: ObjectData):
        """
        Configure and insert an object into the simulation.

        Args:
            obj (ObjectData): The object data containing configuration options.

        Raises:
            NotImplementedError: If the specified type is not supported.
        """
        obj_dict = asdict(obj)
        obj_dict["world"] = self.world

        try:
            object_wrapper = self._create_object_wrapper(obj, obj_dict)
            self._register_object(obj, object_wrapper)
            self._enable_contact_detection(object_wrapper)

            self.node.get_logger().info(
                f"Added object: {obj.name}, relative_to: {obj.relative_to}"
            )
            self.node.get_logger().info(
                f"Position: {obj.position}, Orientation: {obj.orientation}"
            )

            self._initialize_tf_and_markers(obj, object_wrapper)

        except Exception as ex:
            self.node.get_logger().warn(f"Model could not be inserted. Reason: {ex}")

    def _create_object_wrapper(self, obj: ObjectData, obj_dict: dict) -> ModelWrapper:
        """Create the appropriate object wrapper based on the type."""
        if obj.type != "":
            return self._create_by_type(obj.type, obj_dict)
        else:
            return self._create_by_data_class(obj, obj_dict)

    def _create_by_type(self, obj_type: str, obj_dict: dict) -> ModelWrapper:
        match obj_type:
            case "box":
                return Box(**obj_dict)
            case "plane":
                return Plane(**obj_dict)
            case "sphere":
                return Sphere(**obj_dict)
            case "cylinder":
                return Cylinder(**obj_dict)
            case "model":
                return Model(**obj_dict)
            case "mesh":
                return Mesh(**obj_dict)
            case _:
                raise NotImplementedError(f"Unsupported object type: {obj_type}")

    def _create_by_data_class(self, obj: ObjectData, obj_dict: dict) -> ModelWrapper:
        from ..models.configs import (
            BoxObjectData,
            CylinderObjectData,
            MeshData,
            ModelData,
            PlaneObjectData,
            SphereObjectData,
        )

        match obj:
            case BoxObjectData():
                return Box(**obj_dict)
            case PlaneObjectData():
                return Plane(**obj_dict)
            case SphereObjectData():
                return Sphere(**obj_dict)
            case CylinderObjectData():
                return Cylinder(**obj_dict)
            case ModelData():
                return Model(**obj_dict)
            case MeshData():
                return Mesh(**obj_dict)
            case _:
                raise NotImplementedError("Unsupported object data class")

    def _register_object(self, obj: ObjectData, wrapper: ModelWrapper):
        model_name = wrapper.name()
        self.__objects_unique_names.setdefault(obj.name, []).append(model_name)
        self.__object_positions[model_name] = obj.position
        self.object_models.append(wrapper.model)

    def _enable_contact_detection(self, wrapper: ModelWrapper):
        for link_name in wrapper.link_names():
            link = wrapper.to_gazebo().get_link(link_name=link_name)
            link.enable_contact_detection(True)

    def _initialize_tf_and_markers(self, obj: ObjectData, wrapper: ModelWrapper):
        self._setup_marker_array()
        marker = self._create_marker(obj, wrapper, len(self.object_models))
        if marker:
            self.marker_array.markers.append(marker)
            # Start broadcasting once when len of markers == 1
            if len(self.marker_array.markers) == 1:
                self._setup_tf_broadcaster()

    def _setup_tf_broadcaster(self):
        if not hasattr(self, "tf_broadcaster_and_markers"):
            self.tf_broadcaster_and_markers = self.node.create_timer(
                0.1, self.publish_marker
            )

    def _setup_marker_array(self):
        if not hasattr(self, "marker_array"):
            self.marker_array = MarkerArray()
            self.marker_publisher = self.node.create_publisher(
                MarkerArray, "/markers", 10
            )

    def _create_marker(
        self, obj: ObjectData, wrapper: ModelWrapper, id: int
    ) -> Marker | None:
        """Create a visualization marker based on the object type."""
        marker = Marker()
        marker.header.frame_id = obj.relative_to
        marker.header.stamp = self.node.get_clock().now().to_msg()
        marker.ns = "assembly"
        marker.id = id

        type_map = {
            "mesh": Marker.MESH_RESOURCE,
            "model": Marker.MESH_RESOURCE,
            "sphere": Marker.SPHERE,
            "box": Marker.CUBE,
            "cylinder": Marker.CYLINDER,
        }

        if obj.type not in type_map:
            self.node.get_logger().fatal(f"Unsupported object type: {obj.type}")
            return None

        marker.type = type_map[obj.type]
        self._configure_marker_geometry_and_color(marker, obj)

        model_position, model_orientation = self.rotate_around_parent(
            wrapper.base_position(),
            wrapper.base_orientation(),
            (0.0, 0.0, 0.0),
            (1.0, 0.0, 0.0, 0.0),
        )

        position = RosPoint()
        orientation = RosQuat()
        position.x = model_position[0]
        position.y = model_position[1]
        position.z = model_position[2]

        orientation.w = model_orientation[0]
        orientation.x = model_orientation[1]
        orientation.y = model_orientation[2]
        orientation.z = model_orientation[3]

        marker.pose.position = position
        marker.pose.orientation = orientation
        marker.action = Marker.ADD

        return marker

    def _configure_marker_geometry_and_color(self, marker: Marker, obj: ObjectData):
        """Set marker geometry and color."""
        if obj.type in ["mesh", "model"]:
            marker.mesh_resource = f"file://{get_model_meshes_info(obj.name)['visual']}"
            marker.mesh_use_embedded_materials = True
            marker.scale.x = marker.scale.y = marker.scale.z = 1.0
        elif obj.type == "sphere":
            marker.scale.x = marker.scale.y = marker.scale.z = obj.radius
        elif obj.type == "box":
            marker.scale.x, marker.scale.y, marker.scale.z = obj.size
        elif obj.type == "cylinder":
            marker.scale.x = marker.scale.y = obj.radius * 2
            marker.scale.z = obj.length

        marker.color.r, marker.color.g, marker.color.b, marker.color.a = (
            0.0,
            0.5,
            1.0,
            1.0,
        )

    def publish_marker(self):
        if not self.marker_array:
            return

        for idx, model in enumerate(self.object_models):
            if self.marker_array.markers[idx]:
                model_position, model_orientation = (
                    model.base_position(),
                    model.base_orientation(),
                )

                position = RosPoint()
                orientation = RosQuat()
                position.x = model_position[0]
                position.y = model_position[1]
                position.z = model_position[2]

                orientation.w = model_orientation[0]
                orientation.x = model_orientation[1]
                orientation.y = model_orientation[2]
                orientation.z = model_orientation[3]
                self.marker_array.markers[idx].pose.position = position
                self.marker_array.markers[idx].pose.orientation = orientation

                self.tf2_broadcaster_dyn.broadcast_tf(
                    parent_frame_id="world",
                    child_frame_id=model.base_frame(),
                    translation=model_position,
                    rotation=model_orientation,
                    xyzw=False,
                )
            self.marker_publisher.publish(self.marker_array)

    def rotate_around_parent(
        self, position, orientation, parent_position, parent_orientation
    ):
        """
        Apply a 180-degree rotation around the parent's Z-axis to the object's pose.

        Args:
            position: Tuple (x, y, z) - the object's position relative to the parent.
            orientation: Tuple (w, x, y, z) - the object's orientation relative to the parent.
            parent_position: Tuple (x, y, z) - the parent's position.
            parent_orientation: Tuple (w, x, y, z) - the parent's orientation.

        Returns:
            A tuple containing the updated position and orientation:
            (new_position, new_orientation)
        """
        # Define a quaternion for 180-degree rotation around Z-axis (w, x, y, z)
        rotation_180_z = (0.0, 0.0, 0.0, 1.0)
        w1, x1, y1, z1 = orientation
        # Кватернион для 180° поворота вокруг Z
        w2, x2, y2, z2 = (0, 0, 0, 1)

        # Умножение кватернионов
        w3 = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x3 = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y3 = w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2
        z3 = w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2
        new_orientation = (w3, x3, y3, z3)

        # Compute the new orientation: parent_orientation * rotation_180_z * orientation
        # new_orientation = (
        #     parent_orientation[0] * rotation_180_z[0] - parent_orientation[1] * rotation_180_z[1] -
        #     parent_orientation[2] * rotation_180_z[2] - parent_orientation[3] * rotation_180_z[3],
        #     parent_orientation[0] * rotation_180_z[1] + parent_orientation[1] * rotation_180_z[0] +
        #     parent_orientation[2] * rotation_180_z[3] - parent_orientation[3] * rotation_180_z[2],
        #     parent_orientation[0] * rotation_180_z[2] - parent_orientation[1] * rotation_180_z[3] +
        #     parent_orientation[2] * rotation_180_z[0] + parent_orientation[3] * rotation_180_z[1],
        #     parent_orientation[0] * rotation_180_z[3] + parent_orientation[1] * rotation_180_z[2] -
        #     parent_orientation[2] * rotation_180_z[1] + parent_orientation[3] * rotation_180_z[0],
        # )

        # Apply rotation to the position
        rotated_position = (
            -position[0],  # Rotate 180 degrees around Z-axis: x -> -x
            -position[1],  # y -> -y
            position[2],  # z remains unchanged
        )

        # Add parent position
        new_position = (
            parent_position[0] + rotated_position[0],
            parent_position[1] + rotated_position[1],
            parent_position[2] + rotated_position[2],
        )

        return new_position, new_orientation

    def add_light(self):
        """
        Configure and insert a default light source into the simulation.

        This method initializes and adds a light source to the Gazebo simulation based on the
        specified light type. Currently, it supports a "sun" type light. If an unsupported
        light type is specified, it raises an error.

        Raises:
            RuntimeError: If the specified light type is unsupported, including:
                - "random_sun": Currently not implemented.
                - Any other unrecognized light type.

        Notes:
            - The method constructs the light using the properties defined in the `self.light` object,
              which should include direction, color, distance, visual properties, and radius,
              as applicable for the specified light type.
            - After adding the light to the simulation, the method pauses the Gazebo simulation
              to ensure the scene is updated.
        """

        # Create light
        match self.light.type:
            case "sun":
                self.light = Sun(
                    self.world,
                    direction=self.light.direction,
                    color=self.light.color,
                    distance=self.light.distance,
                    visual=self.light.visual,
                    radius=self.light.radius,
                )
            case "random_sun":
                raise RuntimeError("random_sun is not supported yet")
            case _:
                raise RuntimeError(
                    f"Type of light [{self.light.type}] currently not supported"
                )

        self.gazebo_paused_run()

    def add_terrain(self):
        """
        Configure and insert default terrain into the simulation.

        This method initializes and adds a terrain element to the Gazebo simulation using
        the properties specified in the `self.terrain` object. Currently, it creates a
        ground plane, but there are plans to support additional terrain types in the future.

        Raises:
            NotImplementedError: If there are attempts to add terrain types other than
            the default ground.

        Notes:
            - The method utilizes properties such as name, spawn position, spawn quaternion
              (for orientation), and size defined in `self.terrain` to create the terrain.
            - After the terrain is added, contact detection is enabled for all links associated
              with the terrain object.
            - The method pauses the Gazebo simulation after adding the terrain to ensure
              changes are reflected in the environment.
        """

        if self.terrain.type == "flat":
            self.terrain_wrapper = Ground(
                self.world,
                name=self.terrain.name,
                position=self.terrain.spawn_position,
                orientation=quat_to_wxyz(self.terrain.spawn_quat_xyzw),
                size=self.terrain.size,
                ambient=self.terrain.ambient,
                diffuse=self.terrain.diffuse,
                specular=self.terrain.specular,
            )
        elif self.terrain.type == "random_flat":
            self.terrain_wrapper = RandomGround(
                self.world,
                name=self.terrain.name,
                position=self.terrain.spawn_position,
                orientation=quat_to_wxyz(self.terrain.spawn_quat_xyzw),
                size=self.terrain.size,
            )
        else:
            raise ValueError(
                "Type of ground is not supported, supported type is: [ground, random_ground]"
            )

        # Enable contact detection
        for link_name in self.terrain_wrapper.link_names():
            link = self.terrain_wrapper.to_gazebo().get_link(link_name=link_name)
            link.enable_contact_detection(True)

        self.gazebo_paused_run()

    # def randomize_object_color(self, object: ObjectData):
    #     # debugpy.listen(5678)
    #     # debugpy.wait_for_client()
    #     task_object_names = self.__objects_unique_names[object.name]
    #     for object_name in task_object_names:
    #         if not self.world.to_gazebo().remove_model(object_name):
    #             self.node.get_logger().error(f"Failed to remove {object_name}")
    #             raise RuntimeError(f"Failed to remove {object_name}")
    #
    #     self.gazebo_paused_run()
    #     del self.__objects_unique_names[object.name]
    #     self.add_object(object)

    def randomize_object_position(self, object: ObjectData):
        """
        Randomize the position and/or orientation of a specified object in the simulation.

        This method retrieves the names of the object instances from the scene and updates
        their positions and orientations based on the randomization settings defined in
        the `object` parameter. The method applies random transformations to the object's
        pose according to the specified flags in the `ObjectData` instance.

        Args:
            object (ObjectData): An instance of `ObjectData` containing the properties
                                 of the object to be randomized.

        Raises:
            KeyError: If the object name is not found in the unique object names dictionary.

        Notes:
            - If `random_pose` is `True`, both position and orientation are randomized.
            - If `random_orientation` is `True`, only the orientation is randomized, while
              the position remains unchanged.
            - If `random_position` is `True`, only the position is randomized, while the
              orientation remains unchanged.
        """
        task_object_names = self.__objects_unique_names[object.name]
        for object_name in task_object_names:
            position, quat_random = self.get_random_object_pose(object)
            obj = self.world.to_gazebo().get_model(object_name).to_gazebo()
            if object.randomize.random_pose:
                obj.reset_base_pose(position, quat_random)
            elif object.randomize.random_orientation:
                obj.reset_base_pose(object.position, quat_random)
            elif object.randomize.random_position:
                obj.reset_base_pose(position, object.orientation)

            self.__object_positions[object_name] = position

        self.gazebo_paused_run()

    def randomize_camera_pose(self, camera: CameraData):
        """
        Randomize the pose of a specified camera in the simulation.

        This method updates the camera's position and orientation based on the selected
        randomization mode. The new pose is computed relative to either the world frame or
        a specified target model (e.g., the robot). The camera can be detached from the robot
        during the pose update and reattached afterward if needed.

        Args:
            camera (CameraData): An instance of `CameraData` containing properties of the
                                 camera to be randomized.

        Raises:
            TypeError: If the randomization mode specified in `camera.random_pose_mode` is
                       invalid.
            Exception: If the camera cannot be detached from or attached to the robot.

        Notes:
            - Supported randomization modes for camera pose:
              - `"orbit"`: The camera is positioned in an orbital path around the object.
              - `"select_random"`: A random camera pose is selected from predefined options.
              - `"select_nearest"`: The camera is positioned based on the nearest predefined
                camera pose.
        """
        # Get random camera pose, centered at object position (or center of object spawn box)
        if "orbit" == camera.random_pose_mode:
            camera_position, camera_quat_xyzw = self.get_random_camera_pose_orbit(
                camera
            )
        elif "select_random" == camera.random_pose_mode:
            (
                camera_position,
                camera_quat_xyzw,
            ) = self.get_random_camera_pose_sample_random(camera)
        elif "select_nearest" == camera.random_pose_mode:
            (
                camera_position,
                camera_quat_xyzw,
            ) = self.get_random_camera_pose_sample_nearest(camera)
        else:
            raise TypeError("Invalid mode for camera pose randomization.")

        if self.world.to_gazebo().name() == camera.relative_to:
            transformed_camera_position = camera_position
            transformed_camera_quat_wxyz = quat_to_wxyz(camera_quat_xyzw)
        else:
            # Transform the pose of camera to be with respect to robot - but represented in world reference frame for insertion into the world
            (
                transformed_camera_position,
                transformed_camera_quat_wxyz,
            ) = transform_move_to_model_pose(
                world=self.world,
                position=camera_position,
                quat=quat_to_wxyz(camera_quat_xyzw),
                target_model=self.robot_wrapper,
                target_link=camera.relative_to,
                xyzw=False,
            )

        # Detach camera if needed
        if self.__is_camera_attached:
            if not self.robot_wrapper.to_gazebo().detach_link(
                camera.relative_to,
                self.camera_wrapper.name(),
                self.camera_wrapper.link_name,
            ):
                raise Exception("Cannot detach camera link from robot")

        self.gazebo_paused_run()

        # Move pose of the camera
        camera_gazebo = self.camera_wrapper.to_gazebo()
        camera_gazebo.reset_base_pose(
            transformed_camera_position, transformed_camera_quat_wxyz
        )

        self.gazebo_paused_run()

        # Attach to robot again if needed
        if self.__is_camera_attached:
            if not self.robot_wrapper.to_gazebo().attach_link(
                camera.relative_to,
                self.camera_wrapper.name(),
                self.camera_wrapper.link_name,
            ):
                raise Exception("Cannot attach camera link to robot")

        self.gazebo_paused_run()

        # Broadcast tf
        self.tf2_broadcaster.broadcast_tf(
            parent_frame_id=camera.relative_to,
            child_frame_id=self.camera_wrapper.frame_id,
            translation=camera_position,
            rotation=camera_quat_xyzw,
            xyzw=True,
        )

    def get_random_camera_pose_orbit(self, camera: CameraData) -> Pose:
        """
        Generate a random camera pose in an orbital path around a focal point.

        This method computes a random position for the camera based on an orbital distance
        and height range. The camera's orientation is calculated to focus on a specified
        focal point. The generated position ensures that it does not fall within a specified
        arc behind the robot.

        Args:
            camera (CameraData): An instance of `CameraData` containing properties of the
                                 camera, including the orbital distance, height range,
                                 focal point offset, and arc behind the robot to ignore.

        Returns:
            Pose: A tuple containing the randomly generated position and orientation
                  (as a quaternion) of the camera.

        Notes:
            - The method utilizes a random number generator to create a uniformly
              distributed position within specified bounds.
            - The computed camera orientation is represented in the quaternion format.
        """
        rng = np.random.default_rng()

        while True:
            position = rng.uniform(
                low=(-1.0, -1.0, camera.random_pose_orbit_height_range[0]),
                high=(1.0, 1.0, camera.random_pose_orbit_height_range[1]),
            )

            norm = np.linalg.norm(position)
            position /= norm

            if (
                abs(np.arctan2(position[0], position[1]) + np.pi / 2)
                > camera.random_pose_orbit_ignore_arc_behind_robot
            ):
                break

        rpy = np.array(
            [
                0.0,
                np.arctan2(
                    position[2] - camera.random_pose_focal_point_z_offset,
                    np.linalg.norm(position[:2]),
                ),
                np.arctan2(position[1], position[0]) + np.pi,
            ]
        )
        quat_xyzw = Rotation.from_euler("xyz", rpy).as_quat()

        position *= camera.random_pose_orbit_distance
        position[:2] += self.__objects_workspace_centre[:2]

        return tuple(position), tuple(quat_xyzw)

    def get_random_camera_pose_sample_random(self, camera: CameraData) -> Pose:
        """
        Select a random position from the predefined camera position options.

        This method randomly selects one of the available camera position options
        specified in the `camera` object and processes it to generate the camera's pose.

        Args:
            camera (CameraData): An instance of `CameraData` containing the available
                                 position options for random selection.

        Returns:
            Pose: A tuple containing the selected camera position and its orientation
                  (as a quaternion).
        """
        # Select a random entry from the options
        camera.random_pose_select_position_options

        selection = camera.random_pose_select_position_options[
            np.random.randint(len(camera.random_pose_select_position_options))
        ]

        return self.get_random_camera_pose_sample_process(camera, selection)

    def get_random_camera_pose_sample_nearest(self, camera: CameraData) -> Pose:
        """
        Select the nearest entry from the predefined camera position options.

        This method calculates the squared distances of the camera position options
        from a central point and selects the nearest one to generate the camera's pose.

        Args:
            camera (CameraData): An instance of `CameraData` containing the position options
                                 to choose from.

        Returns:
            Pose: A tuple containing the nearest camera position and its orientation
                  (as a quaternion).
        """
        # Select the nearest entry
        dist_sqr = np.sum(
            (
                np.array(camera.random_pose_select_position_options)
                - np.array(camera.random_pose_select_position_options)
            )
            ** 2,
            axis=1,
        )
        nearest = camera.random_pose_select_position_options[np.argmin(dist_sqr)]

        return self.get_random_camera_pose_sample_process(camera, nearest)

    def get_random_camera_pose_sample_process(
        self, camera: CameraData, position: Point
    ) -> Pose:
        """
        Process the selected camera position to determine its orientation.

        This method calculates the camera's orientation based on the selected position,
        ensuring that the camera faces a focal point adjusted by a specified offset.

        Args:
            camera (CameraData): An instance of `CameraData` that contains properties
                                 used for orientation calculations.
            position (Point): A 3D point representing the camera's position.

        Returns:
            Pose: A tuple containing the camera's position and its orientation
                  (as a quaternion).

        Notes:
            - The orientation is computed using roll, pitch, and yaw angles based on
              the position relative to the workspace center.
        """
        # Determine orientation such that camera faces the center
        rpy = [
            0.0,
            np.arctan2(
                position[2] - camera.random_pose_focal_point_z_offset,
                np.linalg.norm(
                    (
                        position[0] - self.__objects_workspace_centre[0],
                        position[1] - self.__objects_workspace_centre[1],
                    ),
                    2,
                ),
            ),
            np.arctan2(
                position[1] - self.__objects_workspace_centre[1],
                position[0] - self.__objects_workspace_centre[0],
            )
            + np.pi,
        ]
        quat_xyzw = Rotation.from_euler("xyz", rpy).as_quat()

        return camera.spawn_position, tuple(quat_xyzw)

    def reset_robot_pose(self):
        """
        Reset the robot's pose to a randomized position or its default spawn position.

        If the robot's randomizer is enabled for pose randomization, this method will
        generate a new random position within the specified spawn volume and assign
        a random orientation. Otherwise, it will reset the robot's pose to its default
        spawn position and orientation.

        The new position is calculated by adding a random offset within the spawn
        volume to the robot's initial spawn position. The orientation is generated
        as a random rotation around the z-axis.

        Raises:
            Exception: If there are issues resetting the robot pose in the Gazebo simulation.

        Notes:
            - The position is a 3D vector represented as [x, y, z].
            - The orientation is represented as a quaternion in the form (x, y, z, w).
        """
        if self.robot.randomizer.pose:
            position = [
                self.robot.spawn_position[0]
                + np.random.RandomState().uniform(
                    -self.robot.randomizer.spawn_volume[0] / 2,
                    self.robot.randomizer.spawn_volume[0] / 2,
                ),
                self.robot.spawn_position[1]
                + np.random.RandomState().uniform(
                    -self.robot.randomizer.spawn_volume[1] / 2,
                    self.robot.randomizer.spawn_volume[1] / 2,
                ),
                self.robot.spawn_position[2]
                + np.random.RandomState().uniform(
                    -self.robot.randomizer.spawn_volume[2] / 2,
                    self.robot.randomizer.spawn_volume[2] / 2,
                ),
            ]
            quat_xyzw = Rotation.from_euler(
                "xyz", (0, 0, np.random.RandomState().uniform(-np.pi, np.pi))
            ).as_quat()
            quat_xyzw = tuple(quat_xyzw)
        else:
            position = self.robot.spawn_position
            quat_xyzw = self.robot.spawn_quat_xyzw

        gazebo_robot = self.robot_wrapper.to_gazebo()
        gazebo_robot.reset_base_pose(position, quat_to_wxyz(quat_xyzw))

        self.gazebo_paused_run()

    def reset_robot_joint_positions(self):
        """
        Reset the robot's joint positions and velocities to their initial values.

        This method retrieves the robot's initial arm joint positions and applies
        them to the robot in the Gazebo simulation. If joint position randomization
        is enabled, it adds Gaussian noise to the joint positions based on the
        specified standard deviation.

        Additionally, the method resets the velocities of both the arm and gripper
        joints to zero. After updating the joint states, it executes an unpaused step
        in Gazebo to process the modifications and update the joint states.

        Raises:
            RuntimeError: If there are issues resetting joint positions or velocities
            in the Gazebo simulation, or if the Gazebo step fails.

        Notes:
            - This method assumes the robot has a gripper only if `self.robot.with_gripper` is true.
            - Joint positions for the gripper are reset only if gripper actuated joint names are available.
        """
        gazebo_robot = self.robot_wrapper.to_gazebo()

        arm_joint_positions = self.robot_wrapper.initial_arm_joint_positions

        # Add normal noise if desired
        if self.robot.randomizer.joint_positions:
            for joint_position in arm_joint_positions:
                joint_position += np.random.RandomState().normal(
                    loc=0.0, scale=self.robot.randomizer.joint_positions_std
                )

        # Arm joints - apply joint positions zero out velocities
        if not gazebo_robot.reset_joint_positions(
            arm_joint_positions, self.robot_wrapper.robot.actuated_joint_names
        ):
            raise RuntimeError("Failed to reset robot joint positions")
        if not gazebo_robot.reset_joint_velocities(
            [0.0] * len(self.robot_wrapper.robot.actuated_joint_names),
            self.robot_wrapper.robot.actuated_joint_names,
        ):
            raise RuntimeError("Failed to reset robot joint velocities")

        # Gripper joints - apply joint positions zero out velocities
        if (
            self.robot.with_gripper
            and self.robot_wrapper.robot.gripper_actuated_joint_names
        ):
            if not gazebo_robot.reset_joint_positions(
                self.robot_wrapper.initial_gripper_joint_positions,
                self.robot_wrapper.robot.gripper_actuated_joint_names,
            ):
                raise RuntimeError("Failed to reset gripper joint positions")
            if not gazebo_robot.reset_joint_velocities(
                [0.0] * len(self.robot_wrapper.robot.gripper_actuated_joint_names),
                self.robot_wrapper.robot.gripper_actuated_joint_names,
            ):
                raise RuntimeError("Failed to reset gripper joint velocities")

        # Execute an unpaused run to process model modification and get new JointStates
        if not self.gazebo.step():
            raise RuntimeError("Failed to execute an unpaused Gazebo step")

        # if self.robot.with_gripper:
        #     if (
        #         self.robot_wrapper.CLOSED_GRIPPER_JOINT_POSITIONS
        #         == self.robot.gripper_joint_positions
        #     ):
        #         self.gripper_controller.close()
        #     else:
        #         self.gripper.open()

    def reset_objects_pose(self, object: ObjectData):
        """
        Reset the pose of specified objects in the simulation.

        This method updates the base position and orientation of all instances of
        the specified object type in the Gazebo simulation to their defined
        positions and orientations.

        Args:
            object (ObjectData): The object data containing the name and the desired
                                 position and orientation for the objects to be reset.

        Raises:
            KeyError: If the specified object's name is not found in the unique names
                       registry.

        Notes:
            - The method retrieves the list of unique names for the specified object
              type and resets the pose of each instance to the provided position and
              orientation from the `object` data.
        """
        task_object_names = self.__objects_unique_names[object.name]
        for object_name in task_object_names:
            obj = self.world.to_gazebo().get_model(object_name).to_gazebo()
            obj.reset_base_pose(object.position, object.orientation)

        self.gazebo_paused_run()

    def get_random_object_pose(
        self,
        obj: ObjectData,
        name: str = "",
        min_distance_to_other_objects: float = 0.2,
        min_distance_decay_factor: float = 0.95,
    ) -> Pose:
        """
        Generate a random pose for an object, ensuring it is sufficiently
        distanced from other objects.

        This method computes a random position within a specified spawn volume for
        the given object. It checks that the new position does not violate minimum
        distance constraints from existing objects in the simulation.

        Args:
            obj (ObjectData): The object data containing the base position,
                              randomization parameters, and relative positioning info.
            name (str, optional): The name of the object being positioned. Defaults to an empty string.
            Mic_distance_to_other_objects (float, optional): The minimum required distance
                                                             from other objects. Defaults to 0.2.
            min_distance_decay_factor (float, optional): Factor by which the minimum distance
                                                          will decay if the new position is
                                                          too close to existing objects. Defaults to 0.95.

        Returns:
            Pose: A tuple containing the randomly generated position (x, y, z) and a
                  randomly generated orientation (quaternion).

        Notes:
            - The method uses a uniform distribution to select random coordinates
              within the object's defined spawn volume.
            - The quaternion is generated to provide a random orientation for the object.
            - The method iteratively checks for proximity to other objects and adjusts the
              minimum distance if necessary.
        """
        is_too_close = True
        # The default value for randomization
        object_position = obj.position
        while is_too_close:
            object_position = (
                obj.position[0]
                + np.random.uniform(
                    -obj.randomize.random_spawn_volume[0] / 2,
                    obj.randomize.random_spawn_volume[0] / 2,
                ),
                obj.position[1]
                + np.random.uniform(
                    -obj.randomize.random_spawn_volume[1] / 2,
                    obj.randomize.random_spawn_volume[1] / 2,
                ),
                obj.position[2]
                + np.random.uniform(
                    -obj.randomize.random_spawn_volume[2] / 2,
                    obj.randomize.random_spawn_volume[2] / 2,
                ),
            )

            if self.world.to_gazebo().name() != obj.relative_to:
                # Transform the pose of camera to be with respect to robot - but represented in world reference frame for insertion into the world
                object_position = transform_move_to_model_position(
                    world=self.world,
                    position=object_position,
                    target_model=self.robot_wrapper,
                    target_link=obj.relative_to,
                )

            # Check if position is far enough from other
            is_too_close = False
            for (
                existing_object_name,
                existing_object_position,
            ) in self.__object_positions.items():
                if existing_object_name == name:
                    # Do not compare to itself
                    continue
                if (
                    distance.euclidean(object_position, existing_object_position)
                    < min_distance_to_other_objects
                ):
                    min_distance_to_other_objects *= min_distance_decay_factor
                    is_too_close = True
                    break

        quat = np.random.uniform(-1, 1, 4)
        quat /= np.linalg.norm(quat)

        return object_position, tuple(quat)

    def is_camera_pose_expired(self, camera: CameraData) -> bool:
        """
        Check if the camera pose has expired and needs to be randomized.

        This method evaluates whether the camera's current pose has reached its
        limit for randomization based on the specified rollout count. If the
        camera has exceeded the allowed number of rollouts, it will reset the
        counter and indicate that the pose should be randomized.

        Args:
            camera (CameraData): The camera data containing the current rollout
                                 counter and the maximum number of rollouts
                                 allowed before a new pose is required.

        Returns:
            bool: True if the camera pose needs to be randomized (i.e., it has
                  expired), False otherwise.
        """

        if camera.random_pose_rollouts_num == 0:
            return False

        camera.random_pose_rollout_counter += 1

        if camera.random_pose_rollout_counter >= camera.random_pose_rollouts_num:
            camera.random_pose_rollout_counter = 0
            return True

        return False

    def gazebo_paused_run(self):
        """
        Execute a run in Gazebo while paused.

        This method attempts to run the Gazebo simulation in a paused state.
        If the operation fails, it raises a RuntimeError to indicate the failure.

        Raises:
            RuntimeError: If the Gazebo run operation fails.
        """
        if not self.gazebo.run(paused=True):
            raise RuntimeError("Failed to execute a paused Gazebo run")

    def check_object_outside_workspace(
        self,
        object_position: Point,
    ) -> bool:
        """
        Check if an object is outside the defined workspace.

        This method evaluates whether the given object position lies outside the
        specified workspace boundaries, which are defined by a parallelepiped
        determined by minimum and maximum bounds.

        Args:
            object_position (Point): A tuple or list representing the 3D position
                                     of the object in the format (x, y, z).

        Returns:
            bool: True if the object is outside the workspace, False otherwise.
        """

        return (
            object_position[0] < self.__workspace_min_bound[0]
            or object_position[1] < self.__workspace_min_bound[1]
            or object_position[2] < self.__workspace_min_bound[2]
            or object_position[0] > self.__workspace_max_bound[0]
            or object_position[1] > self.__workspace_max_bound[1]
            or object_position[2] > self.__workspace_max_bound[2]
        )

    def check_object_overlapping(
        self,
        object: ObjectData,
        allowed_penetration_depth: float = 0.001,
        terrain_allowed_penetration_depth: float = 0.002,
    ) -> bool:
        """
        Check for overlapping objects in the simulation and reset their positions if necessary.

        This method iterates through all objects and ensures that none of them are overlapping.
        If an object is found to be overlapping with another object (or the robot), it resets
        its position to a random valid pose. Collisions or overlaps with terrain are ignored.

        Args:
            object (ObjectData): The object data to check for overlaps.
            allowed_penetration_depth (float, optional): The maximum allowed penetration depth
                between objects. Defaults to 0.001.
            terrain_allowed_penetration_depth (float, optional): The maximum allowed penetration
                depth when in contact with terrain. Defaults to 0.002.

        Returns:
            bool: True if all objects are in valid positions, False if any had to be reset.
        """

        # Update object positions
        for object_name in self.__objects_unique_names[object.name]:
            model = self.world.get_model(object_name).to_gazebo()
            self.__object_positions[object_name] = model.get_link(
                link_name=model.link_names()[0]
            ).position()

        for object_name in self.__objects_unique_names[object.name]:
            obj = self.world.get_model(object_name).to_gazebo()

            # Make sure the object is inside workspace
            if self.check_object_outside_workspace(
                self.__object_positions[object_name]
            ):
                position, quat_random = self.get_random_object_pose(
                    object,
                    name=object_name,
                )
                obj.reset_base_pose(position, quat_random)
                obj.reset_base_world_velocity([0.0, 0.0, 0.0], [0.0, 0.0, 0.0])
                return False

            # Make sure the object is not intersecting other objects
            try:
                for contact in obj.contacts():
                    depth = np.mean([point.depth for point in contact.points])
                    if (
                        self.terrain_wrapper.name() in contact.body_b
                        and depth < terrain_allowed_penetration_depth
                    ):
                        continue
                    if (
                        self.robot_wrapper.name() in contact.body_b
                        or depth > allowed_penetration_depth
                    ):
                        position, quat_random = self.get_random_object_pose(
                            object,
                            name=object_name,
                        )
                        obj.reset_base_pose(position, quat_random)
                        obj.reset_base_world_velocity([0.0, 0.0, 0.0], [0.0, 0.0, 0.0])
                        return False
            except Exception as e:
                self.node.get_logger().error(
                    f"Runtime error encountered while checking objects intersections: {e}"
                )

        return True
