import numpy as np
from gym_gz.scenario.model_wrapper import ModelWrapper
from rclpy.node import Node
from scenario.bindings.core import World
from scipy.spatial.transform import Rotation

from .conversions import orientation_6d_to_quat
from .gazebo import (
    get_model_orientation,
    get_model_pose,
    get_model_position,
    transform_change_reference_frame_orientation,
    transform_change_reference_frame_pose,
    transform_change_reference_frame_position,
)
from .math import quat_mul
from .tf2_listener import Tf2Listener
from .types import Point, Pose, PoseRpy, Quat, Rpy


# Helper functions #
def get_relative_ee_position(self, translation: Point) -> Point:
    # Scale relative action to metric units
    translation = self.scale_relative_translation(translation)
    # Get current position
    current_position = self.get_ee_position()
    # Compute target position
    target_position = (
        current_position[0] + translation[0],
        current_position[1] + translation[1],
        current_position[2] + translation[2],
    )

    # Restrict target position to a limited workspace, if desired
    if self._restrict_position_goal_to_workspace:
        target_position = self.restrict_position_goal_to_workspace(target_position)

    return target_position


def get_relative_ee_orientation(
    self,
    rotation: float | Quat | PoseRpy,
    representation: str = "quat",
) -> Quat:
    current_quat_xyzw = self.get_ee_orientation()

    if representation == "z":
        current_yaw = Rotation.from_quat(current_quat_xyzw).as_euler("xyz")[2]
        current_quat_xyzw = Rotation.from_euler(
            "xyz", [np.pi, 0, current_yaw]
        ).as_quat()

    if isinstance(rotation, tuple):
        if len(rotation) == 4 and representation == "quat":
            relative_quat_xyzw = rotation
        elif len(rotation) == 6 and representation == "6d":
            vectors = (rotation[0:3], rotation[3:6])
            relative_quat_xyzw = orientation_6d_to_quat(vectors[0], vectors[1])
        else:
            raise ValueError("Invalid rotation tuple length for representation.")
    elif isinstance(rotation, float) and representation == "z":
        rotation = self.scale_relative_rotation(rotation)
        relative_quat_xyzw = Rotation.from_euler("xyz", [0, 0, rotation]).as_quat()
    else:
        raise TypeError("Invalid type for rotation or representation.")

    target_quat_xyzw = quat_mul(tuple(current_quat_xyzw), tuple(relative_quat_xyzw))

    target_quat_xyzw = normalize_quaternion(tuple(relative_quat_xyzw))

    return target_quat_xyzw


def normalize_quaternion(
    target_quat_xyzw: tuple[float, float, float, float],
) -> tuple[float, float, float, float]:
    quat_array = np.array(target_quat_xyzw)
    normalized_quat = quat_array / np.linalg.norm(quat_array)
    return tuple(normalized_quat)


def scale_relative_translation(self, translation: Point) -> Point:
    return (
        self.__scaling_factor_translation * translation[0],
        self.__scaling_factor_translation * translation[1],
        self.__scaling_factor_translation * translation[2],
    )


def scale_relative_rotation(
    self, rotation: float | Rpy | np.floating | np.ndarray
) -> float | Rpy:
    scaling_factor = self.__scaling_factor_rotation

    if isinstance(rotation, (int, float, np.floating)):
        return scaling_factor * rotation

    return tuple(scaling_factor * r for r in rotation)


def restrict_position_goal_to_workspace(self, position: Point) -> Point:
    return (
        min(
            self.workspace_max_bound[0],
            max(
                self.workspace_min_bound[0],
                position[0],
            ),
        ),
        min(
            self.workspace_max_bound[1],
            max(
                self.workspace_min_bound[1],
                position[1],
            ),
        ),
        min(
            self.workspace_max_bound[2],
            max(
                self.workspace_min_bound[2],
                position[2],
            ),
        ),
    )


# def restrict_servo_translation_to_workspace(
#     self, translation: tuple[float, float, float]
# ) -> tuple[float, float, float]:
#     current_ee_position = self.get_ee_position()
#
#     translation = tuple(
#         0.0
#         if (
#             current_ee_position[i] > self.workspace_max_bound[i]
#             and translation[i] > 0.0
#         )
#         or (
#             current_ee_position[i] < self.workspace_min_bound[i]
#             and translation[i] < 0.0
#         )
#         else translation[i]
#         for i in range(3)
#     )
#
#     return translation


def get_ee_pose(
    node: Node,
    world: World,
    robot_ee_link_name: str,
    robot_name: str,
    robot_arm_base_link_name: str,
    tf2_listener: Tf2Listener,
) -> Pose | None:
    """
    Return the current pose of the end effector with respect to arm base link.
    """

    try:
        robot_model = world.to_gazebo().get_model(robot_name).to_gazebo()
        ee_position, ee_quat_xyzw = get_model_pose(
            world=world,
            model=robot_model,
            link=robot_ee_link_name,
            xyzw=True,
        )
        return transform_change_reference_frame_pose(
            world=world,
            position=ee_position,
            quat=ee_quat_xyzw,
            target_model=robot_model,
            target_link=robot_arm_base_link_name,
            xyzw=True,
        )
    except Exception as e:
        node.get_logger().warn(
            f"Cannot get end effector pose from Gazebo ({e}), using tf2..."
        )
        transform = tf2_listener.lookup_transform_sync(
            source_frame=robot_ee_link_name,
            target_frame=robot_arm_base_link_name,
            retry=False,
        )
        if transform is not None:
            return (
                (
                    transform.translation.x,
                    transform.translation.y,
                    transform.translation.z,
                ),
                (
                    transform.rotation.x,
                    transform.rotation.y,
                    transform.rotation.z,
                    transform.rotation.w,
                ),
            )
        else:
            node.get_logger().error(
                "Cannot get pose of the end effector (default values are returned)"
            )
            return (
                (0.0, 0.0, 0.0),
                (0.0, 0.0, 0.0, 1.0),
            )


def get_ee_position(
    world: World,
    robot_name: str,
    robot_ee_link_name: str,
    robot_arm_base_link_name: str,
    node: Node,
    tf2_listener: Tf2Listener,
) -> Point:
    """
    Return the current position of the end effector with respect to arm base link.
    """

    try:
        robot_model = world.to_gazebo().get_model(robot_name).to_gazebo()
        ee_position = get_model_position(
            world=world,
            model=robot_model,
            link=robot_ee_link_name,
        )
        return transform_change_reference_frame_position(
            world=world,
            position=ee_position,
            target_model=robot_model,
            target_link=robot_arm_base_link_name,
        )
    except Exception as e:
        node.get_logger().debug(
            f"Cannot get end effector position from Gazebo ({e}), using tf2..."
        )
        transform = tf2_listener.lookup_transform_sync(
            source_frame=robot_ee_link_name,
            target_frame=robot_arm_base_link_name,
            retry=False,
        )
        if transform is not None:
            return (
                transform.translation.x,
                transform.translation.y,
                transform.translation.z,
            )
        else:
            node.get_logger().error(
                "Cannot get position of the end effector (default values are returned)"
            )
            return (0.0, 0.0, 0.0)


def get_ee_orientation(
    world: World,
    robot_name: str,
    robot_ee_link_name: str,
    robot_arm_base_link_name: str,
    node: Node,
    tf2_listener: Tf2Listener,
) -> Quat:
    """
    Return the current xyzw quaternion of the end effector with respect to arm base link.
    """

    try:
        robot_model = world.to_gazebo().get_model(robot_name).to_gazebo()
        ee_quat_xyzw = get_model_orientation(
            world=world,
            model=robot_model,
            link=robot_ee_link_name,
            xyzw=True,
        )
        return transform_change_reference_frame_orientation(
            world=world,
            quat=ee_quat_xyzw,
            target_model=robot_model,
            target_link=robot_arm_base_link_name,
            xyzw=True,
        )
    except Exception as e:
        node.get_logger().warn(
            f"Cannot get end effector orientation from Gazebo ({e}), using tf2..."
        )
        transform = tf2_listener.lookup_transform_sync(
            source_frame=robot_ee_link_name,
            target_frame=robot_arm_base_link_name,
            retry=False,
        )
        if transform is not None:
            return (
                transform.rotation.x,
                transform.rotation.y,
                transform.rotation.z,
                transform.rotation.w,
            )
        else:
            node.get_logger().error(
                "Cannot get orientation of the end effector (default values are returned)"
            )
            return (0.0, 0.0, 0.0, 1.0)


def get_object_position(
    object_model: ModelWrapper | str,
    node: Node,
    world: World,
    robot_name: str,
    robot_arm_base_link_name: str,
) -> Point:
    """
    Return the current position of an object with respect to arm base link.
    Note: Only simulated objects are currently supported.
    """

    try:
        object_position = get_model_position(
            world=world,
            model=object_model,
        )
        return transform_change_reference_frame_position(
            world=world,
            position=object_position,
            target_model=robot_name,
            target_link=robot_arm_base_link_name,
        )
    except Exception as e:
        node.get_logger().error(
            f"Cannot get position of {object_model} object (default values are returned): {e}"
        )
        return (0.0, 0.0, 0.0)


def get_object_positions(
    node: Node,
    world: World,
    object_names: list[str],
    robot_name: str,
    robot_arm_base_link_name: str,
) -> dict[str, tuple[float, float, float]]:
    """
    Return the current position of all objects with respect to arm base link.
    Note: Only simulated objects are currently supported.
    """

    object_positions = {}

    try:
        robot_model = world.to_gazebo().get_model(robot_name).to_gazebo()
        robot_arm_base_link = robot_model.get_link(link_name=robot_arm_base_link_name)
        for object_name in object_names:
            object_position = get_model_position(
                world=world,
                model=object_name,
            )
            object_positions[object_name] = transform_change_reference_frame_position(
                world=world,
                position=object_position,
                target_model=robot_model,
                target_link=robot_arm_base_link,
            )
    except Exception as e:
        node.get_logger().error(
            f"Cannot get positions of all objects (empty Dict is returned): {e}"
        )

    return object_positions


# def substitute_special_frame(self, frame_id: str) -> str:
#     if "arm_base_link" == frame_id:
#         return self.robot_arm_base_link_name
#     elif "base_link" == frame_id:
#         return self.robot_base_link_name
#     elif "end_effector" == frame_id:
#         return self.robot_ee_link_name
#     elif "world" == frame_id:
#         try:
#             # In Gazebo, where multiple worlds are allowed
#             return self.world.to_gazebo().name()
#         except Exception as e:
#             self.get_logger().warn(f"")
#             # Otherwise (e.g. real world)
#             return "rbs_gym_world"
#     else:
#         return frame_id


def move_to_initial_joint_configuration(self):
    pass

    # self.moveit2.move_to_configuration(self.initial_arm_joint_positions)

    # if (
    #     self.robot_model_class.CLOSED_GRIPPER_JOINT_POSITIONS
    #     == self.initial_gripper_joint_positions
    # ):
    #     self.gripper.reset_close()
    # else:
    #     self.gripper.reset_open()


def check_terrain_collision(
    world: World,
    robot_name: str,
    terrain_name: str,
    robot_base_link_name: str,
    robot_arm_link_names: list[str],
    robot_gripper_link_names: list[str],
) -> bool:
    """
    Returns true if robot links are in collision with the ground.
    """
    robot_name_len = len(robot_name)
    terrain_model = world.get_model(terrain_name)

    for contact in terrain_model.contacts():
        body_b = contact.body_b

        if body_b.startswith(robot_name) and len(body_b) > robot_name_len:
            link = body_b[robot_name_len + 2 :]

            if link != robot_base_link_name and (
                link in robot_arm_link_names or link in robot_gripper_link_names
            ):
                return True

    return False


def check_all_objects_outside_workspace(
    workspace: Pose,
    object_positions: dict[str, Point],
) -> bool:
    """
    Returns True if all objects are outside the workspace.
    """

    return all(
        [
            check_object_outside_workspace(workspace, object_position)
            for object_position in object_positions.values()
        ]
    )


def check_object_outside_workspace(workspace: Pose, object_position: Point) -> bool:
    """
    Returns True if the object is outside the workspace.
    """
    workspace_min_bound, workspace_max_bound = workspace
    return any(
        coord < min_bound or coord > max_bound
        for coord, min_bound, max_bound in zip(
            object_position, workspace_min_bound, workspace_max_bound
        )
    )


# def add_parameter_overrides(self, parameter_overrides: Dict[str, any]):
#     self.add_task_parameter_overrides(parameter_overrides)
#     self.add_randomizer_parameter_overrides(parameter_overrides)
#
#
# def add_task_parameter_overrides(self, parameter_overrides: Dict[str, any]):
#     self.__task_parameter_overrides.update(parameter_overrides)
#
#
# def add_randomizer_parameter_overrides(self, parameter_overrides: Dict[str, any]):
#     self._randomizer_parameter_overrides.update(parameter_overrides)
#
#
# def __consume_parameter_overrides(self):
#     for key, value in self.__task_parameter_overrides.items():
#         if hasattr(self, key):
#             setattr(self, key, value)
#         elif hasattr(self, f"_{key}"):
#             setattr(self, f"_{key}", value)
#         elif hasattr(self, f"__{key}"):
#             setattr(self, f"__{key}", value)
#         else:
#             self.get_logger().error(f"Override '{key}' is not supperted by the task.")
#
#     self.__task_parameter_overrides.clear()
