# Note: The `open3d` and `stable_baselines3` modules must be imported prior to `gym_gz`
from env_manager.models.configs import (
    RobotData,
    SceneData,
    SphereObjectData,
    TerrainData,
)
from env_manager.models.configs.objects import ObjectRandomizerData
from env_manager.models.configs.robot import RobotRandomizerData
from env_manager.models.configs.scene import PluginsData
# import open3d  # isort:skip
import stable_baselines3  # isort:skip

# Note: If installed, `tensorflow` module must be imported before `gym_gz`/`scenario`
# Otherwise, protobuf version incompatibility will cause an error
try:
    from importlib.util import find_spec

    if find_spec("tensorflow") is not None:
        import tensorflow
except:
    pass

from os import environ, path
from typing import Any, Dict, Tuple

import numpy as np
from ament_index_python.packages import get_package_share_directory
from gymnasium.envs.registration import register
from rbs_assets_library import get_world_file

from rbs_gym.utils.utils import str2bool

from . import tasks

######################
# Runtime Entrypoint #
######################

RBS_ENVS_TASK_ENTRYPOINT: str = "gym_gz.runtimes.gazebo_runtime:GazeboRuntime"


###################
# Robot Specifics #
###################
# Default robot model to use in the tasks where robot can be static
RBS_ENVS_ROBOT_MODEL: str = "rbs_arm"

######################
# Datasets and paths #
######################
# Path to directory containing base SDF worlds
RBS_ENVS_WORLDS_DIR: str = path.join(get_package_share_directory("rbs_gym"), "worlds")


###########
# Presets #
###########
# Gravity preset for Earth
# GRAVITY_EARTH: Tuple[float, float, float] = (0.0, 0.0, -9.80665)
# GRAVITY_EARTH_STD: Tuple[float, float, float] = (0.0, 0.0, 0.0232)

############################
# Additional Configuration #
############################
# BROADCAST_GUI: bool = str2bool(
#     environ.get("RBS_ENVS_BROADCAST_INTERACTIVE_GUI", default=True)
# )


#########
# Reach #
#########
REACH_MAX_EPISODE_STEPS: int = 100
REACH_KWARGS: Dict[str, Any] = {
    "agent_rate": 4.0,
    "robot_model": RBS_ENVS_ROBOT_MODEL,
    "workspace_frame_id": "world",
    "workspace_centre": (-0.45, 0.0, 0.35),
    "workspace_volume": (0.7, 0.7, 0.7),
    "ignore_new_actions_while_executing": False,
    "use_servo": True,
    "scaling_factor_translation": 8.0,
    "scaling_factor_rotation": 3.0,
    "restrict_position_goal_to_workspace": False,
    "enable_gripper": False,
    "sparse_reward": False,
    "collision_reward": -10.0,
    "act_quick_reward": -0.01,
    "required_accuracy": 0.05,
    "num_threads": 3,
}
REACH_KWARGS_SIM: Dict[str, Any] = {
    "physics_rate": 100,
    "real_time_factor": float(np.finfo(np.float32).max),
    "world": get_world_file("ungravitational"),
}


REACH_RANDOMIZER: str = "rbs_gym.envs.randomizers:ManipulationGazeboEnvRandomizer"

SCENE_CONFIGURATION: SceneData = SceneData(
    physics_rollouts_num=0,
    robot=RobotData(
        name="rbs_arm",
        joint_positions=[0.0, 0.5, 3.14159, 1.5, 0.0, 1.4, 0.0],
        with_gripper=True,
        gripper_joint_positions=0.00,
        randomizer=RobotRandomizerData(joint_positions=True),
    ),
    objects=[
        SphereObjectData(
            name="sphere",
            type="sphere",
            relative_to="base_link",
            position=(0.0, 0.3, 0.5),
            static=True,
            collision=False,
            color=(0.0, 1.0, 0.0, 0.8),
            randomize=ObjectRandomizerData(random_pose=True, models_rollouts_num=2),
        )
    ],
    plugins=PluginsData(
        scene_broadcaster=True, user_commands=True, fts_broadcaster=True
    ),
)


# Task
register(
    id="Reach-v0",
    entry_point=RBS_ENVS_TASK_ENTRYPOINT,
    max_episode_steps=REACH_MAX_EPISODE_STEPS,
    kwargs={
        "task_cls": tasks.Reach,
        **REACH_KWARGS,
    },
)
register(
    id="Reach-ColorImage-v0",
    entry_point=RBS_ENVS_TASK_ENTRYPOINT,
    max_episode_steps=REACH_MAX_EPISODE_STEPS,
    kwargs={
        "task_cls": tasks.ReachColorImage,
        **REACH_KWARGS,
    },
)
register(
    id="Reach-DepthImage-v0",
    entry_point=RBS_ENVS_TASK_ENTRYPOINT,
    max_episode_steps=REACH_MAX_EPISODE_STEPS,
    kwargs={
        "task_cls": tasks.ReachDepthImage,
        **REACH_KWARGS,
    },
)

# Gazebo wrapper
register(
    id="Reach-Gazebo-v0",
    entry_point=REACH_RANDOMIZER,
    max_episode_steps=REACH_MAX_EPISODE_STEPS,
    kwargs={"env": "Reach-v0", **REACH_KWARGS_SIM, "scene_args": SCENE_CONFIGURATION},
)
register(
    id="Reach-ColorImage-Gazebo-v0",
    entry_point=REACH_RANDOMIZER,
    max_episode_steps=REACH_MAX_EPISODE_STEPS,
    kwargs={
        "env": "Reach-ColorImage-v0",
        **REACH_KWARGS_SIM,
        "scene_args": SCENE_CONFIGURATION,
    },
)
register(
    id="Reach-DepthImage-Gazebo-v0",
    entry_point=REACH_RANDOMIZER,
    max_episode_steps=REACH_MAX_EPISODE_STEPS,
    kwargs={
        "env": "Reach-DepthImage-v0",
        **REACH_KWARGS_SIM,
        "scene_args": SCENE_CONFIGURATION,
    },
)
