from typing import Optional
from control_msgs.action import GripperCommand
from rclpy.action import ActionClient


class GripperController:
    def __init__(self, node,
                 open_position: Optional[float] = 0.0,
                 close_position: Optional[float] = 0.0,
                 max_effort: Optional[float] = 0.0,
                 namespace: Optional[str] = ""):
        self._action_client = ActionClient(node, GripperCommand,
                                           namespace + "/gripper_controller/gripper_cmd")
        self._open_position = open_position
        self._close_position = close_position
        self._max_effort = max_effort
        self.is_executed = False

    def open(self):
        self.send_goal(self._open_position)

    def close(self):
        self.send_goal(self._close_position)

    def send_goal(self, goal: float):
        goal_msg = GripperCommand.Goal()
        goal_msg._command.position = goal
        goal_msg._command.max_effort = self._max_effort
        self._action_client.wait_for_server()
        self._send_goal_future = self._action_client.send_goal_async(goal_msg)
        self._send_goal_future.add_done_callback(self.goal_response_callback)

    def goal_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.get_logger().info('Goal rejected :(')
            return

        self.get_logger().info('Goal accepted :)')

        self._get_result_future = goal_handle.get_result_async()
        self._get_result_future.add_done_callback(self.get_result_callback)

    def get_result_callback(self, future):
        result = future.result().result
        self.get_logger().info(f"Gripper position: {result.position}")

    def wait_until_executed(self):
        while not self.is_executed:
            pass
