from array import array
from threading import Lock
from typing import Optional

from rclpy.callback_groups import CallbackGroup
from rclpy.node import Node
from rclpy.qos import (
    QoSDurabilityPolicy,
    QoSHistoryPolicy,
    QoSProfile,
    QoSReliabilityPolicy,
)

from sensor_msgs.msg import JointState

class JointStates:
    def __init__(
        self,
        node: Node,
        topic: str,
        callback_group: Optional[CallbackGroup] = None,
    ):

        self._node = node

        self.__observation = JointState()
        self._node.create_subscription(
            msg_type=JointState,
            topic=topic,
            callback=self.observation_callback,
            qos_profile=QoSProfile(
                reliability=QoSReliabilityPolicy.RELIABLE,
                durability=QoSDurabilityPolicy.VOLATILE,
                history=QoSHistoryPolicy.KEEP_LAST,
                depth=1,
            ),
            callback_group=callback_group,
        )
        self.__observation_mutex = Lock()
        self.__new_observation_available = False

        self.__observation.position

    def observation_callback(self, msg):
        """
        Callback for getting observation.
        """

        self.__observation_mutex.acquire()
        self.__observation = msg
        self.__new_observation_available = True
        self._node.get_logger().debug("New observation received.")
        self.__observation_mutex.release()

    def get_observation(self) -> JointState:
        """
        Get the last received observation.
        """
        self.__observation_mutex.acquire()
        observation = self.__observation
        self.__observation_mutex.release()
        return observation

    def get_positions(self) -> array:
        """
        Get the last recorded observation position
        """
        self.__observation_mutex.acquire()
        observation = self.__observation.position
        self.__observation_mutex.release()
        return observation

    def get_velocities(self) -> array:
        """
        Get the last recorded observation velocity
        """
        self.__observation_mutex.acquire()
        observation = self.__observation.velocity
        self.__observation_mutex.release()
        return observation

    def get_efforts(self) -> array:
        """
        Get the last recorded observation effort
        """
        self.__observation_mutex.acquire()
        observation = self.__observation.effort
        self.__observation_mutex.release()
        return observation

    def reset_new_observation_checker(self):
        """
        Reset checker of new observations, i.e. `self.new_observation_available()`
        """

        self.__observation_mutex.acquire()
        self.__new_observation_available = False
        self.__observation_mutex.release()

    @property
    def new_observation_available(self):
        """
        Check if new observation is available since `self.reset_new_observation_checker()` was called
        """

        return self.__new_observation_available

