import abc

from env_manager.models.configs import (
    SceneData,
)
from env_manager.scene import Scene
from env_manager.utils.logging import set_log_level
from gym_gz import randomizers
from gym_gz.randomizers import gazebo_env_randomizer
from gym_gz.randomizers.gazebo_env_randomizer import MakeEnvCallable
from scenario import gazebo as scenario
from scenario.bindings.gazebo import GazeboSimulator

from rbs_gym.envs import tasks

SupportedTasks = tasks.Reach | tasks.ReachColorImage | tasks.ReachDepthImage


class ManipulationGazeboEnvRandomizer(
    gazebo_env_randomizer.GazeboEnvRandomizer,
    randomizers.abc.PhysicsRandomizer,
    randomizers.abc.TaskRandomizer,
    abc.ABC,
):
    """
    Basic randomizer of environments for robotic manipulation inside Ignition Gazebo. This randomizer
    also populates the simulated world with robot, terrain, lighting and other entities.
    """

    POST_RANDOMIZATION_MAX_STEPS = 50

    metadata = {"render_modes": ["human"]}

    def __init__(
        self,
        env: MakeEnvCallable,
        scene_args: SceneData = SceneData(),
        **kwargs,
    ):
        self._scene_data = scene_args
        # TODO (a lot of work): Implement proper physics randomization.
        if scene_args.physics_rollouts_num != 0:
            raise TypeError(
                "Proper physics randomization at each reset is not yet implemented. Please set `physics_rollouts_num=0`."
            )

        # Update kwargs before passing them to the task constructor (some tasks might need them)
        # TODO: update logic when will need choose between cameras
        cameras: list[dict[str, str | int]] = []
        for camera in self._scene_data.camera:
            camera_info: dict[str, str | int] = {}
            camera_info["name"] = camera.name
            camera_info["type"] = camera.type
            camera_info["width"] = camera.width
            camera_info["height"] = camera.height
            cameras.append(camera_info)

        kwargs.update({"camera_info": cameras})

        # Initialize base classes
        randomizers.abc.TaskRandomizer.__init__(self)
        randomizers.abc.PhysicsRandomizer.__init__(
            self, randomize_after_rollouts_num=scene_args.physics_rollouts_num
        )
        gazebo_env_randomizer.GazeboEnvRandomizer.__init__(
            self, env=env, physics_randomizer=self, **kwargs
        )

        self.__env_initialised = False

    ##########################
    # PhysicsRandomizer impl #
    ##########################

    def init_physics_preset(self, task: SupportedTasks):
        self.set_gravity(task=task)

    def randomize_physics(self, task: SupportedTasks, **kwargs):
        self.set_gravity(task=task)

    def set_gravity(self, task: SupportedTasks):
        if not task.world.to_gazebo().set_gravity(
            (
                task.np_random.normal(loc=self._gravity[0], scale=self._gravity_std[0]),
                task.np_random.normal(loc=self._gravity[1], scale=self._gravity_std[1]),
                task.np_random.normal(loc=self._gravity[2], scale=self._gravity_std[2]),
            )
        ):
            raise RuntimeError("Failed to set the gravity")

    def get_engine(self):
        return scenario.PhysicsEngine_dart

    #######################
    # TaskRandomizer impl #
    #######################

    def randomize_task(self, task: SupportedTasks, **kwargs):
        """
        Randomization of the task, which is called on each reset of the environment.
        Note that this randomizer reset is called before `reset_task()`.
        """

        # Get gazebo instance associated with the task
        if "gazebo" not in kwargs:
            raise ValueError("Randomizer does not have access to the gazebo interface")
        if isinstance(kwargs["gazebo"], GazeboSimulator):
            gazebo: GazeboSimulator = kwargs["gazebo"]
        else:
            raise RuntimeError("Provide GazeboSimulator instance")

        # Initialise the environment on the first iteration
        if not self.__env_initialised:
            self.init_env(task=task, gazebo=gazebo)
            self.__env_initialised = True

        # Perform pre-randomization steps
        self.pre_randomization(task=task)

        self._scene.reset_scene()

        # Perform post-randomization steps
        # TODO: Something in post-randomization causes GUI client to freeze during reset (the simulation server still works fine)
        self.post_randomization(task, gazebo)

    ###################
    # Randomizer impl #
    ###################

    # Initialisation #
    def init_env(self, task: SupportedTasks, gazebo: scenario.GazeboSimulator):
        """
        Initialise an instance of the environment before the very first iteration
        """

        self._scene = Scene(
            task,
            gazebo,
            self._scene_data,
            task.get_parameter("robot_description").get_parameter_value().string_value,
        )

        # Init Scene for task
        task.scene = self._scene
        # Set log level for (Gym) Ignition
        set_log_level(log_level=task.get_logger().get_effective_level().name)

        self._scene.init_scene()

    # Pre-randomization #
    def pre_randomization(self, task: SupportedTasks):
        """
        Perform steps that are required before randomization is performed.
        """

        for obj in self._scene.objects:
            # If desired, select random spawn position from the segments
            # It is performed here because object spawn position might be of interest also for robot and camera randomization

            segments_len = len(obj.randomize.random_spawn_position_segments)
            if segments_len > 1:
                # Randomly select a segment between two points
                start_index = task.np_random.randint(segments_len - 1)
                segment = (
                    obj.randomize.random_spawn_position_segments[start_index],
                    obj.randomize.random_spawn_position_segments[start_index + 1],
                )

                # Randomly select a point on the segment and use it as the new object spawn position
                intersect = task.np_random.random()
                direction = (
                    segment[1][0] - segment[0][0],
                    segment[1][1] - segment[0][1],
                    segment[1][2] - segment[0][2],
                )
                obj.position = (
                    segment[0][0] + intersect * direction[0],
                    segment[0][1] + intersect * direction[1],
                    segment[0][2] + intersect * direction[2],
                )

        # TODO: add bounding box with multiple objects

        # # Update also the workspace centre (and bounding box) if desired
        # if self._object_random_spawn_position_update_workspace_centre:
        #     task.workspace_centre = (
        #         self._object_spawn_position[0],
        #         self._object_spawn_position[1],
        #         # Z workspace is currently kept the same on purpose
        #         task.workspace_centre[2],
        #     )
        #     workspace_volume_half = (
        #         task.workspace_volume[0] / 2,
        #         task.workspace_volume[1] / 2,
        #         task.workspace_volume[2] / 2,
        #     )
        #     task.workspace_min_bound = (
        #         task.workspace_centre[0] - workspace_volume_half[0],
        #         task.workspace_centre[1] - workspace_volume_half[1],
        #         task.workspace_centre[2] - workspace_volume_half[2],
        #     )
        #     task.workspace_max_bound = (
        #         task.workspace_centre[0] + workspace_volume_half[0],
        #         task.workspace_centre[1] + workspace_volume_half[1],
        #         task.workspace_centre[2] + workspace_volume_half[2],
        #     )

    def post_randomization(
        self, task: SupportedTasks, gazebo: scenario.GazeboSimulator
    ):
        """
        Perform steps required once randomization is complete and the simulation can be stepped unpaused.
        """

        def perform_gazebo_step():
            if not gazebo.step():
                raise RuntimeError("Failed to execute an unpaused Gazebo step")

        def wait_for_new_observations():
            attempts = 0
            while True:
                attempts += 1
                if attempts % self.POST_RANDOMIZATION_MAX_STEPS == 0:
                    task.get_logger().debug(
                        f"Waiting for new joint state after reset. Iteration #{attempts}..."
                    )
                else:
                    task.get_logger().debug("Waiting for new joint state after reset.")

                perform_gazebo_step()

                # If camera_sub is defined, ensure all new observations are available
                if hasattr(task, "camera_subs") and not all(
                    sub.new_observation_available for sub in task.camera_subs
                ):
                    continue

                return  # Observations are ready

        attempts = 0
        processed_objects = set()

        # Early exit if the maximum number of steps is already reached
        if self.POST_RANDOMIZATION_MAX_STEPS == 0:
            task.get_logger().error(
                "Robot keeps falling through the terrain. There is something wrong..."
            )
            return

        # Ensure no objects are overlapping
        for obj in self._scene.objects:
            if not obj.randomize.random_pose or obj.name in processed_objects:
                continue

            # Try repositioning until no overlap or maximum attempts reached
            for _ in range(self.POST_RANDOMIZATION_MAX_STEPS):
                task.get_logger().debug(
                    f"Checking overlap for {obj.name}, attempt {attempts + 1}"
                )
                if self._scene.check_object_overlapping(obj):
                    processed_objects.add(obj.name)
                    break  # No overlap, move to next object
                else:
                    task.get_logger().debug(
                        f"Objects overlapping, trying new positions for {obj.name}"
                    )

                perform_gazebo_step()

            else:
                task.get_logger().warn(
                    f"Could not place {obj.name} without overlap after {self.POST_RANDOMIZATION_MAX_STEPS} attempts"
                )
                continue  # Move to next object in case of failure

        # Execute steps until new observations are available
        if hasattr(task, "camera_subs") or task._enable_gripper:
            wait_for_new_observations()

        # Final check if observations are not available within the maximum steps
        if self.POST_RANDOMIZATION_MAX_STEPS == attempts:
            task.get_logger().error("Cannot obtain new observation.")

    # =============================
    # Additional features and debug
    # =============================

    # def visualise_workspace(
    #     self,
    #     task: SupportedTasks,
    #     gazebo: scenario.GazeboSimulator,
    #     color: Tuple[float, float, float, float] = (0, 1, 0, 0.8),
    # ):
    #     # Insert a translucent box visible only in simulation with no physical interactions
    #     models.Box(
    #         world=task.world,
    #         name="_workspace_volume",
    #         position=self._object_spawn_position,
    #         orientation=(0, 0, 0, 1),
    #         size=task.workspace_volume,
    #         collision=False,
    #         visual=True,
    #         gui_only=True,
    #         static=True,
    #         color=color,
    #     )
    #     # Execute a paused run to process model insertion
    #     if not gazebo.run(paused=True):
    #         raise RuntimeError("Failed to execute a paused Gazebo run")
    #
    # def visualise_spawn_volume(
    #     self,
    #     task: SupportedTasks,
    #     gazebo: scenario.GazeboSimulator,
    #     color: Tuple[float, float, float, float] = (0, 0, 1, 0.8),
    #     color_with_height: Tuple[float, float, float, float] = (1, 0, 1, 0.7),
    # ):
    #     # Insert translucent boxes visible only in simulation with no physical interactions
    #     models.Box(
    #         world=task.world,
    #         name="_object_random_spawn_volume",
    #         position=self._object_spawn_position,
    #         orientation=(0, 0, 0, 1),
    #         size=self._object_random_spawn_volume,
    #         collision=False,
    #         visual=True,
    #         gui_only=True,
    #         static=True,
    #         color=color,
    #     )
    #     models.Box(
    #         world=task.world,
    #         name="_object_random_spawn_volume_with_height",
    #         position=self._object_spawn_position,
    #         orientation=(0, 0, 0, 1),
    #         size=self._object_random_spawn_volume,
    #         collision=False,
    #         visual=True,
    #         gui_only=True,
    #         static=True,
    #         color=color_with_height,
    #     )
    #     # Execute a paused run to process model insertion
    #     if not gazebo.run(paused=True):
    #         raise RuntimeError("Failed to execute a paused Gazebo run")
