import abc

import gymnasium as gym
import numpy as np
from gym_gz.utils.typing import Observation, ObservationSpace

from rbs_gym.envs.models.sensors import Camera
from rbs_gym.envs.observation import CameraSubscriber
from rbs_gym.envs.tasks.reach import Reach


class ReachColorImage(Reach, abc.ABC):
    def __init__(
        self,
        camera_info: list[dict] = [],
        monochromatic: bool = False,
        **kwargs,
    ):
        # Initialize the Task base class
        Reach.__init__(
            self,
            **kwargs,
        )

        # Store parameters for later use
        self._camera_width = sum(w.get("width", 0) for w in camera_info)
        self._camera_height = camera_info[0]["height"]
        self._monochromatic = monochromatic
        self._saved = False

        self.camera_subs: list[CameraSubscriber] = []
        for camera in camera_info:
            # Perception (RGB camera)
            self.camera_subs.append(
                CameraSubscriber(
                    node=self,
                    topic=Camera.get_color_topic(camera["name"], camera["type"]),
                    is_point_cloud=False,
                    callback_group=self._callback_group,
                )
            )

    def create_observation_space(self) -> ObservationSpace:
        """
        Creates the observation space for the Imitation Learning algorithm.

        This method returns a dictionary-based observation space that includes:

        - **image**: A 2D image captured from the robot's camera. The image's shape is determined by the camera's
          width and height, and the number of channels (either monochromatic or RGB).
          - Dimensions: `(camera_height, camera_width, channels)` where `channels` is 1 for monochromatic images
            and 3 for RGB images.
          - Pixel values are in the range `[0, 255]` with data type `np.uint8`.

        - **joint_states**: The joint positions of the robot arm, represented as continuous values within the
          range `[-1.0, 1.0]`, where `nDoF` refers to the number of degrees of freedom of the robot arm.

        Returns:
            gym.spaces.Dict: A dictionary defining the observation space for the learning algorithm,
            containing both image and joint_states information.
        """
        return gym.spaces.Dict(
            {
                "image": gym.spaces.Box(
                    low=0,
                    high=255,
                    shape=(
                        self._camera_height,
                        self._camera_width,
                        1 if self._monochromatic else 3,
                    ),
                    dtype=np.uint8,
                ),
                "joint_states": gym.spaces.Box(low=-1.0, high=1.0),
            }
        )

    def get_observation(self) -> Observation:
        # Get the latest images
        image = []

        for sub in self.camera_subs:
            image.append(sub.get_observation())

        image_width = sum(i.width for i in image)
        image_height = image[0].height
        # if image_width == self._camera_width and image_height == self._camera_height:

        # image_data = np.concatenate([i.data for i in image], axis=1)

        assert (
            image_width == self._camera_width and image_height == self._camera_height
        ), f"Error: Resolution of the input image does not match the configured observation space. ({image_width}x{image_height} instead of {self._camera_width}x{self._camera_height})"

        # Reshape and create the observation
        # color_image = np.array([i.data for i in image], dtype=np.uint8).reshape(
        #     self._camera_height, self._camera_width, 3
        # )
        color_image = np.concatenate(
            [
                np.array(i.data, dtype=np.uint8).reshape(i.height, i.width, 3)
                for i in image
            ],
            axis=1,
        )

        # # Debug save images
        # from PIL import Image
        # img_color = Image.fromarray(color_image)
        # img_color.save("img_color.png")

        if self._monochromatic:
            observation = Observation(color_image[:, :, 0])
        else:
            observation = Observation(color_image)

        self.get_logger().debug(f"\nobservation: {observation}")

        # Return the observation
        return observation
