import abc
from typing import Tuple

from geometry_msgs.msg import WrenchStamped
import gymnasium as gym
import numpy as np
from gym_gz.utils.typing import (
    Action,
    ActionSpace,
    Observation,
    ObservationSpace,
    Reward,
)

from rbs_gym.envs.tasks.manipulation import Manipulation
from env_manager.utils.math import distance_to_nearest_point
from std_msgs.msg import Float64MultiArray
from rbs_gym.envs.observation import TwistSubscriber, JointStates
from env_manager.utils.helper import get_object_position


class Reach(Manipulation, abc.ABC):
    def __init__(
        self,
        sparse_reward: bool,
        collision_reward: float,
        act_quick_reward: float,
        required_accuracy: float,
        **kwargs,
    ):
        # Initialize the Task base class
        Manipulation.__init__(
            self,
            **kwargs,
        )

        # Additional parameters
        self._sparse_reward: bool = sparse_reward
        self._act_quick_reward = (
            act_quick_reward if act_quick_reward >= 0.0 else -act_quick_reward
        )
        self._collision_reward = (
            collision_reward if collision_reward >= 0.0 else -collision_reward
        )
        self._required_accuracy: float = required_accuracy

        # Flag indicating if the task is done (performance - get_reward + is_done)
        self._is_done: bool = False
        self._is_truncated: bool = False
        self._is_terminated: bool = False

        # Distance to target in the previous step (or after reset)
        self._previous_distance: float = 0.0
        # self._collision_reward: float = collision_reward

        # self.initial_gripper_joint_positions = self.robot_data.gripper_joint_positions
        self.twist = TwistSubscriber(
            self,
            topic="/cartesian_force_controller/current_twist",
            callback_group=self._callback_group,
        )

        self.joint_states = JointStates(
            self, topic="/joint_states", callback_group=self._callback_group
        )

        self._action = WrenchStamped()
        self._action_array: Action = np.array([])
        self.followed_object_name = "sphere"

    def create_action_space(self) -> ActionSpace:
        # 0:3 - (x, y, z) force
        # 3:6 - (x, y, z) torque
        return gym.spaces.Box(low=-1.0, high=1.0, shape=(3,), dtype=np.float32)

    def create_observation_space(self) -> ObservationSpace:
        # 0:3 - (x, y, z) end effector position
        # 3:6 - (x, y, z) target position
        # 6:9 - (x, y, z) current twist
        # Note: These could theoretically be restricted to the workspace and object spawn area instead of inf
        return gym.spaces.Box(low=-np.inf, high=np.inf, shape=(12,), dtype=np.float32)

    def set_action(self, action: Action):
        # self.get_logger().debug(f"action: {action}")
        # act = Float64MultiArray()
        # act.data = action
        # self.joint_controller.publisher.publish(act)
        if isinstance(action, np.number):
            raise RuntimeError("For the task Reach the action space should be array")

        # Store action for reward function
        self._action_array = action

        # self._action.header.frame_id = self.robot_ee_link_name
        self._action.header.stamp = self.get_clock().now().to_msg()
        self._action.header.frame_id = self.robot.ee_link
        self._action.wrench.force.x = float(action[0]) * 30.0
        self._action.wrench.force.y = float(action[1]) * 30.0
        self._action.wrench.force.z = float(action[2]) * 30.0
        # self._action.wrench.torque.x = float(action[3]) * 10.0
        # self._action.wrench.torque.y = float(action[4]) * 10.0
        # self._action.wrench.torque.z = float(action[5]) * 10.0
        self.controller.publisher.publish(self._action)

    def get_observation(self) -> Observation:
        # Get current end-effector and target positions
        self.object_names = []
        ee_position = self.get_ee_position()
        target_position = self.get_object_position(self.followed_object_name)
        # target_position = self.get_object_position(object_model=self.object_names[0])
        # joint_states = tuple(self.joint_states.get_positions())

        # self.get_logger().warn(f"joint_states: {joint_states[:7]}")

        twist = self.twist.get_observation()
        twt: tuple[float, float, float, float, float, float] = (
            twist.twist.linear.x,
            twist.twist.linear.y,
            twist.twist.linear.z,
            twist.twist.angular.x,
            twist.twist.angular.y,
            twist.twist.angular.z,
        )

        observation: Observation = np.concatenate([ee_position, target_position, twt], dtype=np.float32)
        # Create the observation
        # observation = Observation(
        #     np.concatenate([ee_position, target_position, twt], dtype=np.float32)
        # )

        self.get_logger().debug(f"\nobservation: {observation}")

        # Return the observation
        return observation

    def get_reward(self) -> Reward:
        reward = 0.0

        # Compute the current distance to the target
        current_distance = self.get_distance_to_target()

        # Mark the episode done if target is reached
        if current_distance < self._required_accuracy:
            self._is_terminated = True
            reward += 1.0 if self._sparse_reward else 0.0  # 100.0
        self.get_logger().debug(f"reward_target: {reward}")

        # Give reward based on how much closer robot got relative to the target for dense reward
        if not self._sparse_reward:
            distance_delta = self._previous_distance - current_distance
            reward += distance_delta * 10.0
            self._previous_distance = current_distance
            self.get_logger().debug(f"reward_distance: {reward}")

        if self.check_terrain_collision():
            reward -= self._collision_reward
            self._is_truncated = True
            self.get_logger().debug(f"reward_collision: {reward}")

        # Reward control
        # reward -= np.abs(self._action_array).sum() * 0.01
        # self.get_logger().debug(f"reward_c: {reward}")

        # Subtract a small reward each step to provide incentive to act quickly (if enabled)
        reward += self._act_quick_reward

        self.get_logger().debug(f"reward: {reward}")

        return Reward(reward)

    def is_terminated(self) -> bool:
        self.get_logger().debug(f"terminated: {self._is_terminated}")
        return self._is_terminated

    def is_truncated(self) -> bool:
        self.get_logger().debug(f"truncated: {self._is_truncated}")
        return self._is_truncated

    def reset_task(self):
        Manipulation.reset_task(self)

        self._is_done = False
        self._is_truncated = False
        self._is_terminated = False

        # Compute and store the distance after reset if using dense reward
        if not self._sparse_reward:
            self._previous_distance = self.get_distance_to_target()

        self.get_logger().debug("\ntask reset")

    def get_distance_to_target(self) -> float:
        ee_position = self.get_ee_position()
        object_position = self.get_object_position(object_model=self.followed_object_name)
        self.tf2_broadcaster.broadcast_tf(
            "world", "object", object_position, (0.0, 0.0, 0.0, 1.0), xyzw=True
        )

        return distance_to_nearest_point(origin=ee_position, points=[object_position])
