import abc

import gymnasium as gym
import numpy as np
from gym_gz.utils.typing import Observation, ObservationSpace

from env_manager.models.sensors import Camera
from rbs_gym.envs.observation import CameraSubscriber
from rbs_gym.envs.tasks.reach import Reach


class ReachColorImage(Reach, abc.ABC):
    def __init__(
        self,
        camera_info: list[dict] = [],
        monochromatic: bool = False,
        **kwargs,
    ):

        # Initialize the Task base class
        Reach.__init__(
            self,
            **kwargs,
        )

        # Store parameters for later use
        self._camera_width = sum(w.get('width', 0) for w in camera_info)
        self._camera_height = camera_info[0]["height"]
        self._monochromatic = monochromatic
        self._saved = False

        self.camera_subs: list[CameraSubscriber] = []
        for camera in camera_info:
            # Perception (RGB camera)
            self.camera_subs.append(CameraSubscriber(
                node=self,
                topic=Camera.get_color_topic(camera["name"], camera["type"]),
                is_point_cloud=False,
                callback_group=self._callback_group,
            ))

    def create_observation_space(self) -> ObservationSpace:

        # 0:3*height*width - rgb image
        # 0:1*height*width - monochromatic (intensity) image
        return gym.spaces.Box(
            low=0,
            high=255,
            shape=(
                self._camera_height,
                self._camera_width,
                1 if self._monochromatic else 3,
            ),
            dtype=np.uint8,
        )

    def get_observation(self) -> Observation:
        # Get the latest images
        image = []

        for sub in self.camera_subs:
            image.append(sub.get_observation())

        image_width = sum(i.width for i in image)
        image_height = image[0].height

        assert (
            image_width == self._camera_width and image_height == self._camera_height
        ), f"Error: Resolution of the input image does not match the configured observation space. ({image_width}x{image_height} instead of {self._camera_width}x{self._camera_height})"

        color_image = np.concatenate(
            [np.array(i.data, dtype=np.uint8).reshape(i.height, i.width, 3) for i in image],
            axis=1
        )

        # # Debug save images
        # from PIL import Image
        # img_color = Image.fromarray(color_image)
        # img_color.save("img_color.png")

        if self._monochromatic:
            observation = Observation(color_image[:, :, 0])
        else:
            observation = Observation(color_image)

        self.get_logger().debug(f"\nobservation: {observation}")

        # Return the observation
        return observation
