#!/usr/bin/python3

import rclpy
from ament_index_python.packages import get_package_share_directory
from env_manager.scene import Scene
from env_manager_interfaces.srv import ResetEnv
from rbs_assets_library import get_world_file
from rclpy.node import Node
from scenario.bindings.gazebo import GazeboSimulator
from rclpy.executors import MultiThreadedExecutor, ExternalShutdownException

from .. import scene_config_loader

class GazeboRuntime(Node):
    def __init__(self) -> None:
        super().__init__(node_name="rbs_gz_runtime")
        self.declare_parameter("robot_description", "")
        self.declare_parameter(
            "config_file",
            get_package_share_directory("rbs_runtime")
            + "/config/default-scene-config.yaml",
        )

        self.gazebo = GazeboSimulator(step_size=0.001, rtf=1.0, steps_per_run=1)

        self.gazebo.insert_world_from_sdf(get_world_file("default"))
        if not self.gazebo.initialize():
            raise RuntimeError("Gazebo cannot be initialized")

        config_file = (
            self.get_parameter("config_file").get_parameter_value().string_value
        )

        scene_data = scene_config_loader(config_file)

        self.scene = Scene(
            node=self,
            gazebo=self.gazebo,
            scene=scene_data,
            robot_urdf_string=self.get_parameter("robot_description")
            .get_parameter_value()
            .string_value,
        )
        self.scene.init_scene()

        self.reset_env_srv = self.create_service(
            ResetEnv, "/env_manager/reset_env", self.reset_env
        )
        self.is_env_reset = False
        self.timer = self.create_timer(self.gazebo.step_size(), self.gazebo_run)

    def gazebo_run(self):
        if self.is_env_reset:
            self.scene.reset_scene()
            self.is_env_reset = False
        self.gazebo.run()

    def reset_env(self, req, res):
        self.is_env_reset = True
        res.ok = self.is_env_reset
        return res


def main():
    rclpy.init()

    executor = MultiThreadedExecutor()
    node = GazeboRuntime()
    executor.add_node(node)
    try:
        executor.spin()
    except (KeyboardInterrupt, ExternalShutdownException):
        node.destroy_node()
