#include "geometry_msgs/msg/pose_stamped.hpp"
#include "moveit_msgs/msg/constraints.hpp"
#include "moveit_msgs/msg/orientation_constraint.hpp"
#include "rbs_skill_interfaces/action/moveit_send_pose.hpp"
#include "rbs_skill_interfaces/action/moveit_send_pose_with_constraints.hpp"
#include "rbs_skill_servers/base_skill.hpp"
#include <memory>
#include <moveit/move_group_interface/move_group_interface.hpp>
#include <moveit/utils/moveit_error_code.hpp>
#include <rclcpp/logging.hpp>


namespace rbs_skill_actions {

using MoveitSendPoseOrientationConstraint = rbs_skill_interfaces::action::MoveitSendPoseWithConstraints;

class MoveitMtpOrientationConstraint : public SkillBase<MoveitSendPoseOrientationConstraint> {
public:
  explicit MoveitMtpOrientationConstraint(const rclcpp::NodeOptions &options = rclcpp::NodeOptions())
      : SkillBase("mtp_moveit_orientation_constraint", options) {
    this->declare_parameter("base_link", "");
    this->declare_parameter("ee_link", "");
  }

private:
  std::shared_ptr<moveit::planning_interface::MoveGroupInterface> move_group_;

  std::string requiredActionController() override {
    return "joint_trajectory_controller";
  }

  void executeAction() override {
    RCLCPP_INFO(this->get_logger(), "Executing action goal");
    move_group_ =
        std::make_shared<moveit::planning_interface::MoveGroupInterface>(
            this->shared_from_this(), m_current_goal->robot_name);

    move_group_->startStateMonitor();

    geometry_msgs::msg::PoseStamped pose_stamped;
    pose_stamped.pose = m_current_goal->target_pose;
    pose_stamped.header.frame_id = this->get_parameter("base_link").as_string();
    pose_stamped.header.stamp = this->get_clock()->now();

    moveit_msgs::msg::OrientationConstraint orientation_constraint;
    orientation_constraint.weight = m_current_goal->orientation_constraint.weight;
    orientation_constraint.absolute_x_axis_tolerance = m_current_goal->orientation_constraint.absolute_x_axis_tolerance;
    orientation_constraint.absolute_y_axis_tolerance = m_current_goal->orientation_constraint.absolute_y_axis_tolerance;
    orientation_constraint.absolute_z_axis_tolerance = m_current_goal->orientation_constraint.absolute_z_axis_tolerance;
    orientation_constraint.orientation = pose_stamped.pose.orientation;
    orientation_constraint.header = pose_stamped.header;
    orientation_constraint.link_name = this->get_parameter("ee_link").as_string();

    moveit_msgs::msg::Constraints constraints;
    constraints.orientation_constraints.emplace_back(orientation_constraint);

    move_group_->setPoseTarget(pose_stamped);
    move_group_->setPathConstraints(constraints);
    move_group_->setMaxVelocityScalingFactor(
        m_current_goal->end_effector_velocity);
    move_group_->setMaxAccelerationScalingFactor(
        m_current_goal->end_effector_acceleration);
    move_group_->setPlanningTime(10.0);

    moveit::planning_interface::MoveGroupInterface::Plan plan;
    moveit::core::MoveItErrorCode plan_err_code = move_group_->plan(plan);
    if (plan_err_code == moveit::core::MoveItErrorCode::INVALID_MOTION_PLAN) {
      move_group_->plan(plan);
    }
    if (plan_err_code == moveit::core::MoveItErrorCode::SUCCESS) {
      RCLCPP_INFO(this->get_logger(), "Planning success");
      moveit::core::MoveItErrorCode move_err_code = move_group_->execute(plan);
      if (move_err_code == moveit::core::MoveItErrorCode::SUCCESS) {
        m_current_result->success = true;
        m_current_goal_handle->succeed(m_current_result);
        RCLCPP_INFO(this->get_logger(), "Successfully executed action goal");
      } else {
        RCLCPP_ERROR_STREAM(
            this->get_logger(),
            "Failed to move because:"
                << moveit::core::errorCodeToString(move_err_code));
      }
    } else {
      RCLCPP_WARN_STREAM(this->get_logger(),
                         "Failed to generate plan because: "
                             << moveit::core::errorCodeToString(plan_err_code));
      m_current_result->success = false;
      m_current_goal_handle->abort(m_current_result);
    }

    if (m_current_goal_handle->is_canceling()) {
      m_current_result->success = false;
      m_current_goal_handle->canceled(m_current_result);
      RCLCPP_ERROR(this->get_logger(), "Action goal canceled");
      return;
    }
  }
}; // class MoveToPoseActionServer

} // namespace rbs_skill_actions

#include "rclcpp_components/register_node_macro.hpp"
RCLCPP_COMPONENTS_REGISTER_NODE(rbs_skill_actions::MoveitMtpOrientationConstraint);
