#!/usr/bin/env python3
"""
  recording_demo_lifecycle_node_via_RosBag
  ROS 2 program for recording a demo via RosBag

  @shalenikol release 0.3
  @shalenikol release 0.4 In "Jazzy" writer.close()
"""
import os
import shutil
import json
from functools import partial

import rclpy
from rclpy.lifecycle import Node
from rclpy.lifecycle import State
from rclpy.lifecycle import TransitionCallbackReturn

import rosbag2_py # 0.26.7 (jazzy)
from ros2topic.api import get_topic_names_and_types
from rclpy.serialization import serialize_message

from sensor_msgs.msg import JointState, Image, CompressedImage

from rbs_utils.recording_demo import CommandType

#'get_type_description': 'type_description_interfaces/srv/GetTypeDescription'

NODE_NAME_DEFAULT = "lc_record_demo"
PARAM_SKILL_CFG   = "lc_record_demo_cfg"

OUTPUT_PATH_DEFAULT = "rbs_bag"
RECORD_PATH = "episode"
WRITER_ID = "sqlite3"
TOPIC_FILTERS = {
    "sensor_msgs/msg/JointState": JointState,
    "sensor_msgs/msg/Image": Image,
}

class RecordingDemo(Node):
    """ lifecycle node """

    def __init__(self, **kwargs):
        self.mode: int = 0 # 0 - нет записи, 1 - есть запись
        self.topics = []
        self._sub = []
        self.output_path = OUTPUT_PATH_DEFAULT
        self.compressed = "False"  # default: not compressed image
        self.serialization = "cdr" # default serialization format

        # for other nodes
        kwargs["allow_undeclared_parameters"] = True
        kwargs["automatically_declare_parameters_from_overrides"] = True
        super().__init__(NODE_NAME_DEFAULT, **kwargs)
        str_cfg = self.get_parameter(PARAM_SKILL_CFG).get_parameter_value().string_value

        self.skill_cfg = json.loads(str_cfg)
        sets = self.skill_cfg["Settings"]["output"]["params"]
        for s in sets:
            nam = s["name"]
            val = s["value"]
            if nam == "output_path":
                self.output_path = val
            elif nam == "compressed":
                self.compressed = val
            elif nam == "serialization":
                self.serialization = val

        if self.compressed == "True":
            del TOPIC_FILTERS["sensor_msgs/msg/Image"]
            TOPIC_FILTERS["sensor_msgs/msg/CompressedImage"] = CompressedImage
        
    def get_list_topics(self) -> list:
        topic_names_and_types = get_topic_names_and_types(node=self, include_hidden_topics=False)

        topic_list = []
        for topic_name, topic_types in topic_names_and_types:
            topic_type = topic_types[0]
            if topic_type in TOPIC_FILTERS:
                topic_list.append({
                    "name": topic_name,
                    "type": topic_type,
                    "class": TOPIC_FILTERS[topic_type],
                })

        return topic_list

    def init_output_folder(self, path: str) -> str:
        # check for output folder existence
        if not os.path.isdir(path):
            os.makedirs(path)
        rec_folder = os.path.join(path, RECORD_PATH)

        if os.path.isdir(rec_folder):
            shutil.rmtree(rec_folder)
        return rec_folder

    def save_episode(self) -> None:
        if os.path.isdir(self.rec_folder):
            x = 1
            while os.path.isdir(self.rec_folder + str(x)):
                x += 1
            shutil.move(self.rec_folder, self.rec_folder + str(x))

    def on_configure(self, state: State) -> TransitionCallbackReturn:
        self.rec_folder = self.init_output_folder(self.output_path)
        self.writer = rosbag2_py.SequentialWriter()

        storage_options = rosbag2_py.StorageOptions(uri=self.rec_folder, storage_id=WRITER_ID)
        converter_options = rosbag2_py.ConverterOptions("", "")
        self.writer.open(storage_options, converter_options)
        
        self.topics = self.get_list_topics()
        for id, topic in enumerate(self.topics):
            # for type_hash
            pub_infos = self.get_publishers_info_by_topic(topic["name"])
            topic_info = rosbag2_py.TopicMetadata(
                id,
                name=topic["name"],
                type=topic["type"],
                serialization_format=self.serialization,
                type_description_hash=str(pub_infos[0].topic_type_hash))
            self.writer.create_topic(topic_info)
            self.get_logger().info(f"Topics: {id+1}) {topic}")

        self.get_logger().info("on_configure() is called.")
        return TransitionCallbackReturn.SUCCESS

    def on_activate(self, state: State) -> TransitionCallbackReturn:
        self.get_logger().info('on_activate() is called')
        # Create the subscribers.
        for id, topic in enumerate(self.topics):
            sub = self.create_subscription(topic["class"], topic["name"],
                                           partial(self.topic_callback, topic["name"]),
                                           10)
            self._sub.append(sub)

        self.mode += 1
        return super().on_activate(state)

    def on_deactivate(self, state: State) -> TransitionCallbackReturn:
        self.get_logger().info("on_deactivate() is called")
        # Destroy the subscribers.
        for sub in self._sub:
            self.destroy_subscription(sub)
        self._sub.clear()

        self.mode -= 1
        return super().on_deactivate(state)

    def on_cleanup(self, state: State) -> TransitionCallbackReturn:
        for sub in self._sub:
            self.destroy_subscription(sub)
        self._sub.clear()

        self.writer.close()

        do_param = CommandType.STOP_AND_SAVE # default

        json_param = self.get_parameter(PARAM_SKILL_CFG).get_parameter_value().string_value
        self.skill_cfg = json.loads(json_param)
        for comm in self.skill_cfg["BTAction"]:
            if comm["name"] == "rdStop":
                for par in comm["param"]:
                    if par["type"] == "formBuilder":
                        do_param = par["dependency"]["output"]["do"]
        
        if do_param in [CommandType.STOP_AND_SAVE, CommandType.SAVE_AND_NEXT]:
            self.save_episode()
        elif do_param == CommandType.STOP_AND_CANCEL:
            shutil.rmtree(self.rec_folder)

        self.get_logger().info("on_cleanup() is called")
        return TransitionCallbackReturn.SUCCESS

    def on_shutdown(self, state: State) -> TransitionCallbackReturn:
        for sub in self._sub:
            self.destroy_subscription(sub)
        self._sub.clear()

        self.get_logger().info("on_shutdown() is called")
        return TransitionCallbackReturn.SUCCESS
    
    def topic_callback(self, topic_name, msg):
        self.writer.write(
            topic_name,
            serialize_message(msg),
            self.get_clock().now().nanoseconds)

def main():
    rclpy.init()

    executor = rclpy.executors.SingleThreadedExecutor()
    lc_node = RecordingDemo()
    executor.add_node(lc_node)
    try:
        executor.spin()
    except (KeyboardInterrupt, rclpy.executors.ExternalShutdownException):
        lc_node.destroy_node()

if __name__ == '__main__':
    main()
